/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CROW_CANVASWID_H
#define CROW_CANVASWID_H

#include "widgets.h"
#include "editors.h"

namespace Crow {

class Paint : public Gtk::DrawingArea {

public:
	Paint(const Gdk::Color & f, const Gdk::Color & a) { frame= f; area= a; }

protected:
	bool on_expose_event(GdkEventExpose * event);

private:
	Gdk::Color frame;
	Gdk::Color area;

};

class WidgetCanvasEditorWidget : public EditorWidget {

public:
	WidgetCanvasEditorWidget();
	~WidgetCanvasEditorWidget();

	void setWidget(Gtk::Widget * widget);
	Gtk::Widget * getWidget();
	void unsetWidget();

	void setCursor(Gdk::CursorType cursor);
	void unsetCursor();

	int getGAP() { return GAP; }

	Gtk::EventBox * getGlass() { return &glass; }
	Gtk::EventBox * getGrabber() { return &grabber; }

	void pack(const Point & size=Point(-1, -1));

	void clearPaintList();
	void beginPaint();
	void paint(const Rectangle & R, bool fill, const Gdk::Color & frame, const Gdk::Color & area);
	void endPaint();

	void loadState(PObject state);
	void setupState(Gtk::Allocation &);
	PObject saveState();

private:
	int GAP;

	Gtk::ScrolledWindow scroll;
	Gtk::EventBox glass;
	Gtk::EventBox grabber;
	Gtk::Fixed fixed;
	Gtk::EventBox border;
	Gtk::Alignment container;

	typedef std::list<Paint *> PaintList;
	PaintList paintList;

	Gdk::CursorType lastCursor;

	class State : public Crow::Object {
	public:
		State() { sx= sy= 0; }
		int sx, sy;
	};
	
	typedef Glib::RefPtr<State> PState;

	int scrollx, scrolly;
	sigc::connection scrollc;

};

class WidgetCanvasEditor : public CanvasEditor, public TCreateEditorWidget<WidgetCanvasEditorWidget> {

public:
	WidgetCanvasEditor();

	void created();
	void load(bool reload);
	void deleted();

	PNode getTopMaster();

	bool startPasting(const Glib::ustring & data, bool normal);
	void stopPasting();

protected:
	enum State { Selecting, Selecting1, Dragging, Pasting, Pasting1, Adjusting };
	enum Command { Move, Press, Press2, Release, Paste, Cancel, Enter, Leave };

	typedef std::map<PNode, Rectangle> Frames;

	struct Location {
		int dx, dy, dw, dh;
		PNode widget;
		Location();
		void clear();
		bool manipulator();
		bool border();
		bool internal();
		bool empty();
		bool test(int x, int y, int w, int h);
		Gdk::CursorType cursor();
	};

	void onActionCompleted();

	void read(Command command);

	void beginSelecting();
	void selectingMove();
	void selectingPress();
	void selectingRelease();
	void selectingMenu();
	void beginPasting();
	void pastingMove();
	void pastingLeave();
	void pastingPress();
	void endPasting();
	void beginDragging();
	void draggingMove();
	void endDragging();
	void adjustingMove();
	void endAdjusting();

	void showSelectionFrames();
	void showSelectionFramesIdle();
	bool createSelectionFrames();
	bool createSingleFrame(const Rectangle &);
	void paintFrames(const std::string & frameColorName=std::string());
	void testLocation(const Point & p);
	PNode findWidget(const Point & p, bool ignoreSelection=false);
	PNode findChild(PNode widget, const Point & p);
	PNode findCompositeChild(PNode widget, const Point & p);
	bool isActiveWidget(PNode node);
	bool isInteractive(PNode node);

	Gtk::Widget * getWidget(PNode node);
	Rectangle getWidgetRect(PNode node);
	Rectangle getManipulatorRect(const Rectangle & R, int X, int Y);
	Point toGlass(PNode widget, const Point & p);
	Point toLocal(PNode widget, const Point & p);
	PGtkWidgetView getWidgetView(PNode widget);
	bool isXYChild(PNode widget);

	Point getDesignSize(PNode node);
	void setDesignSize(const Point & size);
	void setXYChildPlacement(PNode widget, const Rectangle & r);

	bool onMouseEnter(GdkEventCrossing *);
	bool onMouseLeave(GdkEventCrossing *);
	bool onMouseMove(GdkEventMotion *);
	bool onMouseButtonPress(GdkEventButton *);
	bool onMouseButtonRelease(GdkEventButton *);
	void mouseMove(int x, int y);
	void buttonPress(int x, int y, Gdk::ModifierType state, int mb);
	void buttonPress2(int mb);
	void buttonRelease(int x, int y, Gdk::ModifierType state, int mb);

	static void customDoEvent(GdkEvent * event, gpointer data);
	void doEvent(GdkEvent * event);
	static gboolean interactionButtonRelease(
		GtkWidget * widget,
		GdkEventButton * event,
		bool * stopInteracting
	);
	void interact();
	void synthesize(GdkWindow * window, GdkEventType type);

	void updateStatus(const Glib::ustring & text);
	void updateStatus(PNode widget, const Point & size);

private:
	State state;
	bool control;
	int button;
	Point press;
	Point mouse;
	Location location;
	Frames frames;

	int GAP;
	int ML;

	Glib::ustring pasteData;
	bool pasteNormal;
	bool selectedOnPress;

	bool interacting;
	bool stopInteracting;
	
	bool deleting;
	PNode topMaster;

};

}

#endif
