/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "containers.h"

#include "util.h"
#include "widgets.h"

namespace Crow {

// Container -------------------------------------------------------------------

Place::Place() {
	a= b= c= d= 0;
}

bool Place::operator == (const Place & p) const {
	return a==p.a && b==p.b && c==p.c && d==p.d;
}

Placeholder::Placeholder() {
	set_size_request(40, 40);
}

bool Placeholder::on_expose_event(GdkEventExpose * event) {
	bool result= Gtk::DrawingArea::on_expose_event(event);
	DrawGrid(get_window(), get_style());
	return result;
}

Container::Container() {
	container= NULL;
}

bool Container::isPlaceholder(PGtkWidget widget) {
	return Glib::RefPtr<Placeholder>::cast_dynamic(widget);
}

PChild Container::createChild() {
	return NewRefPtr<Child>();
}

void Container::erase(Gtk::Widget * widget) {
	if(widget)
		ContainerRemove(getContainer(), widget);
}

void Container::unparent(PGtkWidget widget, bool force) {
	if(widget && widget->get_parent() && (widget->get_parent()!=getContainer() || force))
		ContainerRemove(widget->get_parent(), GetPtr(widget));
}

PGtkWidget Container::createPlaceholder() {
	PGtkWidget widget= NewRefPtr<Placeholder>();
	widget->show();
	return widget;
}

Gtk::PackOptions Container::getPackOptions(bool expand, bool fill) {
	if(!expand)
		return Gtk::PACK_SHRINK;
	else
		return fill ? Gtk::PACK_EXPAND_WIDGET : Gtk::PACK_EXPAND_PADDING;
}

bool Container::getExpand(Gtk::PackOptions packOptions) {
	return packOptions != Gtk::PACK_SHRINK;
}

bool Container::getFill(Gtk::PackOptions packOptions) {
	return packOptions == Gtk::PACK_EXPAND_WIDGET;
}

// XYContainer -----------------------------------------------------------------

XYChild::XYChild() {
	SetData(this, "CrowTypeHint", "CrowXYChild");
}

void XYContainer::set(const Children & children) {

	if(children.empty()) {
		ContainerClear(getContainer());
		return;
	}

	Children old= get(false);

	for(int i=0; i<(int)old.size(); ++i) {
		PChild child= find(old[i]->widget, children);
		if(!child)
			erase(GetPtr(old[i]->widget));
	}

	for(int i=0; i<(int)children.size(); ++i) {

		PChild child= find(children[i]->widget, old);

		if(child) {

			if(child->place != children[i]->place)
				move(children[i]);

		}	else {

			insert(children[i]);

		}

	}

}

Children XYContainer::get(bool placeholders) {
	Children result;
	std::vector<Gtk::Widget *> widgets= getContainer()->get_children();
	result.resize(widgets.size());
	Point shift= getShift();
	for(int i=0; i<(int)result.size(); ++i) {
		result[i]= Container::createChild();
		result[i]->widget= MakeRefPtr(widgets[i]);
		Point xy= getXY(widgets[i]);
		result[i]->place.a= xy.x - shift.x;
		result[i]->place.b= xy.y - shift.y;
	}
	return result;
}

Point XYContainer::getShift() {
	return Point(0, 0);
}

PChild XYContainer::find(PGtkWidget widget, const Children & children) {
	for(int i=0; i<(int)children.size(); ++i)
		if(children[i]->widget == widget)
			return children[i];
	return PChild();
}

Point XYContainer::getXY(Gtk::Widget * widget) {
	Point xy;
	xy.x= getChildProperty<int>(widget, "x");
	xy.y= getChildProperty<int>(widget, "y");
	return xy + getShift();
}

// FixedContainer --------------------------------------------------------------

Point FixedContainer::getShift() {
	int bw= getContainer()->get_border_width();
	return Point(bw, bw);
}

void FixedContainer::insert(PChild child) {
	unparent(child->widget);
	getContainer()->put(*GetPtr(child->widget), child->place.a, child->place.b);
}

void FixedContainer::move(PChild child) {
	getContainer()->move(*GetPtr(child->widget), child->place.a, child->place.b);
}

// LayoutContainer -------------------------------------------------------------

Point LayoutContainer::getShift() {
	int sx= (int)getContainer()->get_hadjustment()->get_value();
	int sy= (int)getContainer()->get_vadjustment()->get_value();
	return Point(-sx, -sy);
}

void LayoutContainer::insert(PChild child) {
	unparent(child->widget);
	getContainer()->put(*GetPtr(child->widget), child->place.a, child->place.b);
}

void LayoutContainer::move(PChild child) {
	getContainer()->move(*GetPtr(child->widget), child->place.a, child->place.b);
}

// RowContainer ----------------------------------------------------------------

RowChild::RowChild() {
	SetData(this, "CrowTypeHint", "CrowRowChild");
}

void RowContainer::update(int item, PChild child, bool placeholders) {
	Gtk::Widget * widget= getWidget(item);

	if(!child) {

		if(!placeholders) {

			setWidget(item, NULL);

		} else {

			if(widget) {
				if(!isPlaceholder(MakeRefPtr(widget))) {
					setWidget(item, GetPtr(createPlaceholder()));
					updateChild(item, child);
				}
			} else {
				setWidget(item, GetPtr(createPlaceholder()));
				updateChild(item, child);
			}

		}

	} else {

		if(widget != GetPtr(child->widget)) {
			unparent(child->widget, true);
			setWidget(item, GetPtr(child->widget));
			updateChild(item, child);
		} else
			updateChild(item, child);

	}

}

// BinContaner -----------------------------------------------------------------

void BinContainer::set(const Children & children, bool placeholders) {
	Check(children.size()<=1);
	if(children.empty()) {
		update(0, PChild(), placeholders);
	} else
		update(0, children.front(), placeholders);
}

Children BinContainer::get(bool placeholders) {
	Children result;
	Gtk::Widget * child= getWidget(0);
	if(child && (placeholders || !isPlaceholder(MakeRefPtr(child)))) {
		result.push_back(Container::createChild());
		result.front()->widget= MakeRefPtr(child);
	}
	return result;
}

Gtk::Widget * BinContainer::getWidget(int item) {
	Check(item==0);
	return getContainer()->get_child();
}

void BinContainer::setWidget(int item, Gtk::Widget * widget) {
	Check(item==0);
	BinRemove(getContainer());
	if(widget)
		getContainer()->add(*widget);
}

// ScrolledWindowContainer -----------------------------------------------------

Gtk::Widget * ScrolledWindowContainer::getWidget(int item) {
	Check(item==0);
	Gtk::Widget * widget= getContainer()->get_child();
	if(widget && GetData(widget, "CrowAutoViewportHint")!=NULL) {
		Gtk::Viewport * viewport= dynamic_cast<Gtk::Viewport *>(widget);
		return viewport->get_child();
	} else
		return widget;
}

void ScrolledWindowContainer::setWidget(int item, Gtk::Widget * widget) {
	Check(item==0);
	Gtk::Widget * child= getContainer()->get_child();
	if(child) {
		child->reference();
		BinRemove(getContainer());
		if(GetData(child, "CrowAutoViewportHint")!=NULL)
			BinRemove(dynamic_cast<Gtk::Bin *>(child));
		g_signal_handlers_block_matched(getContainer()->get_vadjustment()->gobj(), G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, child->gobj());
		g_signal_handlers_block_matched(getContainer()->get_hadjustment()->gobj(), G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, child->gobj());
		child->unreference();	
	}
	if(widget) {
		getContainer()->add(*widget);
		if(getContainer()->get_child() != widget) {
			Gtk::Viewport * viewport= dynamic_cast<Gtk::Viewport *>(getContainer()->get_child());
			SetData(viewport, "CrowAutoViewportHint", "true");
		}
	}
}

// DesignDialogContainer -------------------------------------------------------

DesignDialog * DesignDialogContainer::getContainer() { 
	return static_cast<DesignDialog *>(Container::getContainer()); 
}

Gtk::Widget * DesignDialogContainer::getWidget(int item) {
	Check(item==0);
	return getContainer()->getBin()->get_child();
}

void DesignDialogContainer::setWidget(int item, Gtk::Widget * widget) {
	Check(item==0);
	BinRemove(getContainer()->getBin());
	if(widget)
		getContainer()->getBin()->add(*widget);
}

// PanedContainer --------------------------------------------------------------

PanedChild::PanedChild() {
	SetData(this, "CrowTypeHint", "CrowPanedChild");
	resize= true;
	shrink= true;
}

void PanedContainer::set(const Children & children, bool placeholders) {
	Check(children.size()<=2);
	if(children.empty()) {
		update(0, PChild(), placeholders);
		update(1, PChild(), placeholders);
	} else if(children.size()==2) {
		update(children[0]->place.a, children[0], placeholders);
		update(children[1]->place.a, children[1], placeholders);
	} else {
		PChild child= children.front();
		if(child->place.a==0) {
			update(0, child, placeholders);
			update(1, PChild(), placeholders);
		} else {
			update(0, PChild(), placeholders);
			update(1, child, placeholders);
		}
	}
}

Children PanedContainer::get(bool placeholders) {
	Children result;
	for(int i=0; i<2; ++i) {
		Gtk::Widget * child= getWidget(i);
		if(child && (placeholders || !isPlaceholder(MakeRefPtr(child)))) {
			PChild R= Container::createChild();
			R->widget= MakeRefPtr(child);
			R->place.a= i;
			result.push_back(R);
		}
	}
	return result;
}

PChild PanedContainer::createChild() {
	return NewRefPtr<PanedChild>();
}

Gtk::Widget * PanedContainer::getWidget(int item) {
	Check(item<=1);
	return item==0 ? getContainer()->get_child1() : getContainer()->get_child2();
}

void PanedContainer::setWidget(int item, Gtk::Widget * widget) {
	Check(item<=1);
	if(item==0) {
		erase(getContainer()->get_child1());
		if(widget)
			getContainer()->add1(*widget);
	} else {
		erase(getContainer()->get_child2());
		if(widget)
			getContainer()->add2(*widget);
	}
}

void PanedContainer::updateChild(int item, PChild child) {
	Gtk::Widget * widget= getWidget(item);
	PPanedChild pchild= PPanedChild::cast_static(child);
	bool resize= pchild ? pchild->resize : true;
	bool shrink= pchild ? pchild->shrink : true;
	if(getChildProperty<bool>(widget, "resize") != resize)
		setChildProperty<bool>(widget, "resize", resize);
	if(getChildProperty<bool>(widget, "shrink") != shrink)
		setChildProperty<bool>(widget, "shrink", shrink);
}

// IndexContainer --------------------------------------------------------------

void IndexContainer::process(int capacity) {

	for(int i=0; i<capacity; ++i) {

		PGtkWidget widget= input(i)->widget;

		int index= find(widget, i);

		if(index == -1) {

			insert(i);

		} else {

			if(index != i)
				reorder(index, i);

		}

		update(i);

	}

	shrink(capacity);

}

// BoxContainer ----------------------------------------------------------------

BoxChild::BoxChild() {
	SetData(this, "CrowTypeHint", "CrowBoxChild");
	place.b= Gtk::PACK_START;
	options= Gtk::PACK_EXPAND_WIDGET;
	padding= 0;
}

void BoxContainer::set(const Children & children, int capS, int capE) {

	sorted.resize(capS + capE);

	for(int i=0; i<(int)children.size(); ++i) {

		PChild child= children[i];

		if(child->place.b == Gtk::PACK_START) {
			Check(child->place.a < capS);
			Check(!sorted[child->place.a]);
			sorted[child->place.a]= child;
		} else {
			Check(child->place.b == Gtk::PACK_END);
			Check(child->place.a < capE);
			Check(!sorted[capS + child->place.a]);
			sorted[capS + child->place.a]= child;
		}

	}

	for(int i=0; i<capS+capE; ++i)
		if(!sorted[i]) {
			sorted[i]= createChild();
			if(i<capS) {
				sorted[i]->place.a= i;
				sorted[i]->place.b= Gtk::PACK_START;
			} else {
				sorted[i]->place.a= i-capS;
				sorted[i]->place.b= Gtk::PACK_END;
			}
		}

	process(capS + capE);

	sorted.clear();

}

Children BoxContainer::get(bool placeholders) {

	Children result;
	Gtk::Box_Helpers::BoxList::iterator child= getContainer()->children().begin();
	int i=0;
	result.reserve(getContainer()->children().size());
	int capS= 0;
	int capE= 0;

	for(; child!=getContainer()->children().end(); ++child, ++i) {

		if(!child->get_pack()) {
			Check(capE==0);
			++capS;
		} else
			++capE;

		if(!placeholders && isPlaceholder(MakeRefPtr(child->get_widget())))
			continue;

		PChild R= Container::createChild();
		result.push_back(R);

		R->widget= MakeRefPtr(child->get_widget());
		R->place.a= child->get_pack() ? i - capS : i;
		R->place.b= child->get_pack();

	}

	Check(capS + capE == (int)getContainer()->children().size());

	return result;
}

PChild BoxContainer::createChild() {
	return NewRefPtr<BoxChild>();
}

int BoxContainer::find(PGtkWidget widget, int start) {

	Gtk::Box_Helpers::BoxList::iterator it= getContainer()->children().begin();
	int i= 0;

	for(; it!=getContainer()->children().end(); ++it, ++i)
		if(i >= start) {

			if(widget) {

				if(it->get_widget() == GetPtr(widget))
					return i;

			} else {

				if(isPlaceholder(MakeRefPtr(it->get_widget())))
					return i;

			}

		}

	return -1;
}

void BoxContainer::reorder(int index1, int index2) {
	Gtk::Widget * widget= getContainer()->children()[index1].get_widget();
	getContainer()->reorder_child(*widget, index2);
}

void BoxContainer::shrink(int size) {
	while((int)getContainer()->children().size() > size)
		getContainer()->children().pop_back();
}

PChild BoxContainer::input(int i) {
	Check(0 <= i && i < (int)sorted.size());
	return sorted[i];
}

void BoxContainer::insert(int index) {
	Gtk::Box_Helpers::BoxList::iterator it= getContainer()->children().begin();
	for(int i=0; i<index; ++i)
		++it;
	PBoxChild S= PBoxChild::cast_static(sorted[index]);
	PGtkWidget widget= S->widget ? S->widget : createPlaceholder();
	unparent(widget);
	getContainer()->children().insert(it, Gtk::Box_Helpers::Element(
		*GetPtr(widget),
		S->options,
		S->padding,
		(Gtk::PackType)S->place.b
	));
}

void BoxContainer::update(int index) {
	Gtk::Box_Helpers::BoxList::iterator child= getContainer()->children().begin();
	for(int i=0; i<index; ++i)
		++child;
	PBoxChild S= PBoxChild::cast_static(sorted[index]);
	if(S->widget)
		Check(GetPtr(S->widget) == child->get_widget());
	else
		Check(isPlaceholder(MakeRefPtr(child->get_widget())));
	if((Gtk::PackType)child->get_pack() != (Gtk::PackType)S->place.b)
		setChildProperty(child->get_widget(), "pack-type", S->place.b);
	if(getPackOptions(child->get_expand(), child->get_fill()) != S->options) {
		setChildProperty(child->get_widget(), "expand", getExpand(S->options));
		setChildProperty(child->get_widget(), "fill", getFill(S->options));
	}
	if(child->get_padding() != S->padding)
		setChildProperty(child->get_widget(), "padding", S->padding);
	postUpdate(child, index);
}

// ButtonBoxContainer ----------------------------------------------------------

ButtonBoxChild::ButtonBoxChild() {
	SetData(this, "CrowTypeHint", "CrowButtonBoxChild");
	secondary= false;
}

PChild ButtonBoxContainer::createChild() {
	return NewRefPtr<ButtonBoxChild>();
}

void ButtonBoxContainer::postUpdate(Gtk::Box_Helpers::BoxList::iterator child, int index) {
	PButtonBoxChild S= PButtonBoxChild::cast_static(sorted[index]);
	if(getContainer()->get_child_secondary(*child->get_widget()) != S->secondary)
		getContainer()->set_child_secondary(*child->get_widget(), S->secondary);
}

// NotebookContainer -----------------------------------------------------------

NotebookChild::NotebookChild() {
	SetData(this, "CrowTypeHint", "CrowNotebookChild");
	packOptions= Gtk::PACK_SHRINK;
}

void NotebookContainer::set(const Children & children, int capacity) {
	int page= getContainer()->get_current_page();
	sorted.resize(capacity);
	for(int i=0; i<(int)children.size(); ++i) {
		PChild child= children[i];
		Check(child->place.a < capacity);
		Check(!sorted[child->place.a]);
		sorted[child->place.a]= child;
	}
	for(int i=0; i<capacity; ++i)
		if(!sorted[i]) {
			sorted[i]= createChild();
			sorted[i]->place.a= i;
		}
	process(capacity);
	sorted.clear();
	if(0 <= page && page < getContainer()->get_n_pages())
		getContainer()->set_current_page(page);
}

Children NotebookContainer::get(bool placeholders) {

	Children result;
	Gtk::Notebook_Helpers::PageList::iterator child= getContainer()->pages().begin();
	int i=0;
	result.reserve(getContainer()->pages().size());

	for(; child!=getContainer()->pages().end(); ++child, ++i) {
		if(!placeholders && isPlaceholder(MakeRefPtr(child->get_child())))
			continue;
		PChild R= Container::createChild();
		result.push_back(R);
		R->widget= MakeRefPtr(child->get_child());
		R->place.a= i;
	}

	return result;
}

PChild NotebookContainer::createChild() {
	return NewRefPtr<NotebookChild>();
}

int NotebookContainer::find(PGtkWidget widget, int start) {

	Gtk::Notebook_Helpers::PageList::iterator it= getContainer()->pages().begin();
	int i= 0;

	for(; it!=getContainer()->pages().end(); ++it, ++i)
		if(i >= start) {

			if(widget) {

				if(it->get_child() == GetPtr(widget))
					return i;

			} else {

				if(isPlaceholder(MakeRefPtr(it->get_child())))
					return i;

			}

		}

	return -1;
}

void NotebookContainer::reorder(int index1, int index2) {
	Gtk::Widget * widget= getContainer()->pages()[index1].get_child();
	getContainer()->reorder_child(*widget, index2);
}

void NotebookContainer::shrink(int size) {
	while((int)getContainer()->pages().size() > size)
		getContainer()->pages().pop_back();
}

PChild NotebookContainer::input(int i) {
	Check(0 <= i && i < (int)sorted.size());
	return sorted[i];
}

void NotebookContainer::insert(int index) {
	PNotebookChild S= PNotebookChild::cast_static(sorted[index]);
	PGtkWidget widget= S->widget ? S->widget : createPlaceholder();
	unparent(widget);
	getContainer()->insert_page(*GetPtr(widget), index);
}

void NotebookContainer::update(int index) {
	PNotebookChild S= PNotebookChild::cast_static(sorted[index]);
	Gtk::Notebook_Helpers::PageList::iterator child= getContainer()->pages().find(index);

	if(S->widget)
		Check(GetPtr(S->widget) == child->get_child());
	else
		isPlaceholder(MakeRefPtr(child->get_child()));

	if(S->tabLabelWidget) {
		if(child->get_tab_label() != GetPtr(S->tabLabelWidget))
			getContainer()->set_tab_label(*child->get_child(), *GetPtr(S->tabLabelWidget));
	} else {
		if(S->tabLabel.empty())
			getContainer()->set_tab_label_text(*child->get_child(), "Page " + ToString(index));
		else
			getContainer()->set_tab_label_text(*child->get_child(), S->tabLabel);
	}

	if(S->menuLabelWidget) {
		if(child->get_menu_label() != GetPtr(S->menuLabelWidget))
			getContainer()->set_menu_label(*child->get_child(), *GetPtr(S->menuLabelWidget));
	} else {
		if(S->menuLabel.empty())
			gtk_container_child_set(
				Container::getContainer()->gobj(), child->get_child()->gobj(),
				"menu-label", NULL, NULL
			);
		else
			getContainer()->set_menu_label_text(*child->get_child(), S->menuLabel);
	}

	bool expand= getChildProperty<bool>(child->get_child(), "tab-expand");
	bool fill= getChildProperty<bool>(child->get_child(), "tab-fill");
	if(getPackOptions(expand, fill) != S->packOptions) {
		setChildProperty(child->get_child(), "tab-expand", getExpand(S->packOptions));
		setChildProperty(child->get_child(), "tab-fill", getFill(S->packOptions));
	}

}

// TableContainer --------------------------------------------------------------

TableChild::TableChild() {
	SetData(this, "CrowTypeHint", "CrowTableChild");
	place.c= 1;
	place.d= 1;
	paddingX= 0;
	paddingY= 0;
	xExpand= true;
	xFill= true;
	xShrink= false;
	yExpand= true;
	yFill= true;
	yShrink= false;
}

TableContainer::TableContainer() {
	skip= Container::createChild();
	capX= capY= 0;
}

void TableContainer::set(const Children & children, int capX, int capY, bool placeholders) {
	sorted.resize(capX*capY);
	this->capX= capX;
	this->capY= capY;
	for(int i=0; i<(int)children.size(); ++i) {
		PChild child= children[i];
		put(child->place.a, child->place.b, child);
	}
	for(int i=0; i<(int)children.size(); ++i) {
		PChild child= children[i];
		int x= child->place.a;
		int y= child->place.b;
		int sx=1, sy=1;
		while(sx<child->place.c && mark(x+sx, y, 1))
			++sx;
		while(sy<child->place.d && mark(x, y+sy, sx))
			++sy;
		child->place.c= sx;
		child->place.d= sy;
	}
	for(int x=0; x<capX; ++x)
		for(int y=0; y<capY; ++y)
			if(!input(x, y)) {
				PChild child= createChild();
				child->place.a= x;
				child->place.b= y;
				put(x, y, child);
			}
	process(placeholders);
	sorted.clear();
	this->capX= 0;
	this->capY= 0;
}

Children TableContainer::get(bool placeholders) {
	Children result;
	result.reserve(getContainer()->children().size());
	Gtk::Table_Helpers::TableList::iterator child= getContainer()->children().begin();
	for(; child!=getContainer()->children().end(); ++child)
		if(placeholders || !isPlaceholder(MakeRefPtr(child->get_widget()))) {
			PChild R= Container::createChild();
			R->widget= MakeRefPtr(child->get_widget());
			R->place.a= child->get_left_attach();
			R->place.b= child->get_top_attach();
			R->place.c= child->get_right_attach() - child->get_left_attach();
			R->place.d= child->get_bottom_attach() - child->get_top_attach();
			result.push_back(R);
		}
	std::sort(result.begin(), result.end(), less());
	return result;
}

bool TableContainer::less::operator () (const PChild & ch1, const PChild & ch2) const {
	const Place & p1= ch1->place;
	const Place & p2= ch2->place;
	return p1.b < p2.b || (p1.b==p2.b && p1.a < p2.a);
}

PChild TableContainer::createChild() {
	return NewRefPtr<TableChild>();
}

PTableChild TableContainer::input(int x, int y) {
	Check(x < capX && y < capY);
	return PTableChild::cast_static(sorted[x+capX*y]);
}

void TableContainer::put(int x, int y, PChild child) {
	Check(x < capX && y < capY);
	Check(!input(x, y));
	sorted[x+capX*y]= child;
}

bool TableContainer::mark(int x, int y, int s) {
	if(y>=capY || x+s-1>=capX)
		return false;
	for(int i=0; i<s; ++i)
		if(input(x+i, y))
			return false;
	for(int i=0; i<s; ++i)
		put(x+i, y, skip);
	return true;
}

void TableContainer::process(bool placeholders) {

	getContainer()->resize(capX, capY);

	Children old= get(true);

	for(int i=0; i<(int)sorted.size(); ++i)
		if(sorted[i]!=skip && (placeholders || sorted[i]->widget)) {

			PTableChild tchild= PTableChild::cast_static(sorted[i]);
			PChild child;
			for(int n=0; n<(int)old.size(); ++n)
				if(
					(old[n]->widget && old[n]->widget == sorted[i]->widget) ||
					(!sorted[i]->widget && isPlaceholder(old[n]->widget))
				)
					child= old[n];

			if(child) {
				tchild->widget= child->widget;
				child->widget.clear();
				if(child->place != tchild->place)
					move(tchild);
			}	else {
				insert(tchild);
			}

			update(tchild);
		}

	for(int i=0; i<(int)old.size(); ++i)
		if(old[i]->widget)
			erase(GetPtr(old[i]->widget));

}

void TableContainer::move(PTableChild child) {
	setChildProperty(GetPtr(child->widget), "left-attach", child->place.a);
	setChildProperty(GetPtr(child->widget), "top-attach", child->place.b);
	setChildProperty(GetPtr(child->widget), "right-attach", child->place.a + child->place.c);
	setChildProperty(GetPtr(child->widget), "bottom-attach", child->place.b + child->place.d);
}

void TableContainer::insert(PTableChild child) {
	if(!child->widget)
		child->widget= createPlaceholder();
	unparent(child->widget);
	getContainer()->attach(*GetPtr(child->widget),
		child->place.a, child->place.a + child->place.c,
		child->place.b, child->place.b + child->place.d,
		getAttachOptions(child->xExpand, child->xFill, child->xShrink),
		getAttachOptions(child->yExpand, child->yFill, child->yShrink),
		child->paddingX, child->paddingY
	);
}

void TableContainer::update(PTableChild child) {
	if(getChildProperty<unsigned>(GetPtr(child->widget), "x-padding") != child->paddingX)
		setChildProperty(GetPtr(child->widget), "x-padding", child->paddingX);
	if(getChildProperty<unsigned>(GetPtr(child->widget), "y-padding") != child->paddingY)
		setChildProperty(GetPtr(child->widget), "y-padding", child->paddingY);
	Gtk::AttachOptions xOptions= getAttachOptions(child->xExpand, child->xFill, child->xShrink);
	if(getChildProperty<Gtk::AttachOptions>(GetPtr(child->widget), "x-options") != xOptions)
		setChildProperty(GetPtr(child->widget), "x-options", xOptions);
	Gtk::AttachOptions yOptions= getAttachOptions(child->yExpand, child->yFill, child->yShrink);
	if(getChildProperty<Gtk::AttachOptions>(GetPtr(child->widget), "y-options") != yOptions)
		setChildProperty(GetPtr(child->widget), "y-options", yOptions);
}

Gtk::AttachOptions TableContainer::getAttachOptions(bool expand, bool fill, bool shrink) {
	Gtk::AttachOptions result= Gtk::AttachOptions();
	if(expand)
		result |= Gtk::EXPAND;
	if(fill)
		result |= Gtk::FILL;
	if(shrink)
		result |= Gtk::SHRINK;
	return result;
}

}
