/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CROW_CONTROLLER_H
#define CROW_CONTROLLER_H

#include "crow_impl.h"
#include "model.h"

namespace Crow {

class Context {

public:
	virtual ~Context() { }
	
	virtual void sweep() = 0;
	virtual Gtk::Window * getWindow() = 0;
	virtual Settings * getSettings() = 0;
	virtual int runDialog(Gtk::Window * window) = 0;
	virtual void showMessage(const Glib::ustring & str) = 0;

	template <class T> T getKeyValue(const std::string & key, const T & def=T()) {
		Settings::const_iterator it= getSettings()->find(key);
		return it==getSettings()->end() ? def : FromString<T>(it->second);
	}

	template <class T> void setKeyValue(const std::string & key, const T & val) {
		(*getSettings())[key]= ToString<T>(val);
	}

};

enum LoadResult { lrSuccess, lrFailed, lrChanged };

class Controller {

public:
	typedef std::map<PNode, PEntityView> ViewMap;

	Controller();
	~Controller();

	Context * getContext() { return context; }
	void setContext(Context * context) { this->context= context; }

	Model * getModel() { return &model; }

	void clearState();
	void begin(UpdateMode um=umNormal);
	void rollback();
	void commit();

	void undo();
	void redo();
	bool canUndo();
	bool canRedo();

	LoadResult load(const Glib::ustring & data);
	void clear();
	Glib::ustring save(const Nodes & nodes);

	Property * findViewProperty(PNode node);
	NodeTest willSave(PNode node);
	bool canDefault(PNode node);

	PEntityView findView(Any value);
	const ViewMap * getViews() { return &views; }
	PEntityView getView(PNode node);
	template <class T> Glib::RefPtr<T> getView(PNode node) { return Glib::RefPtr<T>::cast_dynamic(getView(node)); }

	void touch(PNode node);
	bool isTouched(PNode node);

	void objectDestroyed(PGlibObject object);

	struct EmitterInfo {
		GlibObjectView * view;
		std::string signature;
		Emitter emitter;
	};
	typedef std::multimap<std::string, EmitterInfo> Emitters;
	Emitters getEmitters();

	bool modelGetScalar(PNode node, Any * result, PNode owner, const std::string & name, StateFlags sf);
	bool modelGetVector(PNode node, Any * result, PNode owner, const std::string & name, TypeId base, StateFlags sf);
	bool modelGetEntity(PNode node, Any * result, PNode owner, const std::string & name, TypeId base, Property * property);

	PNode modelSetScalar(PNode owner, const std::string & name, Any value, Property * property);
	PNode modelSetVector(PNode owner, TypeId type, const std::string & name, Any value);
	PNode modelSetEntity(PNode owner, const std::string & name, Any value, Property * property);

protected:
	PNode makeNodeForValue(PNode old, PNode owner, const std::string & name, Any value, Property * property);
	PNode makeLink(PNode old, PNode owner, const std::string & name, Property * property);
	void ensureAndWrite(PNode node, Any value, Property * property);
	PEntityView ensureView(PNode node, Any value=Any());

	void synchronize();
	Nodes getOrderedModified();
	Nodes update();
	void write(const Nodes & updated);
	void setView(PNode node, PEntityView view);
	void removeView(PNode node);
	static void * destroyNotify(void * data);
	void sweepViews();

	bool isVoidEntity(Any value);
	bool isEntity(TypeId type);
	bool isWidget(PNode node);

private:
	Model model;
	Context * context;
	NodeSet touched;

	ViewMap views;

	std::list<PGlibObject> destroyed;

};

}

#endif
