/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "designer.h"

#include "crow_impl.h"
#include "editors.h"
#include "views.h"
#include "palette.h"
#include "canvaswid.h"

namespace Crow {

// DesignerImpl ----------------------------------------------------------------

DesignerImpl::DesignerImpl() :
	states(50)
{
	initialized= false;

	createActionGroup();
	createUIManager();

	hPaned.pack1(vPanedHP, false, true);
	vPanedHP.show();

	hPaned.pack2(hPanedCE, true, true);
	hPanedCE.show();

	paletteWidget.signal_palette_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPaletteEvent));
	vPanedHP.pack2(paletteWidget, true, true);
	paletteWidget.show();

	hPanedCE.pack2(explorerVBox, false, true);
	explorerVBox.show();

	explorerVBox.add(explorerWidget);
	explorerWidget.show();

	vPanedHP.pack1(hierarchyWidget, true, true);
	hierarchyWidget.show();

	hPanedCE.pack1(canvasWidget, true, true);
	canvasWidget.show();

	explorerWidget.getTreeView(ExplorerWidget::epProperties)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));

	Gtk::Toolbar * explorerToolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/ExplorerToolbar"));
	explorerToolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	explorerVBox.pack_end(*explorerToolbar, Gtk::PACK_SHRINK);
	explorerToolbar->show();

	hierarchySessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onHierarchySessionEvent));
	canvasSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onCanvasSessionEvent));
	propertiesSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPropertiesSessionEvent));
	propertiesSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));
	signalsSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onSignalsSessionEvent));
	signalsSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));
	packingSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPackingSessionEvent));
	packingSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));

	manager.getController()->setContext(this);
	manager.addSupplier(&hierarchySessionSupplier);
	manager.addSupplier(&canvasSessionSupplier);
	manager.addSupplier(&propertiesSessionSupplier);
	manager.addSupplier(&signalsSessionSupplier);
	manager.addSupplier(&packingSessionSupplier);
	manager.acceptSuppliers();

	manager.signal_action_started().connect(sigc::mem_fun(*this, &DesignerImpl::onManagerActionStarted));
	manager.signal_action_completed().connect(sigc::mem_fun(*this, &DesignerImpl::onManagerActionCompleted));
	manager.signal_before_sessions_update().connect(sigc::mem_fun(*this, &DesignerImpl::onBeforeSessionsUpdate));
	manager.signal_after_sessions_update().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterSessionsUpdate));

	hierarchySessionSupplier.signal_popup_menu().connect(sigc::mem_fun(*this, &DesignerImpl::onPopupMenu));

	canvasSessionSupplier.signal_pasting_finished().connect(sigc::mem_fun(*this, &DesignerImpl::onPastingFinished));
	canvasSessionSupplier.signal_status_changed().connect(sigc::mem_fun(*this, &DesignerImpl::onStatusChanged));
	canvasSessionSupplier.signal_popup_menu().connect(sigc::mem_fun(*this, &DesignerImpl::onPopupMenu));

	explorerWidget.signal_selection_changed().connect(sigc::mem_fun(*this, &DesignerImpl::onExplorerSelectionChanged));

	clipboardTarget= "application/x-gui";
	savePoint= -1;

	Palette::get()->signal_leaks().connect(sigc::mem_fun(*this, &DesignerImpl::onLeaks));
	onLeaks();

	actions= 0;
	signal_actions_updated().connect(sigc::mem_fun(*this, &DesignerImpl::onActionsUpdated));

}

void DesignerImpl::apply_settings(const Settings & set) {

	if(!initialized) {
		hPaned.show();
		initialized= true;
	}

	for(Settings::const_iterator it= set.begin(); it!=set.end(); ++it)
		settings[it->first]= it->second;
	
	if(set.find("hpaned")!=set.end())
		hPaned.set_position(getKeyValue<int>("hpaned"));
	if(set.find("hpaned_ce")!=set.end())
		hPanedCE.set_position(getKeyValue<int>("hpaned_ce"));
	if(set.find("vpaned_hp")!=set.end())
		vPanedHP.set_position(getKeyValue<int>("vpaned_hp"));
	if(set.find("hier_split")!=set.end())
		hierarchyWidget.setSplit(getKeyValue<int>("hier_split"));
	if(set.find("expl_split")!=set.end())
		explorerWidget.setSplit(getKeyValue<int>("expl_split"));
	
	if(set.find("$wce_vc")==set.end())
		settings["$wce_vc"]= ToString(Gdk::Color("forest green"));
	if(set.find("$wce_wc")==set.end())
		settings["$wce_wc"]= ToString(Gdk::Color("indian red"));
	
	paletteWidget.setShowDW(getKeyValue("$show_dw", false));
	manager.applySettings();

}

void DesignerImpl::store_settings(Settings * set) {
	Check(initialized);
	settings["hpaned"]= ToString(hPaned.get_position());
	settings["hpaned_ce"]= ToString(hPanedCE.get_position());
	settings["vpaned_hp"]= ToString(vPanedHP.get_position());
	settings["hier_split"]= ToString(hierarchyWidget.getSplit());
	settings["expl_split"]= ToString(explorerWidget.getSplit());
	manager.storeSettings();
	for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it)
		(*set)[it->first]= it->second;
}

void DesignerImpl::clear() {
	Check(initialized);
	savePoint= -1;
	manager.clear();
	updateSaveActions();
}

void DesignerImpl::load(const Glib::ustring & str) {
	Check(initialized);

	Glib::ustring data;

	if(!str.empty() && str[0]=='"') {
		std::vector<Glib::ustring> lines= Split(str, '\n');
		for(int i=0; i<(int)lines.size(); ++i) {
			int sz= lines[i].size();
			if(sz >= 2)
				data += Glib::strcompress(lines[i].substr(1, sz-2));
			else
				break;
		}
	} else
		data= str;

	savePoint= -1;
	LoadResult result= manager.load(data);
	savePoint= result==lrChanged ? -1 : 0;
	updateSaveActions();
	hierarchyWidget.getEditorWidget()->collapse();
	if(result == lrChanged)
		showChangedWarning();
}

Glib::ustring DesignerImpl::save() {
	Check(initialized);
	Glib::ustring data= manager.save();
	return data;
}

Glib::ustring DesignerImpl::save_header() {
	Check(initialized);

	Glib::ustring data= manager.save();

	std::vector<Glib::ustring> lines= Split(data, '\n');
	data.clear();
	for(int i=0; i<(int)lines.size(); ++i)
		if(!lines[i].empty())
			data += '"' + Glib::strescape(lines[i]+'\n') + "\"\n";
	
	Nodes strings= manager.getModel()->query(FindType("string"));
	for(Nodes::iterator it= strings.begin(); it!=strings.end(); ++it) {
		PNode node= *it;
		const Glib::ustring & value= *node->getValue()->getString();
		if(!value.empty() && manager.getController()->willSave(node).first) {
			Translatable meta(*node->getMeta());
			if(meta.getTranslate()) {
				if(!meta.getComments()->empty())
					data += "\n/*TRANSLATORS: " + *meta.getComments() + "*/";
				Glib::ustring prefix= meta.getPrefix()->empty() ? Glib::ustring() : *meta.getPrefix() + "|";					
				data += "\nGUIXML_(\"" + Glib::strescape(prefix + value)  + "\")\n";
			}
		}
	}
		
	return data;
}

void DesignerImpl::saved() {
	Check(initialized);
	savePoint= manager.getModel()->getPointer();
	updateSaveActions();
}

Gtk::Widget * DesignerImpl::get_widget() {
	return &hPaned;
}

CrowActions DesignerImpl::get_enabled_actions() {
	Check(initialized);
	return actions;
}

void DesignerImpl::run_action(CrowActions action) {
	Glib::signal_idle().connect(sigc::bind<CrowActions>(sigc::bind_return(sigc::mem_fun(*this, &DesignerImpl::doRunAction), false), action));
}

void DesignerImpl::doRunAction(CrowActions action) {
	Check(initialized);
	Check(action & (CROW_ACTIONS_BASIC|CROW_ACTIONS_EXPLORER));
	switch(action) {
		case CROW_ACTION_UNDO:            onActionUndo();            break;
		case CROW_ACTION_REDO:            onActionRedo();            break;
		case CROW_ACTION_CUT:             onActionCut();             break;
		case CROW_ACTION_COPY:            onActionCopy();            break;
		case CROW_ACTION_PASTE:           onActionPaste();           break;
		case CROW_ACTION_PASTE_TO_ROOT:   onActionPasteToRoot();     break;
		case CROW_ACTION_DELETE:          onActionDelete();          break;
		case CROW_ACTION_ABOUT:           onActionAbout();           break;
		case CROW_ACTION_PREVIEW:         onActionPreview();         break;
		case CROW_ACTION_DEFAULT_SIZE:    onActionDefaultSize();     break;
		case CROW_ACTION_SHIFT_LEFT:      onActionShiftLeft();       break;
		case CROW_ACTION_SHIFT_RIGHT:     onActionShiftRight();      break;
		case CROW_ACTION_TRANSLATABLES:   onActionTranslatables();   break;
		case CROW_ACTION_DEFAULT:         onActionDefault();         break;
		case CROW_ACTION_ADD:             onActionAdd();             break;
		case CROW_ACTION_REMOVE:          onActionRemove();          break;
		case CROW_ACTION_MOVE_UP:         onActionMoveUp();          break;
		case CROW_ACTION_MOVE_DOWN:       onActionMoveDown();        break;
		default: Check(false);
	}
}

void DesignerImpl::sweep() {
	PWidgetCanvasEditor editor= PWidgetCanvasEditor::cast_dynamic(canvasSessionSupplier.getEditor());
	if(editor) {
		PNode node= editor->getSession()->getNode1();
		if(!node->isInactive() && manager.getModel()->raiseMaster(node) != editor->getTopMaster())
			editor->getEditorWidget()->unsetWidget();
	}
}

Gtk::Window * DesignerImpl::getWindow() {
	return dynamic_cast<Gtk::Window *>(get_widget()->get_toplevel());
}

int DesignerImpl::runDialog(Gtk::Window * window) {
	int result= Gtk::RESPONSE_OK;
	enterModal();
	if(GTK_IS_PLUG(get_widget()->get_parent()->gobj()))
		window->set_keep_above(true);
	if(Gtk::Dialog * dialog= dynamic_cast<Gtk::Dialog *>(window))
		result= dialog->run();
	else
		Gtk::Main::run(*window);
	leaveModal();
	return result;
}

void DesignerImpl::showMessage(const Glib::ustring & str) {
	Gtk::MessageDialog dialog(*getWindow(), str);
	runDialog(&dialog);
}

void DesignerImpl::showChangedWarning() {
	showMessage("Warning: one or more properties were changed");
}

bool DesignerImpl::canSave() {
	return savePoint != manager.getModel()->getPointer();
}

bool DesignerImpl::shouldSave() {
	return canSave() && !manager.getModel()->emptyHistory();
}

void DesignerImpl::onActionUndo() {
	checkAction(CROW_ACTION_UNDO);
	manager.undo();
}

void DesignerImpl::onActionRedo() {
	checkAction(CROW_ACTION_REDO);
	manager.redo();
}

void DesignerImpl::onActionCut() {
	checkAction(CROW_ACTION_CUT);
	onActionCopy();
	onActionDelete();
}

void DesignerImpl::onActionCopy() {
	checkAction(CROW_ACTION_COPY);
	clipboardStore= manager.copy();
	std::vector<Gtk::TargetEntry> targets;
	targets.reserve(2);
	targets.push_back(Gtk::TargetEntry(clipboardTarget));
	targets.push_back(Gtk::TargetEntry("UTF8_STRING"));
	Gtk::Clipboard::get()->set(
		targets,
		sigc::mem_fun(*this, &DesignerImpl::onClipboardGet),
		sigc::mem_fun(*this, &DesignerImpl::onClipboardClear)
	);
	Gtk::Clipboard::get()->set_can_store();
	Gtk::Clipboard::get()->store();
}

void DesignerImpl::onActionPaste() {
	checkAction(CROW_ACTION_PASTE);
	Gtk::Clipboard::get()->request_contents(
		clipboardTarget,
		sigc::bind<bool>(sigc::mem_fun(*this, &DesignerImpl::onClipboardReceived), false)
	);
}

void DesignerImpl::onActionPasteToRoot() {
	checkAction(CROW_ACTION_PASTE_TO_ROOT);
	Gtk::Clipboard::get()->request_contents(
		clipboardTarget,
		sigc::bind<bool>(sigc::mem_fun(*this, &DesignerImpl::onClipboardReceived), true)
	);
}

void DesignerImpl::onActionDelete() {
	checkAction(CROW_ACTION_DELETE);
	manager.deleteSelection();
}

void DesignerImpl::onActionAbout() {
	checkAction(CROW_ACTION_ABOUT);

	struct Tmp { 
		Gtk::Dialog * dialog;
		DesignerImpl * designer;
		static void onShowURL(GtkAboutDialog * about, const gchar * link, gpointer data) {
			Tmp * tmp= static_cast<Tmp *>(data); 
			tmp->dialog->response(Gtk::RESPONSE_CLOSE);
			Glib::signal_idle().connect(sigc::bind<Glib::ustring>(sigc::bind_return(sigc::mem_fun(tmp->designer->signalShowUrl, &SignalShowUrl::emit), false), link));
		}
	};

	Gtk::AboutDialog dialog;
	Tmp tmp;
	tmp.dialog= &dialog;
	tmp.designer= this;

	gtk_about_dialog_set_url_hook(&Tmp::onShowURL, &tmp, NULL);

	int major= crow_version()/1000000;
	int minor= crow_version()/1000 - major*1000;
	int micro= crow_version() - major*1000000 - minor*1000;
	Glib::ustring version= ToString(major) + "." + ToString(minor) + "." + ToString(micro);

	#ifdef CROW_TIMESTAMP
		version += "-" + ToString(CROW_TIMESTAMP);
	#endif

	dialog.set_name("Crow");
	dialog.set_version(version);
	dialog.set_comments("A modern GUI designer for the GTK+ toolkit");
	dialog.set_copyright("Copyright \302\251 2005 Maxim Udushlivy");
	dialog.set_website(CrowProjectHome);
	dialog.set_authors(std::vector<std::string>(1, "Maxim Udushlivy <maxim_udushlivy@users.sourceforge.net>"));
	dialog.set_license(
		"Crow is free software; you can redistribute it and/or modify\n"
		"it under the terms of the GNU General Public License as published by\n"
		"the Free Software Foundation; either version 2 of the License, or\n"
		"(at your option) any later version.\n\n"
		"Crow is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with Crow; if not, write to the Free Software\n"
		"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA"
	);
	dialog.set_transient_for(*getWindow());

	runDialog(&dialog);

}

void DesignerImpl::onActionPreview() {
	checkAction(CROW_ACTION_PREVIEW);
	crow_enter_check("DesignerImpl::doPreview()");
	try {
		doPreview();
	} catch(const RuntimeError & x) {
		showMessage("Error in preview: " + x.what());
	}
	crow_leave();
	Palette::get()->checkLeaks();
}

void DesignerImpl::onActionDefaultSize() {
	checkAction(CROW_ACTION_DEFAULT_SIZE);
	setDefaultSize();
}

void DesignerImpl::onActionShiftLeft() {
	checkAction(CROW_ACTION_SHIFT_LEFT);
	doShift(-1);
}

void DesignerImpl::onActionShiftRight() {
	checkAction(CROW_ACTION_SHIFT_RIGHT);
	doShift(1);
}

void DesignerImpl::onActionTranslatables() {
	checkAction(CROW_ACTION_TRANSLATABLES);

	TranslatablesDialog dialog;
	dialog.set_transient_for(*getWindow());

	TranslatablesDialog::Strings strings;
	Nodes nodes= manager.getModel()->query(FindType("string"));
	strings.reserve(nodes.size());
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		PNode node= *it;
		bool element= node->getOwner()->getRole()==nrVector;
		Property * property= manager.getController()->findViewProperty(node);
		if(property->hasFlag(sfTranslatable) && property->canEdit()) {
			TranslatablesDialog::String string;
			string.node= node;
			string.path= manager.getModel()->getNodePath(node);
			string.string= *node->getValue()->getString();
			if(!element && property->getDefault())
				string.defval= *property->getDefault()->getString();
			string.meta= Translatable(*node->getMeta());
			strings.push_back(string);
		}
	}
	
	std::sort(strings.begin(), strings.end());

	dialog.setStrings(strings);
	
	if(runDialog(&dialog) == Gtk::RESPONSE_OK) {
		manager.begin();
		strings= dialog.getStrings();
		for(int i= 0; i<(int)strings.size(); ++i) {
			manager.getModel()->setScalar(strings[i].node, CAny::createString(strings[i].string));
			manager.getModel()->setMeta(strings[i].node, strings[i].meta.getMeta());
		}
		manager.commit();
	}

}

void DesignerImpl::onActionDefault() {
	checkAction(CROW_ACTION_DEFAULT);
	doDefault();
}

void DesignerImpl::onActionAdd() {
	checkAction(CROW_ACTION_ADD);
	doAdd();
}

void DesignerImpl::onActionRemove() {
	checkAction(CROW_ACTION_REMOVE);
	doRemove();
}

void DesignerImpl::onActionMoveUp() {
	checkAction(CROW_ACTION_MOVE_UP);
	doMoveUp();
}

void DesignerImpl::onActionMoveDown() {
	checkAction(CROW_ACTION_MOVE_DOWN);
	doMoveDown();
}

void DesignerImpl::createActionGroup() {

	PGtkActionGroup actions= Gtk::ActionGroup::create();

	actions->add(Gtk::Action::create("Cut", Gtk::Stock::CUT, "Cu_t"), sigc::mem_fun(*this, &DesignerImpl::onActionCut));
	actions->add(Gtk::Action::create("Copy", Gtk::Stock::COPY, "_Copy"), sigc::mem_fun(*this, &DesignerImpl::onActionCopy));
	actions->add(Gtk::Action::create("Paste", Gtk::Stock::PASTE, "_Paste..."), sigc::mem_fun(*this, &DesignerImpl::onActionPaste));
	actions->add(Gtk::Action::create("Delete", Gtk::Stock::DELETE, "_Delete"), sigc::mem_fun(*this, &DesignerImpl::onActionDelete));
	actions->add(Gtk::Action::create("PasteToRoot", "Paste to _Root"), sigc::mem_fun(*this, &DesignerImpl::onActionPasteToRoot));
	
	actions->add(Gtk::Action::create("DefaultSize", Gtk::Stock::ZOOM_100, "D_efault Size"), sigc::mem_fun(*this, &DesignerImpl::onActionDefaultSize));
	actions->add(Gtk::Action::create("ShiftLeft", Gtk::Stock::GO_BACK, "Shift _Left"), sigc::mem_fun(*this, &DesignerImpl::onActionShiftLeft));
	actions->add(Gtk::Action::create("ShiftRight", Gtk::Stock::GO_FORWARD, "Shift _Right"), sigc::mem_fun(*this, &DesignerImpl::onActionShiftRight));
	
	actions->add(Gtk::Action::create("Default", Gtk::Stock::CLEAR, "_Default", "Set the property to its default value"), sigc::mem_fun(*this, &DesignerImpl::onActionDefault));
	actions->add(Gtk::Action::create("Add", Gtk::Stock::ADD, "_Add", "Add a New Element"), sigc::mem_fun(*this, &DesignerImpl::onActionAdd));
	actions->add(Gtk::Action::create("Remove", Gtk::Stock::REMOVE, "Re_move", "Remove the element"), sigc::mem_fun(*this, &DesignerImpl::onActionRemove));
	actions->add(Gtk::Action::create("MoveUp", Gtk::Stock::GO_UP, "Move _Up", "Move the element up"), sigc::mem_fun(*this, &DesignerImpl::onActionMoveUp));
	actions->add(Gtk::Action::create("MoveDown", Gtk::Stock::GO_DOWN, "Move Do_wn", "Move the element down"), sigc::mem_fun(*this, &DesignerImpl::onActionMoveDown));

	{
		std::vector<Glib::RefPtr<Gtk::Action> > V= actions->get_actions();
		for(int i=0; i<(int)V.size(); ++i)
			V[i]->set_sensitive(false);
	}

	actionGroup= actions;
}

void DesignerImpl::createUIManager() {

	PGtkUIManager manager= Gtk::UIManager::create();
	manager->insert_action_group(actionGroup);

	Glib::ustring uistr=
		"<ui>"
		"  <toolbar name='ExplorerToolbar'>"
		"    <toolitem action='Default'/>"
		"    <separator/>"
		"    <toolitem action='Add'/>"
		"    <toolitem action='Remove'/>"
		"    <toolitem action='MoveUp'/>"
		"    <toolitem action='MoveDown'/>"
		"  </toolbar>"
		"  <popup name='MainPopup'>"
		"    <menuitem action='Cut'/>"
		"    <menuitem action='Copy'/>"
		"    <menuitem action='Paste'/>"
		"    <menuitem action='Delete'/>"
		"    <separator/>"
		"    <menuitem action='PasteToRoot'/>"
		"    <separator/>"
		"    <menuitem action='DefaultSize'/>"
		"    <menuitem action='ShiftLeft'/>"
		"    <menuitem action='ShiftRight'/>"
		"  </popup>"
		"</ui>"
	;

	try {
		manager->add_ui_from_string(uistr);
	} catch(const Glib::Error & x) {
		FatalError("Creating UIManager failed: " + x.what());
	}

	uiManager= manager;
}

void DesignerImpl::updateActions() {

	CrowActions old= getActionsForMask(CROW_ACTIONS_BASIC);

	setupAction(CROW_ACTION_PASTE, true);
	setupAction(CROW_ACTION_PASTE_TO_ROOT, true);
	setupAction(CROW_ACTION_ABOUT, true);
	setupAction(CROW_ACTION_TRANSLATABLES, true);

	setupAction(CROW_ACTION_UNDO, manager.canUndo());
	setupAction(CROW_ACTION_REDO, manager.canRedo());

	bool selempty= manager.getSelection()->empty();
	setupAction(CROW_ACTION_CUT, !selempty);
	setupAction(CROW_ACTION_COPY, !selempty);
	setupAction(CROW_ACTION_DELETE, !selempty);

	setupAction(CROW_ACTION_PREVIEW, canPreview());
	setupAction(CROW_ACTION_DEFAULT_SIZE, canDefaultSize());
	setupAction(CROW_ACTION_SHIFT_LEFT, canShift());
	setupAction(CROW_ACTION_SHIFT_RIGHT, canShift());

	if(CrowActions changed = getChangedActions(CROW_ACTIONS_BASIC, old))
		signalActionsUpdated.emit(changed);

	updateSaveActions();
	updateExplorerActions();
}

void DesignerImpl::updateSaveActions() {
	CrowActions old= getActionsForMask(CROW_ACTIONS_SAVE);
	setupAction(CROW_ACTION_SAVE, canSave());
	setupAction(CROW_ACTION_SHOULD_SAVE, shouldSave());
	if(CrowActions changed = getChangedActions(CROW_ACTIONS_SAVE, old))
		signalActionsUpdated.emit(changed);
}

void DesignerImpl::updateExplorerActions() {
	CrowActions old= getActionsForMask(CROW_ACTIONS_EXPLORER);
	ExplorerTest et= testExplorer();
	setupAction(CROW_ACTION_DEFAULT, et & etCanDefault);
	setupAction(CROW_ACTION_ADD, et & etCanAdd);
	setupAction(CROW_ACTION_REMOVE, et & etCanRemove);
	setupAction(CROW_ACTION_MOVE_UP, et & etCanMoveUp);
	setupAction(CROW_ACTION_MOVE_DOWN, et & etCanMoveDown);
	if(CrowActions changed = getChangedActions(CROW_ACTIONS_EXPLORER, old))
		signalActionsUpdated.emit(changed);
}

void DesignerImpl::onActionsUpdated(CrowActions changed) {
	if(changed & CROW_ACTIONS_BASIC) {
		setupGtkAction(changed, CROW_ACTION_CUT, "Cut");
		setupGtkAction(changed, CROW_ACTION_COPY, "Copy");
		setupGtkAction(changed, CROW_ACTION_PASTE, "Paste");
		setupGtkAction(changed, CROW_ACTION_PASTE_TO_ROOT, "PasteToRoot");
		setupGtkAction(changed, CROW_ACTION_DELETE, "Delete");
		setupGtkAction(changed, CROW_ACTION_DEFAULT_SIZE, "DefaultSize");
		setupGtkAction(changed, CROW_ACTION_SHIFT_LEFT, "ShiftLeft");
		setupGtkAction(changed, CROW_ACTION_SHIFT_RIGHT, "ShiftRight");
	}
	if(changed & CROW_ACTIONS_EXPLORER) {
		setupGtkAction(changed, CROW_ACTION_DEFAULT, "Default");
		setupGtkAction(changed, CROW_ACTION_ADD, "Add");
		setupGtkAction(changed, CROW_ACTION_REMOVE, "Remove");
		setupGtkAction(changed, CROW_ACTION_MOVE_UP, "MoveUp");
		setupGtkAction(changed, CROW_ACTION_MOVE_DOWN, "MoveDown");
	}
}

void DesignerImpl::setupGtkAction(CrowActions changed, CrowActions ca, const char * str) {
	if(changed & ca)
		actionGroup->get_action(str)->set_sensitive(actions & ca);
} 

void DesignerImpl::setupAction(CrowActions action, bool enabled) {
	if(enabled)
		actions |= action;
	else
		actions &= ~action;
}

CrowActions DesignerImpl::getActionsForMask(CrowActions mask) {
	return actions & mask;
}

CrowActions DesignerImpl::getChangedActions(CrowActions mask, CrowActions old) {
	return getActionsForMask(mask) ^ old;
}

void DesignerImpl::checkAction(CrowActions action) {
	Check((actions & action) == action);
}

void DesignerImpl::onClipboardGet(Gtk::SelectionData & data, guint) {
	std::string target= data.get_target();
	if(target == clipboardTarget)
		data.set(clipboardTarget, clipboardStore);
	else if(target == "UTF8_STRING")
		data.set_text(clipboardStore);
	else if(target == "SAVE_TARGETS")
		;
	else
		showMessage("Unexpected clipboard target format");
}

void DesignerImpl::onClipboardClear() {
}

void DesignerImpl::onClipboardReceived(const Gtk::SelectionData & data, bool root) {
	std::string target= data.get_target();
	if(target == clipboardTarget) {
 		Glib::ustring clipboard= data.get_data_as_string();
		if(!clipboard.empty()) {
			if(!root)
				paletteWidget.unpush();
			paste(clipboard, root, false);
		}
	} else
		showMessage("Unexpected clipboard target format");
}

void DesignerImpl::onManagerActionStarted() {
	savedSelection= ToNodeSet(*manager.getSelection());
	if(savedSelection.empty())
		savedState.clear();
	else
		savedState= explorerWidget.getTreeView(ExplorerWidget::epProperties)->saveState();
}

void DesignerImpl::onManagerActionCompleted() {
	Palette::get()->checkLeaks();
	updateActions();
	NodeSet current= ToNodeSet(*manager.getSelection());
	if(current != savedSelection) {
		if(!savedSelection.empty())
			states.set(savedSelection, savedState);
		if(!current.empty()) {
			PObject state= states.get(current);
			if(state)
				explorerWidget.getTreeView(ExplorerWidget::epProperties)->loadState(state);
		}
	}
}

void DesignerImpl::onExplorerSelectionChanged() {
	updateExplorerActions();
}

void DesignerImpl::onBeforeSessionsUpdate() {
	explorerWidget.getTreeView(ExplorerWidget::epProperties)->prepare();
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->prepare();
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->prepare();
}

void DesignerImpl::onAfterSessionsUpdate() {
	explorerWidget.getTreeView(ExplorerWidget::epProperties)->submit();
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->submit();
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->submit();
}

void DesignerImpl::onHierarchySessionEvent(PSession session) {
	if(session->getOperation()==soInsert)
		hierarchyWidget.setEditorWidget(GetPtr(session->getEditor()->getEditorWidget()));
	else if(session->getOperation()==soRemove)
		hierarchyWidget.unsetEditorWidget();
}

void DesignerImpl::onCanvasSessionEvent(PSession session) {
	if(session->getOperation()==soInsert)
		canvasWidget.setEditorWidget(GetPtr(session->getEditor()->getEditorWidget()));
	else if(session->getOperation()==soRemove) {
		onStatusChanged("");
		canvasWidget.unsetEditorWidget();
	}
}

void DesignerImpl::onPropertiesSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epProperties)->update(*session->getPath(), session);
}

void DesignerImpl::onSignalsSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epSignals)->update(*session->getPath(), session);
}

void DesignerImpl::onPackingSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epPacking)->update(*session->getPath(), session);
}

void DesignerImpl::onAfterLoad(PSession session) {
	PPolycell pcvalue= PPolycell::cast_dynamic(session->getEditor()->getEditorWidget());
	PPolycellText pclabel= pcvalue->getElement()->getValue<PolycellText>(0);

	if(session->modelEditable())
		pclabel->changeAttribute(Pango::Attribute::create_attr_style(Pango::STYLE_NORMAL));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_style(Pango::STYLE_ITALIC));

	if(session->hasNoUndo() && session->modelEditable())
		pclabel->changeAttribute(Pango::Attribute::create_attr_strikethrough(true));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_strikethrough(false));

	if(session->hasHidden())
		pclabel->changeAttribute(Pango::Attribute::create_attr_underline(Pango::UNDERLINE_ERROR));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_underline(Pango::UNDERLINE_NONE));

	if(session->hasVectorWithMasterLink() || (session->hasMasterLink() && !session->isElement()))
		pclabel->changeAttribute(Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_weight(Pango::WEIGHT_NORMAL));

	if(session->isElement())
		pclabel->setText(session->getProperty1()->getLabel(session->getNode1()));
	else
		pclabel->setText((session->hasToSave() ? "*" : "") + session->getName());
}

void DesignerImpl::setRowValues(PPolyelem elem) {
	PSession session= PSession::cast_static(elem->getObject());
	PPolycellText pclabel= NewRefPtr<PolycellText>();
	elem->setValue(0, pclabel);
	PPolycell pcvalue= PPolycell::cast_dynamic(session->getEditor()->getEditorWidget());
	elem->setValue(1, pcvalue);
}

void DesignerImpl::onPaletteEvent(TypeId type, PaletteEvent event) {

	if(event==peCancel) {

		canvasSessionSupplier.stopPasting();

	} else {

		Check(event==peRoot || event==peCanvas);

		Glib::ustring data= Palette::get()->createPasteData(type);
		paste(data, event==peRoot, true);

	}

}

void DesignerImpl::paste(const Glib::ustring & data, bool root, bool normal) {

	if(!root) {

		if(canvasSessionSupplier.startPasting(data, normal))
			return;
		else
			paletteWidget.unpush();

	}

	try {
		manager.pasteToRoot(data, normal);
	} catch(const Glib::Exception & x) {
		showMessage("Pasting failed: " + x.what());
	}

}

void DesignerImpl::onPastingFinished() {
	paletteWidget.unpush();
}

void DesignerImpl::onStatusChanged(const Glib::ustring & status) {
	if(statusText != status)
		signalStatusUpdated.emit(statusText = status);
}

void DesignerImpl::onPopupMenu() {
	Gtk::Menu * menu= dynamic_cast<Gtk::Menu *>(uiManager->get_widget("/MainPopup"));
	menu->popup(0, GDK_CURRENT_TIME);
}

bool DesignerImpl::canDefaultSize() {
	PNode node= manager.getSingleSelection();
	if(node) {
		Any size;
		if(!manager.getModel()->findMaster(node))
			size= manager.getModel()->getScalar(node, "design-size");
		else
			size= manager.getModel()->getScalar(node, "size-request");
		return size && *size->getPoint() != Point(-1, -1);
	} else
		return false;
}

void DesignerImpl::setDefaultSize() {
	PNode node= manager.getSingleSelection();
	if(node) {
		PNode size;
		if(!manager.getModel()->findMaster(node))
			size= manager.getModel()->find(node, "design-size");
		else
			size= manager.getModel()->find(node, "size-request");
		if(size) {
			manager.begin();
			manager.getModel()->setScalar(size, CAny::createPoint(Point(-1, -1)));
			manager.commit();
		}
	}
}

DesignerImpl::Selected DesignerImpl::getSelected() {
	PSession session= explorerWidget.getSelectedRow();
	if(session) {
		if(!session->isMultiple())
			return std::make_pair(session, session->getNode1());
		else
			return std::make_pair(session, PNode());
	} else if(signalsPageCurrent()) {
		PNode node= manager.getSingleSelection();
		if(node) {
			PNode signals= manager.getModel()->find(node, "signals");
			return std::make_pair(PSession(), signals);
		}
	}
	return Selected();
}

bool DesignerImpl::signalsPageCurrent() {
	return explorerWidget.getCurrentPage() == ExplorerWidget::epSignals;
}

ExplorerTest DesignerImpl::testExplorer() {
	ExplorerTest result= 0;
	Selected selected= getSelected();
	const PSession & session= selected.first;
	const PNode & node= selected.second;

	if(session && session->canDefault())
		result |= etCanDefault;

	if(node) {

		PNode owner= node->getOwner();
		Property * property= manager.getController()->findViewProperty(node);

		if(property->canEdit()) {

			if(node->getRole()==nrVector || owner->getRole()==nrVector)
				result |= etCanAdd;

			if(owner->getRole()==nrVector) {
				result |= etCanRemove;

				int index= FromString<int>(*node->getName());
				int size= owner->getOut()->size();

				if(index > 0)
					result |= etCanMoveUp;

				if(index < size-1)
					result |= etCanMoveDown;

			}

		}

	}

	return result;
}

void DesignerImpl::doDefault() {
	Selected selected= getSelected();
	const PSession & session= selected.first;
	manager.begin();
	if(session->getRole()==srVector)
		session->resetVector();
	else if(session->getRole()==srScalar)
		session->resetScalar();
	else {
		Check(session->getRole()==srObject);
		session->resetObject();
	}
	manager.commit();
}

void DesignerImpl::doAdd() {
	PNode node= getSelected().second;
	PNode vector;
	int index;
	if(node->getRole()==nrVector) {
		vector= node;
		index= 0;
	} else {
		vector= node->getOwner();
		index= FromString<int>(*node->getName()) + 1;
	}
	Property * property= manager.getController()->findViewProperty(vector);
	manager.begin();
	PNode inserted= property->insertElement(index);
	manager.commit();
	if(inserted)
		explorerWidget.setSelectedRow(manager.findSession(inserted));
}

void DesignerImpl::doRemove() {
	PNode node= getSelected().second;
	int index= FromString<int>(*node->getName());
	PNode vector= node->getOwner();
	manager.begin();
	manager.getModel()->vectorRemove(node);
	manager.commit();
	index= std::min(index, (int)vector->getOut()->size() - 1);
	if(explorerWidget.getCurrentPage()!=ExplorerWidget::epSignals || index>=0) {
		PNode nsel= index<0 ? vector : manager.getModel()->find(vector, ToString(index));
		explorerWidget.setSelectedRow(manager.findSession(nsel));
	}
}

void DesignerImpl::doMoveUp() {
	PNode node= getSelected().second;
	manager.begin();
	manager.getModel()->vectorMove(node, -1);
	manager.commit();
	explorerWidget.setSelectedRow(manager.findSession(node));
}

void DesignerImpl::doMoveDown() {
	PNode node= getSelected().second;
	manager.begin();
	manager.getModel()->vectorMove(node, 1);
	manager.commit();
	explorerWidget.setSelectedRow(manager.findSession(node));
}

bool DesignerImpl::canPreview() {
	PNode node= manager.raiseSelectionMaster();
	return node ? Palette::get()->isWidget(node->getType()) : false;
}

void DesignerImpl::doPreview() {
	Gtk::Window hwindow;

	PNode master= manager.raiseSelectionMaster();
	const std::string & name= *master->getName();

	Glib::ustring data= manager.save();
	Gui::PLoader loader= Gui::Loader::create();
	loader->set_preview_mode(true);
	loader->load_from_string(data);
	Gtk::Widget * widget= NULL;
	loader->bind_object(name, &widget);

	Gtk::Window * window= dynamic_cast<Gtk::Window *>(widget);
	if(!window) {
		window= &hwindow;
		window->add(*widget);
		window->set_title("Preview Window");
	} else
		window->set_title("Preview Window [" + window->get_title() + "]");

	gtk_window_set_accept_focus(window->gobj(), true);
	window->set_decorated(true);
	window->set_deletable(true);
	window->set_resizable(true);
	gtk_window_set_focus_on_map(window->gobj(), true);
	window->set_modal(true);
	window->set_type_hint(Gdk::WINDOW_TYPE_HINT_NORMAL);
	window->set_transient_for(*getWindow());
	window->set_position(Gtk::WIN_POS_CENTER_ON_PARENT);
	int W, H;
	window->get_default_size(W, H);
	if(W==-1 || H==-1)
		window->set_default_size(320, 200);

	runDialog(window);
}

void DesignerImpl::onLeaks() {
	int leaks= Palette::get()->getLeaks();
	Glib::ustring info= leaks==0 ? Glib::ustring() : ("Leaks: " + ToString(leaks));
	if(infoText != info)
		signalInfoUpdated.emit(infoText = info);
}

int DesignerImpl::getShiftCapacity(PNode child) {
	PNode parent= manager.getModel()->findParent(child);
	Any capacity= manager.getModel()->getScalar(parent, "capacity");
	::Crow::Container * container= manager.getContainer(parent);
	if(dynamic_cast<BoxContainer *>(container)) {
		PNode node= manager.getModel()->findChildNode(child);
		int pack= manager.getModel()->getScalar(node, "pack")->getInt();
		return pack==0 ? capacity->getPoint()->x : capacity->getPoint()->y;
	} else
		return capacity->getInt();
}

PNode DesignerImpl::findShiftNeighbour(PNode child, int index) {
	PNode parent= manager.getModel()->findParent(child);
	::Crow::Container * container= manager.getContainer(parent);
	int pack= -1;
	if(dynamic_cast<BoxContainer *>(container)) {
		PNode node= manager.getModel()->findChildNode(child);
		pack= manager.getModel()->getScalar(node, "pack")->getInt();
	}
	PNode children= manager.getModel()->find(parent, "children");
	const Nodes & out= *children->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		int cindex= manager.getModel()->find(*it, "index")->getValue()->getInt();
		if(cindex==index) {
			if(pack==-1)
				return manager.getModel()->getWidgetNode(*it);
			else {
				int cpack= manager.getModel()->find(*it, "pack")->getValue()->getInt();
				if(cpack==pack)
					return manager.getModel()->getWidgetNode(*it);
			}
		}
	}
	return PNode();
}

int DesignerImpl::getChildIndex(PNode child) {
	PNode node= manager.getModel()->findChildNode(child);
	return manager.getModel()->getScalar(node, "index")->getInt();
}

void DesignerImpl::setChildIndex(PNode child, int index) {
	PNode node= manager.getModel()->findChildNode(child);
	manager.getModel()->setScalar(node, "index", CAny::createInt(index));
}

bool DesignerImpl::canShift() {
	PNode node= manager.getSingleSelection();
	if(!node)
		return false;
	if(!manager.isIndexChild(node))
		return false;
	int capacity= getShiftCapacity(node);
	return capacity > 1;
}

void DesignerImpl::doShift(int shift) {
	PNode child= manager.getSingleSelection();
	int index= getChildIndex(child);
	int capacity= getShiftCapacity(child);
	int index2= (index+shift+capacity)%capacity;
	PNode child2= findShiftNeighbour(child, index2);
	manager.begin();
	setChildIndex(child, index2);
	if(child2)
		setChildIndex(child2, index);
	manager.commit();
}

}
