/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CROW_DESIGNER_H
#define CROW_DESIGNER_H

#include "crow.h"
#include "session.h"
#include "widgets.h"

namespace Crow {

typedef unsigned ExplorerTest;
const ExplorerTest etCanAdd       =   1;
const ExplorerTest etCanRemove    =   2;
const ExplorerTest etCanMoveUp    =   4;
const ExplorerTest etCanMoveDown  =   8;
const ExplorerTest etCanDefault   =  16;

class DesignerImpl : public Object, public Designer, public Context {

public:
	DesignerImpl();

	void apply_settings(const Settings & settings);
	void store_settings(Settings * settings);
	void clear();
	void load(const Glib::ustring & data);
	Glib::ustring save();
	Glib::ustring save_header();
	void saved();
	Gtk::Widget * get_widget();
	CrowActions get_enabled_actions();
	void run_action(CrowActions action);

	void doRunAction(CrowActions action);

	SignalActionsUpdated & signal_actions_updated() { return signalActionsUpdated; }
	SignalStatusUpdated & signal_status_updated() { return signalStatusUpdated; }
	SignalInfoUpdated & signal_info_updated() { return signalInfoUpdated; }
	SignalShowUrl & signal_show_url() { return signalShowUrl; }
	SignalModalEnter & signal_modal_enter() { return signalModalEnter; }
	SignalModalLeave & signal_modal_leave() { return signalModalLeave; }

	void sweep();
	Gtk::Window * getWindow();
	Settings * getSettings() { return &settings; }
	int runDialog(Gtk::Window * window);
	void showMessage(const Glib::ustring & str);
	void enterModal() { signalModalEnter.emit(); }
	void leaveModal() { signalModalLeave.emit(); }

protected:
	void showChangedWarning();
	bool canSave();
	bool shouldSave();

	void onActionUndo();
	void onActionRedo();
	void onActionCut();
	void onActionCopy();
	void onActionPaste();
	void onActionPasteToRoot();
	void onActionDelete();
	void onActionAbout();
	void onActionPreview();
	void onActionDefaultSize();
	void onActionShiftLeft();
	void onActionShiftRight();
	void onActionTranslatables();
	void onActionDefault();
	void onActionAdd();
	void onActionRemove();
	void onActionMoveUp();
	void onActionMoveDown();

	void createActionGroup();
	void createUIManager();

	void updateActions();
	void updateSaveActions();
	void updateExplorerActions();
	void onActionsUpdated(CrowActions changed);
	void setupGtkAction(CrowActions changed, CrowActions da, const char * str);

	void setupAction(CrowActions action, bool enabled);
	CrowActions getActionsForMask(CrowActions mask);
	CrowActions getChangedActions(CrowActions mask, CrowActions old);
	void checkAction(CrowActions action);

	void onClipboardGet(Gtk::SelectionData & data, guint);
	void onClipboardClear();
	void onClipboardReceived(const Gtk::SelectionData & data, bool root);

	void onManagerActionStarted();
	void onManagerActionCompleted();
	void onExplorerSelectionChanged();

	void onBeforeSessionsUpdate();
	void onAfterSessionsUpdate();
	void onHierarchySessionEvent(PSession);
	void onCanvasSessionEvent(PSession);
	void onPropertiesSessionEvent(PSession);
	void onSignalsSessionEvent(PSession);
	void onPackingSessionEvent(PSession);
	void onAfterLoad(PSession session);

	void setRowValues(PPolyelem elem);

	void onPaletteEvent(TypeId type, PaletteEvent event);
	void paste(const Glib::ustring & data, bool root, bool normal);
	void onPastingFinished();
	void onStatusChanged(const Glib::ustring & text);
	void onPopupMenu();

	bool canDefaultSize();
	void setDefaultSize();

	typedef std::pair<PSession, PNode> Selected;
	Selected getSelected();

	bool signalsPageCurrent();
	ExplorerTest testExplorer();

	void doDefault();
	void doAdd();
	void doRemove();
	void doMoveUp();
	void doMoveDown();

	bool canPreview();
	void doPreview();

	void onLeaks();

	int getShiftCapacity(PNode child);
	PNode findShiftNeighbour(PNode child, int index);
	int getChildIndex(PNode child);
	void setChildIndex(PNode child, int index);
	bool canShift();
	void doShift(int shift);

private:
	bool initialized;
	int savePoint;
	Settings settings;

	Glib::ustring statusText, infoText;

	Pool<NodeSet> states;
	PObject savedState;
	NodeSet savedSelection;

	Glib::ustring clipboardStore;
	std::string clipboardTarget;

	CrowActions actions;

	SignalActionsUpdated signalActionsUpdated;
	SignalStatusUpdated signalStatusUpdated;
	SignalInfoUpdated signalInfoUpdated;
	SignalShowUrl signalShowUrl;
	SignalModalEnter signalModalEnter;
	SignalModalLeave signalModalLeave;

	PGtkActionGroup actionGroup;
	PGtkUIManager uiManager;

	Gtk::HPaned hPaned;
	Gtk::VPaned vPanedHP;
	Gtk::HPaned hPanedCE;
	Gtk::VBox explorerVBox;

	PaletteWidget paletteWidget;

	CanvasWidget canvasWidget;
	HierarchyWidget hierarchyWidget;
	ExplorerWidget explorerWidget;

	HierarchySessionSupplier hierarchySessionSupplier;
	CanvasSessionSupplier canvasSessionSupplier;
	PropertiesSessionSupplier propertiesSessionSupplier;
	SignalsSessionSupplier signalsSessionSupplier;
	PackingSessionSupplier packingSessionSupplier;

	SessionManager manager;

};

}

#endif
