/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "document.h"

#include "model.h"
#include "palette.h"
#include "crow_impl.h"

namespace Crow {

Document::Document() {
	model= NULL;
}

Glib::ustring Document::save(Model * mdl, const Nodes & nodes, const NodeSave & nsave) {

	model= mdl;
	nodeSet= model->growSlaves(nodes);
	nodeSave= nsave;

	Glib::ustring doc;
	doc += "<?xml version='1.0' encoding='UTF-8'?>\n";
	doc += "<gui namespace='GTK+' version='" + ToString(CROW_VERSION_GUIXML) + "'>\n";

	NodesSorted sorted(nodeSet.begin(), nodeSet.end());
	Check(sorted.size()==nodeSet.size());
	for(NodesSorted::iterator it=sorted.begin(); it!=sorted.end(); ++it)
		doc += save(*it, 1);

	doc += "</gui>\n";

	nodeSet.clear();
	nodeSave= NodeSave();
	model= NULL;

	return doc;
}

Glib::ustring Document::save(PNode node, int indent) {
	Glib::ustring child;
	Glib::ustring tabs(indent, '\t');

	NodeTest test= nodeSave(node);
	
	Glib::ustring meta= *node->getMeta();

	Glib::ustring attr;
	if(node->getOwner()->getRole()!=nrVector)
		attr += " name='" + *node->getName() + "'";
	if(node->getRole()!=nrLink && node->getRole()!=nrVector)
		attr += " type='" + *Palette::get()->getEntry(node->getType())->getTypeName() + "'";
	if(!meta.empty())
		attr += " meta='" + XmlEncode(meta, true) + "'";

	switch(node->getRole()) {

		case nrScalar: {
			if(test.first) {
				child += tabs + "<scalar" + attr + ">";
				
				if(test.second->hasFlag(sfXml)) {
					
					Glib::ustring str= node->getValue()->toString();
					child += "<![CDATA[\n";
					
					std::vector<Glib::ustring> lines= Split(str, '\n');
					for(int i=0; i<(int)lines.size(); ++i)
						if(!lines[i].empty())
							child += tabs + '\t' + lines[i] + '\n';
					
					child += tabs + "]]>";
				
				} else
					child += XmlEncode(node->getValue()->toString());
				
				child += "</scalar>\n";
			}
			break;
		}

		case nrVector: {
			if(test.first) {
				child += tabs + "<vector" + attr + ">\n";
				NodesSorted sorted(node->getOut()->begin(), node->getOut()->end());
				Check(sorted.size()==node->getOut()->size());
				for(NodesSorted::const_iterator it= sorted.begin(); it!=sorted.end(); ++it)
					child += save(*it, indent+1);
				child += tabs + "</vector>\n";
			}
			break;
		}

		case nrEntity: {
			child += tabs + "<entity" + attr + ">\n";
			NodesSorted sorted(node->getOut()->begin(), node->getOut()->end());
			Check(sorted.size()==node->getOut()->size());
			Glib::ustring domain;
			for(NodesSorted::const_iterator it= sorted.begin(); it!=sorted.end(); ++it)
				domain += save(*it, indent+1);
			if(!test.first && domain.empty())
				child= "";
			else
				child += domain + tabs + "</entity>\n";
			break;
		}

		case nrLink: {
			if(test.first) {
				PNode target= node->getLink();
				child += tabs + "<link" + attr;
				if(target && nodeSet.find(target)!=nodeSet.end())
					child += ">" + *target->getName() + "</link>\n";
				else {
					Check(!target || node->isWeakLink());
					child += "/>\n";
				}
			}
			break;
		}

		default:
			Check(false);

	}

	return child;
}

bool Document::load(Model * mdl, const Glib::ustring & data, Nodes & loaded) {

	model= mdl;
	error= "";
	bool success= true;

	try {

		try {
			Parser::parse(this, data);
		} catch(const Glib::Error & x) {
			if(x.what()=="upgrade") {
				Glib::ustring upgraded= GuiUpgrade().upgrade(data);
				Parser::parse(this, upgraded);
			} else
				throw;
		}

		Check(stack.empty());

		for(Links::iterator it= links.begin(); it!=links.end(); ++it) {
			std::string name= names[it->second];
			PNode target= model->findObject(name);
			if(target)
				model->setLink(it->first, target);
		}

		loaded= objects;

	} catch(const Glib::Error & x) {
		error= x.what();
		success= false;
	}

	stack.clear();
	names.clear();
	links.clear();
	objects.clear();

	model= NULL;

	return success;
}

StateFlags Document::getNodeState(const std::string & name) {
	Stack::iterator it= stack.end();
	--it;
	std::string property;
	if(it->node->getRole()==nrEntity)
		property= name;
	else {
		Check(it->node->getRole()==nrVector);
		property= *it->node->getName();
		--it;
		Check(it->node->getRole()==nrEntity);
	}
	return Palette::get()->getEntry<EntityEntry>(it->node->getType())->getNodeState(property);
}

void Document::onStartElement(const Glib::ustring & element, const Glib::Markup::Parser::AttributeMap & attributes) {

	if(element == "gui") {

		Glib::Markup::Parser::AttributeMap::const_iterator it;
		it= attributes.find("namespace");
		std::string nspace= it==attributes.end() ? "GTK+" : it->second;
		if(nspace!="GTK+")
			RaiseError("GuiXml namespace check failed");
		it= attributes.find("version");
		std::string version= it==attributes.end() ? "" : it->second;
		checkVersion(version);
		
		stack.push_back(SEntry(model->getRoot(), ""));

	} else {

		++stack.back().leafs;

		std::string name;
		std::string type;
		Glib::ustring meta;

		PNode node;

		for(Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.begin(); it!=attributes.end(); ++it)
			if(it->first == "name")
				name= it->second;
			else if(it->first == "type")
				type= it->second;
			else if(it->first == "meta")
				meta= it->second;
			else
				RaiseError("unknown attribute '" + it->first + "'");

		if(stack.size()==1) {
			std::string unique= model->makeUnique(name);
			names[name]= unique;
			name= unique;
		} else if(stack.back().node->getRole()==nrVector) {
			Check(name.empty());
			name= ToString(stack.back().leafs - 1);
		}

		if(element=="scalar")
			node= model->createScalar(stack.back().node, FindType(type), name, getNodeState(name));
		else if(element=="vector")
			node= model->createVector(stack.back().node, name);
		else if(element=="entity")
			node= model->createEntity(stack.back().node, FindType(type), name);
		else if(element=="link")
			node= model->createLink(stack.back().node, name, getNodeState(name));
		else
			RaiseError("unknown element '" + element + "'");

		model->setMeta(node, meta);

		if(stack.size()==1)
			objects.push_back(node);

		stack.push_back(SEntry(node, ""));

	}

}

void Document::onEndElement(const Glib::ustring & name) {
	PNode node= stack.back().node;
	const Glib::ustring & text= stack.back().text;
	if(node->getRole()==nrScalar) {
		Any value= CAny::createFromString(node->getType(), text);
		model->setScalar(node, value);
	} else if(node->getRole()==nrLink) {
		links.push_back(std::make_pair(node, text));
	}
	stack.pop_back();
}

void Document::onText(const Glib::ustring & text) {
	stack.back().text += text;
}

void Document::onPassthrough(const Glib::ustring & text) {
	std::string str= text;
	if(Begins(str, std::string("<![CDATA[\n")) && Ends(str, std::string("]]>"))) {
		Glib::ustring sub= str.substr(10, str.size()-10-3);
		int skip= sub.find_first_not_of('\t');
		std::vector<Glib::ustring> lines= Split(sub, '\n');
		Glib::ustring shifted;
		for(int i=0; i<(int)lines.size(); ++i)
			if((int)lines[i].size() >= skip)
				shifted += lines[i].substr(skip, lines[i].size()-skip) + '\n';
		onText(shifted);
	}
}

void Document::checkVersion(const std::string & vstr) {
	int version= FromString<int>(vstr);
	if(CROW_VERSION_GUIXML == version)
		return;
	if(CROW_VERSION_GUIXML < version)
		RaiseError("gui version check failed");
	if(version <= GuiNeedsUpgrade)
		throw Upgrade();
}

bool Document::idorder::operator () (PNode n1, PNode n2) { 
	return LessSplit(*n1->getName(), *n2->getName()); 
}

// GuiUpgrade ------------------------------------------------------------------

Glib::ustring GuiUpgrade::upgrade(const Glib::ustring & data) {

	try {
		Parser::parse(this, data);
	} catch(const Glib::Error & x) {
		std::cerr << "UI parsing failed: " << x.what() << std::endl;
		return data;
	}

	switch(version) {

		case 1:
			upgrade1();

		case 2:
			upgrade2();

		case 3:
		case 4:
			upgrade4();

		case 5:
			upgrade5();

		case 6:
			upgrade6();

		case 7:
		case 8:
			upgrade8();

	}

	return write();
}

void GuiUpgrade::onStartElement(const Glib::ustring & element, const Glib::Markup::Parser::AttributeMap & attributes) {

	if(element == "gui") {

		Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.find("version");
		std::string vstr= it==attributes.end() ? "" : it->second;
		version= FromString<int>(vstr);
		model.push_back(Node());
		stack.push_back(&model.front());

	} else {

		Node node;

		for(Glib::Markup::Parser::AttributeMap::const_iterator it= attributes.begin(); it!=attributes.end(); ++it)
			if(it->first == "name")
				node.name= it->second;
			else if(it->first == "type")
				node.type= it->second;
			else if(it->first == "flags") {
				std::vector<Glib::ustring> flags= Split(it->second, '|');
				std::copy(flags.begin(), flags.end(), std::back_inserter(node.flags));
			} else if(it->first == "meta")
				node.meta= it->second;
			else
				RaiseError("unknown attribute '" + it->first + "'");

		node.element= element;
		node.owner= stack.back();
		model.push_back(node);
		stack.back()->domain.push_back(&model.back());
		stack.push_back(&model.back());

	}

}

void GuiUpgrade::onEndElement(const Glib::ustring & name) {
	stack.pop_back();
}

void GuiUpgrade::onText(const Glib::ustring & text) {
	stack.back()->text += text;
}

void GuiUpgrade::onPassthrough(const Glib::ustring & text) {
	std::string str= text;
	if(Begins(str, std::string("<![CDATA[\n")) && Ends(str, std::string("]]>"))) {
		Glib::ustring sub= str.substr(10, str.size()-10-3);
		stack.back()->cdata += sub;
	}
}

Glib::ustring GuiUpgrade::write() {

	Glib::ustring doc;
	doc += "<?xml version='1.0' encoding='UTF-8'?>\n";
	doc += "<gui namespace='GTK+' version='" + ToString(CROW_VERSION_GUIXML) + "'>\n";

	std::list<Node *> & nodes= model.front().domain;
	for(std::list<Node *>::iterator it=nodes.begin(); it!=nodes.end(); ++it)
		doc += write(*it, 1);

	doc += "</gui>\n";

	return doc;
}

Glib::ustring GuiUpgrade::write(Node * node, int indent) {

	Glib::ustring child;
	Glib::ustring tabs(indent, '\t');

	Glib::ustring attr;
	attr += " name='" + node->name + "'";
	if(!node->type.empty())
		attr += " type='" + node->type + "'";
	if(!node->meta.empty())
		attr += " meta='" + XmlEncode(node->meta, true) + "'";

	if(node->element=="scalar" || node->element=="link") {
		child += tabs + "<" + node->element + attr + ">";
		if(!node->cdata.empty())
			child += "<![CDATA[\n" + node->cdata + "]]>";
		else
			child += XmlEncode(node->text);
		child += "</" + node->element + ">\n";
	}

	else if(node->element=="vector" || node->element=="entity") {
		child += tabs + "<" + node->element + attr + ">\n";
		std::list<Node *> & nodes= node->domain;
		for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it)
			child += write(*it, indent+1);
		child += tabs + "</" + node->element + ">\n";
	}

	else
		Check(false);

	return child;
}

void GuiUpgrade::clean() {
	std::list<Node>::iterator it= model.begin();
	for(++it; it!=model.end(); ++it)
		if(!it->owner)
			clean(&*it);
	for(--it; it!=model.begin(); --it)
		if(!it->owner)
			it= model.erase(it);
}

void GuiUpgrade::clean(Node * node) {
	std::list<Node *> nodes= node->domain;
	for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		remove(*it);
		clean(*it);
	}
}

GuiUpgrade::Node * GuiUpgrade::find(Node * node, const std::string & name) {
	if(node) {
		std::list<Node *> & nodes= node->domain;
		for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it)
			if((*it)->name==name)
				return *it;
	}
	return NULL;
}

void GuiUpgrade::remove(Node * node) {
	if(node) {
		Check(FindAndErase(node->owner->domain, node));
		node->owner= NULL;
	}
}

void GuiUpgrade::reown(Node * node, Node * owner) {
	if(node && owner) {
		remove(node);
		owner->domain.push_back(node);
		node->owner= owner;
	}
}

void GuiUpgrade::upgrade1() {
	std::list<Node *> elems;
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		FindAndErase(node->flags, "SensitiveLink");
		FindAndErase(node->flags, "ModifyingLink");
		if(node->name=="name" && node->owner->type=="Gtk::ActionGroup")
			FindAndErase(node->flags, "Construct");
		if(node->type=="Gideon::UIElement")
			elems.push_back(node);
	}
	for(std::list<Node *>::iterator it= elems.begin(); it!=elems.end(); ++it) {
		Node * uielement= *it;
		Node * widget= find(uielement, "widget");
		Node * manager= find(uielement, "manager");
		Node * element= find(uielement, "element");

		if(manager)
			FindAndErase(manager->flags, "Construct");
		if(element)
			FindAndErase(element->flags, "Construct");

		std::string ui_element;
		std::vector<Node *> ui_definitions;

		std::string element_str= element ? element->text : "";
		Node * manager_node= manager ? find(&model.front(), manager->text) : NULL;
		if(manager_node) {
			Node * ui_defs= find(manager_node, "ui-definitions");
			if(ui_defs) {
				std::list<Node *> & defs= ui_defs->domain;
				for(std::list<Node *>::iterator it= defs.begin(); it!=defs.end(); ++it)
					ui_definitions.push_back(find(&model.front(), (*it)->text));
			}
		}
		
		for(int i=0; i<(int)ui_definitions.size(); ++i) {
			Node * ui= find(ui_definitions[i], "ui");
			if(ui) {
				ui_element= UiElementExtract().extract(ui->cdata)[element_str];
				if(!ui_element.empty())
					break;
			}
		}
		
		if(ui_element=="menubar")
			uielement->type= "Gtk::MenuBar";
		else
			uielement->type= "Gtk::Toolbar";

		if(widget) {

			reown(find(widget, "border-width"), uielement);
			reown(find(widget, "events"), uielement);
			reown(find(widget, "extension-events"), uielement);
			reown(find(widget, "orientation"), uielement);
			reown(find(widget, "show-arrow"), uielement);
			reown(find(widget, "toolbar-style-set"), uielement);
			reown(find(widget, "toolbar-style"), uielement);
			reown(find(widget, "tooltips"), uielement);
			
			Node * wsignals= find(widget, "signals");
			if(wsignals) {
				Node * signals= find(uielement, "signals");
				if(!signals)
					reown(wsignals, uielement);
				else {
					int shift= signals->domain.size();
					std::list<Node *> nodes= wsignals->domain;
					for(std::list<Node *>::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
						Node * node= *it;
						int index= FromString<int>(node->name);
						node->name= ToString(shift + index);
						reown(node, signals);
					}
				}
			}
			
			remove(widget);
		}
	}

	clean();		
}

void GuiUpgrade::upgrade2() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		bool found= FindAndErase(node->flags, "Translate");
		if(found)
			node->meta= "translate";
	}
}

void GuiUpgrade::upgrade4() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(
			node->name=="position-set" && (
				node->owner->type=="Gtk::HPaned" ||
				node->owner->type=="Gtk::VPaned"
			)
		)
			FindAndErase(node->flags, "NoUndo");
	}
}

std::string GuiUpgrade::replaceColons(const std::string & str, const std::string & rep) {
	std::string result= str;
	std::string::size_type ch= result.find(':');
	if(ch!=std::string::npos) {
		Check(result[ch+1]==':');
		result.replace(ch, 2, rep);
	}
	return result;
}

void GuiUpgrade::upgrade5() {
	std::map<std::string, std::string> rename;
	std::set<std::string> enums;

	rename["Glib::ustring"] = "string";
	rename["Gideon::UniChar"] = "char";
	rename["Gtk::BuiltinIconSize"] = "GtkIconSize";

	enums.insert("GdkExtensionMode");
	enums.insert("GdkGravity");
	enums.insert("GdkWindowTypeHint");
	enums.insert("GtkShadowType");
	enums.insert("GtkPolicyType");
	enums.insert("GtkCornerType");
	enums.insert("GtkPositionType");
	enums.insert("GtkPackOptions");
	enums.insert("GtkButtonBoxStyle");
	enums.insert("GtkWindowType");
	enums.insert("GtkWindowPosition");
	enums.insert("GtkOrientation");
	enums.insert("GtkToolbarStyle");
	enums.insert("GtkResponseType");
	enums.insert("GtkFileChooserAction");
	enums.insert("GtkPackDirection");
	enums.insert("GtkJustification");
	enums.insert("GtkReliefStyle");
	enums.insert("GtkSpinButtonUpdatePolicy");
	enums.insert("GtkUpdateType");
	enums.insert("GtkProgressBarOrientation");
	enums.insert("GtkIconSize");
	enums.insert("GtkSizeGroupMode");
	enums.insert("GtkWrapMode");
	enums.insert("GtkSelectionMode");
	enums.insert("GtkArrowType");
	enums.insert("PangoEllipsizeMode");
	
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		std::map<std::string, std::string>::iterator jt= rename.find(node->type);
		if(jt!=rename.end())
			node->type= jt->second;
		else	
			node->type= replaceColons(node->type);
		if(enums.find(node->type)!=enums.end())
			node->text= Glib::ustring(replaceColons(node->text, "_")).uppercase();
	}
}

void GuiUpgrade::upgrade6() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		node->flags.clear();
		if(node->owner && node->owner->element=="vector") {
			node->name= "";
			if(node->owner->name=="signals") {
				Emitter E= FromString<Emitter>(node->text);
				if(E.signame=="get-accessible")
					remove(node);
				else if(E.signame=="parent-changed") {
					E.signame= "parent-set";
					node->text= ToString(E);
				} else if(E.signame=="style-changed") {
					E.signame= "style-set";
					node->text= ToString(E);
				} else if(E.signame=="toolbar-style-changed") {
					E.signame= "style-changed";
					node->text= ToString(E);
				}
			}
		}
		if(node->type=="GtkPackOptions") {
			node->type= "GideonPackOptions";
			node->text.replace(0, 3, "GIDEON");
		}	
	}
	clean();
}

void GuiUpgrade::upgrade8() {
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(Begins(node->type, std::string("Gideon")))
			node->type.replace(0, 6, "Crow");
		if(node->type=="CrowPackOptions")
			node->text.replace(0, 6, "CROW");
	}
	for(std::list<Node>::iterator it= model.begin(); it!=model.end(); ++it) {
		Node * node= &(*it);
		if(
			node->type == "CrowBoxChild" && (
				node->owner->owner->type == "GtkHButtonBox" || 
				node->owner->owner->type == "GtkVButtonBox"
			)
		)
			node->type= "CrowButtonBoxChild";
	}
}

}
