/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CROW_EDITORS_H
#define CROW_EDITORS_H

#include "util.h"
#include "widgets.h"

namespace Crow {

class Editor : public Object {

public:
	void init(PSession session);
	void done();

	PEditorWidget getEditorWidget();
	template <class T> Glib::RefPtr<T> getEditorWidget() { return Glib::RefPtr<T>::cast_static(getEditorWidget()); }

	virtual void created() { }
	virtual void load(bool reload) { }
	virtual void applySettings() { }
	virtual void storeSettings() { }
	virtual void deleted() { }

	PSession getSession();

	typedef sigc::signal<void> SignalPopupMenu;
	SignalPopupMenu & signal_popup_menu() { return signalPopupMenu; }

	Controller * getController();

protected:
	SessionManager * getManager();

	virtual PEditorWidget createEditorWidget() = 0;
	void setEditorWidget(PEditorWidget);

	void popupMenu() { signalPopupMenu.emit(); }

private:
	PSession session;
	PEditorWidget widget;
	SignalPopupMenu signalPopupMenu;

};

template <class T> class TCreateEditorWidget : public virtual Editor {

public:
	Glib::RefPtr<T> getEditorWidget() { return Editor::getEditorWidget<T>(); }

protected:
	PEditorWidget createEditorWidget() { return NewRefPtr<T>(); }

};

class HierarchyEditor : public TCreateEditorWidget<HierarchyEditorWidget> {

public:
	void created();
	void load(bool reload);
	void deleted();

protected:
	void setRowValues(PPolyelem elem);
	void onManagerSelectionChanged();
	void onTreeViewSelectionChanged();
	bool validator(const Glib::ustring & oldtext, const Glib::ustring & newtext);
	void onRename(PolycellInput * polycell, PNode node);

};

class CanvasEditor : public virtual Editor {

public:
	virtual bool startPasting(const Glib::ustring & data, bool normal) { return false; }
	virtual void stopPasting() { }

	void pastingFinished() { signalPastingFinished.emit(); }
	void statusChanged(const Glib::ustring & str) { signalStatusChanged.emit(str); }

	typedef sigc::signal<void> SignalPastingFinished;
	typedef sigc::signal<void, Glib::ustring> SignalStatusChanged;

	SignalPastingFinished & signal_pasting_finished() { return signalPastingFinished; }
	SignalStatusChanged & signal_status_changed() { return signalStatusChanged; }

private:
	SignalPastingFinished signalPastingFinished;
	SignalStatusChanged signalStatusChanged;

};

class NullCanvasEditor : public CanvasEditor {

protected:
	PEditorWidget createEditorWidget();

};

class PropertyEditor : public virtual Editor {

public:
	PPropertyEditorWidget getEditorWidget();

	void created();

protected:
	virtual void onAccepted() { }

	TypeId getEditingFinalType();
	Property * getProperty1();

	Any getScalar();
	void setScalar(Any value);
	void setScalarAndMeta(Any value, const Glib::ustring & meta);

};

class VectorPropertyEditor : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetText> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetText>::getEditorWidget;

	void load(bool reload);

};

class ObjectPropertyEditor : public PropertyEditor, TCreateEditorWidget<PropertyEditorWidgetInput> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetInput>::getEditorWidget;

	void created();
	void load(bool reload);

protected:
	void onPopup();
	void onMenuClear();
	void onMenuSelect(PNode object);
	void onMenuCreate(TypeId type);

};

class PropertyEditorInput : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetInput> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetInput>::getEditorWidget;

	void created();
	void load(bool reload);

protected:
	void onAccepted();

	virtual Glib::ustring toText(Any value);
	virtual Any fromText(const Glib::ustring & text);

private:
	void onPopup();

};

class PropertyEditorChoice : public PropertyEditor, public TCreateEditorWidget<PropertyEditorWidgetChoice> {

public:
	using TCreateEditorWidget<PropertyEditorWidgetChoice>::getEditorWidget;

	void load(bool reload);

};

class BoolPropertyEditor : public PropertyEditorChoice {

public:
	void created();

protected:
	void onAccepted();

};

class EnumPropertyEditor : public PropertyEditorChoice {

public:
	void created();

protected:
	void onAccepted();

};

class FlagsPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class EmitterPropertyEditor : public PropertyEditorInput {

public:
	EmitterPropertyEditor();
	void created();

protected:
	std::string getMySignature();
	bool validator(const Glib::ustring & oldtext, const Glib::ustring & newtext);
	Glib::ustring toText(Any value);
	Any fromText(const Glib::ustring & text);
	void onPopup();
	void onOrderSelect(Gtk::Menu * menu);
	void onMenuSelect(const std::string & name);

private:
	bool lock;

};

class ColorPropertyEditor : public PropertyEditor, public TCreateEditorWidget<ColorPropertyEditorWidget> {

public:
	using TCreateEditorWidget<ColorPropertyEditorWidget>::getEditorWidget;

	void load(bool reload);

protected:
	void onAccepted();

};

class StockIDPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class IconNamePropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();

};

class ElementPropertyEditor : public PropertyEditorInput {

public:
	void created();

private:
	void onPopup();
	void onMenuSelect(const std::string & choice);

};

}

#endif
