/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef CROW_MODEL_H
#define CROW_MODEL_H

#include "util.h"

namespace Crow {

enum NodeRole { nrNone, nrScalar, nrVector, nrEntity, nrLink };

class Node : public Object {

friend class Model;

public:
	NodeRole getRole() { return role; }
	TypeId getType() { Check(role!=nrVector && role!=nrLink); return type; }
	StateFlags getState() { Check(role!=nrVector && role!=nrEntity); return state; }

	const std::string * getName() { return &name; }
	Any getValue();
	const Glib::ustring * getMeta() { return &meta; }

	bool isInactive();
	bool isModified() { return modified; }

	bool isWeakLink() { return state & sfWeakLink; }
	bool isNoUndo() { return state & sfNoUndo; }
	bool isConstruct() { return state & sfConstruct; }

	const Nodes * getOut() { Check(role!=nrLink && role!=nrScalar); return &out; }
	const Nodes * getIn() { return &in; }

	PNode getOwner();
	PNode getLink();

protected:
	Node(NodeRole r, TypeId t, StateFlags s=sfNone);
	static PNode create(NodeRole r, TypeId t, StateFlags s=sfNone);

	void setModified(bool b) { modified= b; }

	void setName(const std::string & n) { name= n; }
	void setValue(Any any);
	void setMeta(const Glib::ustring & m) { meta= m; }

	void pushOut(PNode node);
	void pushIn(PNode node);

	void eraseOut(PNode node);
	void eraseIn(PNode node);

	void setOwner(PNode owner);
	void setLink(PNode link);

private:
	NodeRole role;
	TypeId type;
	StateFlags state;

	bool modified;

	std::string name;
	Any value;
	Glib::ustring meta;
	Nodes in, out;

	PNode owner;
	PNode link;

};

enum UpdateMode { umNone, umNormal, umSeek, umLoad, umClear, umPaste };

enum OperationCode {
	ocNone,
	ocSetOwner,
	ocModified,
	ocSetName,
	ocSetMeta,
	ocSetValue,
	ocPushIn,
	ocPushOut,
	ocEraseIn,
	ocEraseOut,
	ocSetLink
};

struct Operation : public Object {
	OperationCode code;
	PNode node;

	static POperation create(OperationCode oc, PNode node);
	static POperation create(OperationCode oc, PNode node, const std::string & oldname, const std::string & newname);
	static POperation create(OperationCode oc, PNode node, const Glib::ustring & oldmeta, const Glib::ustring & newmeta);
	static POperation create(OperationCode oc, PNode node, Any oldvalue, Any newvalue);
	static POperation create(OperationCode oc, PNode node, PNode link);
	static POperation create(OperationCode oc, PNode node, PNode oldnode, PNode newnode);

};

struct OperationName : public Operation {
	std::string oldname;
	std::string newname;
};

struct OperationMeta : public Operation {
	Glib::ustring oldmeta;
	Glib::ustring newmeta;
};

struct OperationValue : public Operation {
	Any oldvalue;
	Any newvalue;
};

struct OperationLink : public Operation {
	PNode link;
};

struct OperationNode : public Operation {
	PNode oldnode;
	PNode newnode;
};

typedef unsigned ChangeType;
const ChangeType ctNone   = 0;
const ChangeType ctAppend = 1;
const ChangeType ctRemove = 2;
const ChangeType ctEdit   = 4;

class Model {

public:
	Model();
	~Model();

	PNode getRoot();
	bool empty();
	bool emptyHistory();
	UpdateMode getUpdateMode() { return mode; }
	bool testChange(ChangeType ctype);

	bool validName(const std::string & name);
	std::string makeUnique(const std::string & name);
	std::string suggestName(TypeId type);

	bool isModified();
	void tagModified(PNode node=PNode());

	IdPath getNodePath(PNode node, PNode cut=PNode());
	IdPath getChildPath(PNode node);

	Nodes getNested(PNode node, bool all=true);
	Nodes getChildren(PNode node, bool all=true);
	Nodes getSlaves(PNode node, bool all=true);
	NodeSet growSlaves(const Nodes & nodes);
	Nodes selectMasters(const Nodes & nodes);
	Nodes selectParents(const Nodes & nodes);

	Nodes query(TypeId baseType, bool all=true);
	PNode findObject(const std::string & name);
	PNode find(PNode owner, const std::string & name);
	PNode findParent(PNode node);
	PNode findMaster(PNode node);
	PNode raiseOwner(PNode node);
	PNode raiseParent(PNode node);
	PNode raiseMaster(PNode node);
	bool hasParent(PNode child, PNode parent);
	bool hasMaster(PNode child, PNode master);
	PNode findChildNode(PNode node);
	PNode getWidgetNode(PNode node);
	Any getScalar(PNode owner, const std::string & name);
	bool vectorEmptyOrNullLinks(PNode node);

	PNode createScalar(PNode owner, const std::string & name, Any value, StateFlags sf=sfNone);
	PNode createScalar(PNode owner, TypeId type, const std::string & name, StateFlags sf=sfNone);
	PNode createVector(PNode owner, const std::string & name);
	PNode createEntity(PNode owner, TypeId type, const std::string & name);
	PNode createLink(PNode owner, const std::string & name, StateFlags sf);

	PNode vectorCreateEntity(PNode owner, TypeId type);
	PNode vectorCreateScalar(PNode owner, Any value, StateFlags sf);
	PNode vectorCreateLink(PNode owner, StateFlags sf);

	void cut(PNode node);
	void paste(PNode node, PNode owner);
	void remove(PNode node);
	void rootRemove(PNode entity);
	void vectorClear(PNode node);
	void vectorRemove(PNode node);
	void vectorMove(PNode node, int shift);
	void ensureOut(PNode node, const std::vector<std::string> & names);

	void clearLink(PNode node);
	void setLink(PNode node, PNode target);
	void setScalar(PNode node, Any value);
	void setScalar(PNode owner, const std::string & name, Any value);
	void setName(PNode node, const std::string & name);
	void setMeta(PNode node, const Glib::ustring & meta);
	void setChildNode(PNode widget, PNode child);

	void clearState();
	void begin(UpdateMode mode);
	void clean();
	void seek(int ptr);
	void rollback();
	void commit();

	void clear();

	int getPointer() { return hpointer; }
	int getMarkCount() { return hsize+1; }

	void setReadOnly(bool ro) { readOnly= ro; }

	void clearLinksCleared() { gatherLinks= true; }
	bool getLinksCleared();

protected:
	bool isChildrenVector(PNode node);
	void getNested(PNode node, Nodes * nodes, bool all);
	void getChildren(PNode node, Nodes * nodes, bool all);
	void getSlaves(PNode node, Nodes * nodes, bool all);
	void query(PNode node, TypeId baseType, Nodes * nodes, bool all);

	PNode createNode(NodeRole role, TypeId type, StateFlags sf=sfNone);

	void op_setOwner(PNode node, PNode owner);
	void op_modified(PNode node);
	void op_setName(PNode node, const std::string & name);
	void op_setMeta(PNode node, const Glib::ustring & meta);
	void op_setValue(PNode node, Any value);
	void op_pushIn(PNode node, PNode push);
	void op_pushOut(PNode node, PNode push);
	void op_eraseIn(PNode node, PNode erase);
	void op_eraseOut(PNode node, PNode erase);
	void op_setLink(PNode node, PNode link);

	void setModified(PNode node);
	void clearModified(PNode node);

	bool redundantLog();
	void undo();
	void redo();
	void appendHistory();
	void clearHistory();
	void undo(POperation op);
	void redo(POperation op);

private:
	PNode root;
	UpdateMode mode;
	ChangeType changeType;

	typedef std::vector<POperation> Log;
	typedef std::list<Log> History;

	History history;
	Log log;

	History::iterator pointer;
	int hpointer;
	int hsize;

	bool dirty;
	bool readOnly;
	
	bool gatherLinks;
	Nodes linksCleared;

};

}

#endif
