/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "util.h"

#include "palette.h"
#include "crow_impl.h"
#include "model.h"

namespace Crow {

void CheckFailed(const char * expr, const char * file, unsigned line) {

	#ifdef CROW_SANDBOX
		assert(0);
	#endif

	std::ostringstream oss;
	oss << "Check failed: " << expr << ", file: " << file << ", line: " << line;
	FatalError(oss.str());
}

void FatalError(const std::string & message) {
	g_critical(message.c_str());
	Gtk::MessageDialog dialog(
		"Fatal Error [" G_LOG_DOMAIN "]", false, 
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

std::string GetTypeName(const std::type_info & ti) {
	std::string result;
	#ifdef __GNUC__
		int status;
		char * buf= abi::__cxa_demangle(ti.name(), NULL, NULL, &status);
		result= buf;
		free(buf);
	#else
		result= ti.name();
	#endif
	return result;
}

TypeId FindType(const std::string typeName) {
	return Palette::get()->findType(typeName);
}

PaletteEntry * GetEntry(TypeId type) {
	return Palette::get()->getEntry(type);
}

void DumpParents(const gchar * name) {
	GType type= g_type_from_name(name);
	GType parent= g_type_parent(type);
	if(parent!=0) {
		const gchar * pname= g_type_name(parent);
		std::cout << pname << std::endl;
		DumpParents(pname);
	}
}

void NoOp() {
}

IdPath GetParentPath(const IdPath & path) {
	Check(!path.empty());
	return IdPath(path.begin(), path.end()-1);
}

const char * GetData(Glib::Object * object, const std::string & key) {
	return static_cast<char *>(object->get_data(Glib::QueryQuark(key)));
}

void SetData(Glib::Object * object, const std::string & key, const char * value) {
	object->set_data(Glib::Quark(key), const_cast<char *>(value));
}

GType GetGType(Glib::Object * object) {
	if(object==NULL)
		return G_TYPE_INVALID;
	GType gtype= G_OBJECT_TYPE(object->gobj());
	std::string name= g_type_name(gtype);
	if(Begins(name, std::string("gtkmm__")))
		return g_type_parent(gtype);
	else
		return gtype;
}

Point Transform(const Point & p, Gtk::Widget * src, Gtk::Widget * dest) {
	Point r;
	Check(src->is_realized());
	Check(dest->is_realized());
	Check(src->translate_coordinates(*dest, p.x, p.y, r.x, r.y));
	return r;
}

void SplitNumber(const std::string & name, std::string * str, std::string * num) {
	num->clear();
	int i;
	for(i= name.size()-1; i>=0; --i)
		if(isdigit(name[i]))
			*num += name[i];
		else
			break;
	while(num->size() > 1)
		if((*num)[num->size()-1]=='0' || num->size() > 9) {
			num->resize(num->size()-1);
			++i;
		} else
			break;
	*str= name.substr(0, i+1);
	std::reverse(num->begin(), num->end());
}

void SplitNumber(const std::string & name, std::string * str, int * num) {
	std::string str_num;
	SplitNumber(name, str, &str_num);
	*num= str_num.empty() ? -1 : FromString<int>(str_num);
}

bool LessSplit(const std::string & name1, const std::string & name2) {
	std::string str1, str2;
	int num1, num2;
	SplitNumber(name1, &str1, &num1);
	SplitNumber(name2, &str2, &num2);
	return str1==str2 ? num1 < num2 : str1 < str2;
}

std::string RightPart(const std::string & str, char ch) {
	std::string::size_type pos= str.rfind(ch, str.length()-1);
	std::string name;
	if(pos == std::string::npos)
		return str;
	else
		return str.substr(pos+1, str.length()-pos-1);
}

std::vector<Glib::ustring> Split(const Glib::ustring & str, gunichar ch, int count) {
	std::vector<Glib::ustring> result;
	gchar ** vec= g_strsplit(str.c_str(), Glib::ustring(1, ch).c_str(), count);
	int len= g_strv_length(vec);
	result.reserve(len);
	for(int i=0; i<len; ++i)
		result.push_back(vec[i]);
	g_strfreev(vec);
	return result;
}

Glib::ustring XmlEncode(const Glib::ustring & str, bool attr) {
	Glib::ustring result;
	for(Glib::ustring::const_iterator it= str.begin(); it!=str.end(); ++it) {
		gunichar ch= *it;
		if(Glib::Unicode::isprint(ch)) {
			switch(ch) {
				case '<':
					result += "&lt;";
					continue;
				case '>':
					result += "&gt;";
					continue;
				case '&':
					result += "&amp;";
					continue;
				case '"':
					if(!attr)
						break;
					result += "&quot;";
					continue;
				case '\'':
					if(!attr)
						break;
					result += "&apos;";
					continue;
			}
			result += ch;
		} else {
			std::ostringstream oss;
			oss << "&#x" << std::hex << safe_cast<unsigned>(ch) << ";";
			result += oss.str();
		}
	}
	return result;
}

bool ValidName(const std::string & name) {
	if(name.empty())
		return false;
	if(name[0] != '_' && !isalpha(name[0]))
		return false;
	for(int i=1; i<(int)name.size(); ++i)
		if(!isalpha(name[i]) && !isdigit(name[i]) && name[i] != '_')
			return false;
	return true;
}

void DrawGrid(PGdkWindow window, PGtkStyle style, bool squares) {
	PGdkGC gc= Gdk::GC::create(window);
	int W, H;
	window->get_size(W, H);
	gc->set_foreground(GetDark(style));
	window->draw_rectangle(gc, false, 1, 1, W-3, H-3);
	if(squares)
		gc->set_foreground(GetLight(style));
	int GS= 9;
	int GW= W-4;
	int GH= H-4;
	int xx= GW/GS;
	int yy= GH/GS;
	int GX= 2 + (GW-xx*GS)/2;
	int GY= 2 + (GH-yy*GS)/2;
	if(!squares) {
		for(int x=0; x<xx; ++x)
			for(int y=0; y<yy; ++y)
				window->draw_point(gc, GX+x*GS+GS/2, GY+y*GS+GS/2);
	} else {
		for(int x=0; x<xx-1; ++x)
			for(int y=0; y<yy-1; ++y)
				if((x+y) % 2 == 0)
					window->draw_rectangle(gc, true, GX+x*GS+GS/2, GY+y*GS+GS/2, GS, GS);
	}
}

void RaiseError(const Glib::ustring & msg) {
	throw RuntimeError(msg);
}

int GetRefCount(const Glib::ObjectBase * object) {
	return object->gobj()->ref_count;
}

void ContainerClear(Gtk::Container * container) {
	std::vector<Gtk::Widget *> widgets= container->get_children();
	for(int i=0; i<(int)widgets.size(); ++i)
		ContainerRemove(container, widgets[i]);
}

void ContainerRemove(Gtk::Container * container, Gtk::Widget * widget) {
	gtk_container_remove(container->gobj(), widget->gobj());
}

void BinRemove(Gtk::Bin * bin) {
	Gtk::Widget * widget= bin->get_child();
	if(widget)
		gtk_container_remove(bin->Gtk::Container::gobj(), widget->gobj());
}

NodeSet ToNodeSet(const Nodes & nodes) {
	NodeSet result(nodes.begin(), nodes.end());
	Check(result.size()==nodes.size());
	return result;
}

Gdk::Color GetWhite(PGtkStyle style) {
	return style->get_light(Gtk::STATE_ACTIVE);
}

Gdk::Color GetBlack(PGtkStyle style) {
	return style->get_dark(Gtk::STATE_SELECTED);
}

Gdk::Color GetLight(PGtkStyle style) {
	return style->get_mid(Gtk::STATE_NORMAL);
}

Gdk::Color GetDark(PGtkStyle style) {
	return style->get_dark(Gtk::STATE_NORMAL);
}

std::string IdPathToString(const IdPath & path, bool dots) {
	std::string str;
	if(!dots)
		str += '/';
	for(int i=0; i<(int)path.size(); ++i) {
		str += path[i].toString();
		if(i < (int)path.size()-1)
			str += dots ? '.' : '/';
	}
	return str;
}

void SortByName(Nodes * nodes) {
	struct tmp { static bool less(PNode a, PNode b) {
		return LessSplit(*a->getName(), *b->getName());
	}};
	nodes->sort(sigc::ptr_fun(&tmp::less));
}

void SetColumnWidth(Gtk::TreeViewColumn * column, int width) {
	column->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
	column->set_fixed_width(width);
	column->set_resizable(true);
}

std::string Trim(const std::string & str) {
	std::string::size_type p0= str.find_first_not_of(' ');
	if(p0 != std::string::npos) {
		std::string::size_type p1= str.find_last_not_of(' ');
		return str.substr(p0, p1-p0+1);
	}
	return std::string();
}

bool Equal(const PObject & v1, const PObject & v2) {
	return (v1 && v2) ? v1->equals(v2) : (!v1 && !v2);
}

Settings HashTableToSettings(GHashTable * table) {
	Settings result;
	struct tmp { static void add(gpointer key, gpointer value, gpointer user_data) {
		(*static_cast<Settings *>(user_data))[static_cast<gchar *>(key)]= static_cast<gchar *>(value);
	} };
	g_hash_table_foreach(table, &tmp::add, &result);
	return result;
}

void SettingsToHashTable(const Settings & settings, GHashTable * table) {
	for(Settings::const_iterator it= settings.begin(); it!=settings.end(); ++it) {
		gchar * key= g_strdup(it->first.c_str());
		gchar * value= g_strdup(it->second.c_str());
		g_hash_table_insert(table, key, value);
	}
}

void PrepareGlibObject(Glib::Object * obj) {
	Gtk::Object * gtkobj= dynamic_cast<Gtk::Object *>(obj);
	if(gtkobj) {
		Check(!dynamic_cast<Gtk::Window *>(gtkobj));
		gtkobj->set_manage();
		Check(g_object_is_floating(obj->gobj()));
		g_object_ref_sink(obj->gobj());
	}
}

int TypeDist(GType sub, GType type) {
	Check(g_type_is_a(sub, type));
	int dist= 0;
	GType gt= sub;
	for(;;) {
		gt= g_type_parent(gt);
		if(gt == G_TYPE_INVALID || !g_type_is_a(gt, type))
			return dist;
		++dist;
	}
}

template <> Glib::ustring ToString<bool>(const bool & value) {
	return value ? "true" : "false";
}

template <> bool FromString<bool>(const Glib::ustring & str) {
	if(str=="true")
		return true;
	if(str=="false")
		return false;
	RaiseError("FromString() failed");
	return false;
}

template <> Glib::ustring ToString<Gdk::Color>(const Gdk::Color & value) {
	unsigned pixel=
		((value.get_red() >> 8) << 16) |
		((value.get_green() >> 8) << 8) |
		(value.get_blue() >> 8);
	std::ostringstream oss;
	oss << "#";
	oss << std::setw(6) << std::setfill('0');
	oss << std::hex << std::uppercase;
	oss << pixel;
	return oss.str();
}

template <> Gdk::Color FromString<Gdk::Color>(const Glib::ustring & str) {
	Gdk::Color color;
	if(!color.set(str))
		RaiseError("FromString() failed");
	return color;
}

template <> Glib::ustring ToString<Gtk::StockID>(const Gtk::StockID & value) {
	return value.get_string();
}

template <> Gtk::StockID FromString<Gtk::StockID>(const Glib::ustring & str) {
	return Gtk::StockID(str);
}

// RuntimeError ----------------------------------------------------------------

RuntimeError::RuntimeError(const Glib::ustring & msg) :
	Glib::Error(CROW_ERROR, CROW_ERROR_RUNTIME, msg)
{
}

RuntimeError::RuntimeError() :
	Glib::Error(CROW_ERROR, CROW_ERROR_RUNTIME, "")
{
}

// Object ----------------------------------------------------------------------

Object::Object() {
	refcounter= 0;
	CrowInit::get()->objectCreated(this);
}

Object::~Object() {
	Check(refcounter==0);
	CrowInit::get()->objectDeleted(this);
}

void Object::reference() { 
	++refcounter; 
}

void Object::unreference() {
	Check(refcounter > 0);
	--refcounter;
	if(refcounter==0)
		delete this;
}

bool Object::equals(PObject po) const {
	return this == GetPtr(po);
}

// UniChar ---------------------------------------------------------------------

template <> Glib::ustring ToString<UniChar>(const UniChar & value) {
	gunichar ch= value;
	return Glib::ustring(1, ch);
}

template <> UniChar FromString<UniChar>(const Glib::ustring & str) {
	if(str.size()!=1)
		RaiseError("FromString() failed");
	gunichar ch= str[0];
	return UniChar(ch);
}

template <> UniChar ExtractCppValue<UniChar>(const Glib::ValueBase & val) {
	gunichar ch= ExtractCppValue<gunichar>(val);
	return UniChar(ch);
}

template <> Glib::ValueBase MakeGValue<UniChar>(const UniChar & val) {
	gunichar ch= val;
	return MakeGValue<gunichar>(ch);
}

// Translatable ----------------------------------------------------------------

Translatable::Translatable(const Glib::ustring & meta) {
	if(meta.empty()) {
		trans= false;
		valid= true;
	} else {
		std::vector<Glib::ustring> vstr= Split(meta, '|', 3);
		if(vstr.size() >= 1) {
			trans= vstr[0]=="translate";
			if(vstr.size() >= 2) {
				prefix= vstr[1];
				if(vstr.size() >= 3)
					comments= vstr[2];
			}
		}
		testValid();
	}
	Check(valid);
}

Translatable::Translatable(bool trans, const Glib::ustring & prefix, const Glib::ustring & comments) {
	this->trans= trans;
	this->prefix= prefix;
	this->comments= comments;
	testValid();
}

void Translatable::testValid() {
	valid= 
		std::string(prefix).find('|')==std::string::npos && 
		std::string(comments).find("*/")==std::string::npos
	;
}

Glib::ustring Translatable::getMeta() const {
	Check(valid);
	Glib::ustring result;
	if(trans)
		result += "translate";
	if(!prefix.empty() || !comments.empty())
		result += '|';
	if(!prefix.empty())
		result += prefix;
	if(!comments.empty())
		result += '|' + comments;
	return result;
}

// Point -----------------------------------------------------------------------

Point::Point() {
	x= 0;
	y= 0;
}

Point::Point(int x, int y) {
	this->x= x;
	this->y= y;
}

bool Point::inside(const Rectangle & r) const {
	return
		r.x <= x && x <= r.x + r.w - 1 &&
		r.y <= y && y <= r.y + r.h - 1
	;
}

Point Point::operator + (const Point & p) const {
	return Point(x+p.x, y+p.y);
}

Point Point::operator - (const Point & p) const {
	return Point(x-p.x, y-p.y);
}

bool Point::operator == (const Point & p) const {
	return x==p.x && y==p.y;
}

Point Point::operator - () const {
	return Point(-x, -y);
}

template <> Glib::ustring ToString<Point>(const Point & value) {
	return ToString(value.x) + "," + ToString(value.y);
}

template <> Point FromString<Point>(const Glib::ustring & str) {
	std::vector<Glib::ustring> split= Split(str, ',');
	if(split.size()==1)
		return Point(FromString<int>(split.front()), 0);
	if(split.size()!=2)
		RaiseError("FromString() failed");
	if(Trim(split.front()).empty())
		return Point(0, FromString<int>(split.back()));
	else
		return Point(FromString<int>(split.front()), FromString<int>(split.back()));
}

// Rectangle -------------------------------------------------------------------

Rectangle::Rectangle() {
	x= y= 0;
	w= h= 0;
}

Rectangle::Rectangle(int x, int y, int w, int h) {
	this->x= x;
	this->y= y;
	this->w= w;
	this->h= h;
}

Rectangle::Rectangle(const Point & point, const Point & size) {
	x= point.x;
	y= point.y;
	w= size.x;
	h= size.y;
}

Rectangle::Rectangle(const Gdk::Rectangle & rect) {
	x= rect.get_x();
	y= rect.get_y();
	w= rect.get_width();
	h= rect.get_height();
}

Rectangle::operator Gdk::Rectangle () const {
	return Gdk::Rectangle(x, y, w, h);
}

bool Rectangle::operator == (const Rectangle & r) const {
	return x==r.x && y==r.y && w==r.w && h==r.h;
}

void Rectangle::grow(int dx, int dy, int dw, int dh) {
	x += dx;
	y += dy;
	w += dw;
	h += dh;
}

void Rectangle::grow(int dx, int dy, int dw, int dh, int minw, int minh) {
	int L= x;
	int T= y;
	int R= x+w;
	int B= y+h;
	growSegment(&L, &R, dx, dx+dw, minw);
	growSegment(&T, &B, dy, dy+dh, minh);
	x= L;
	y= T;
	w= R-L;
	h= B-T;
}

void Rectangle::growSegment(int * a, int * b, int da, int db, int min) {
	if(*b+db-*a-da < min) {
		int k1= min+*a-*b;
		int k2= db-da;
		int da1= (da*k1)/k2;
		int db1= (db*k1)/k2;
		Check(da1*k2 == da*k1);
		Check(db1*k2 == db*k1);
		*a += da1;
		*b += db1;
	} else {
		*a += da;
		*b += db;
	}
}

bool Rectangle::empty() const {
	return w <= 0 || h <= 0;
}

// Border ----------------------------------------------------------------------

Border::Border() {
	left= right= top= bottom= 0;
}

bool Border::operator == (const Border & b) const {
	return
		left == b.left &&
		right == b.right &&
		top == b.top &&
		bottom == b.bottom;
}

Border::operator Gtk::Border () const {
	Gtk::Border border;
	border.left= left;
	border.right= right;
	border.top= top;
	border.bottom= bottom;
	return border;
}

template <> Glib::ustring ToString<Border>(const Border & value) {
	return 
		ToString(value.left) + "," + 
		ToString(value.right) + "," + 
		ToString(value.top) + "," + 
		ToString(value.bottom);
}

template <> Border FromString<Border>(const Glib::ustring & str) {
	std::vector<Glib::ustring> split= Split(str, ',');
	if(split.size()!=4)
		RaiseError("FromString() failed");
	Border result;
	result.left= FromString<int>(split[0]);
	result.right= FromString<int>(split[1]);
	result.top= FromString<int>(split[2]);
	result.bottom= FromString<int>(split[3]);
	return result;
}

// Emitter ---------------------------------------------------------------------

Emitter::Emitter(const std::string & sname, const std::string & mname, bool after) {
	signame= sname;
	mapname= mname;
	this->after= after;
}

bool Emitter::operator == (const Emitter & e) const {
	return signame==e.signame && mapname==e.mapname && after==e.after;
}

template <> Glib::ustring ToString<Emitter>(const Emitter & value) {
	return value.signame + "," + value.mapname + "," + ToString(value.after);
}

template <> Emitter FromString<Emitter>(const Glib::ustring & str) {
	std::vector<Glib::ustring> split= Split(str, ',');
	if(split.size()!=3)
		RaiseError("FromString() failed");
	return Emitter(split[0], split[1], FromString<bool>(split[2]));
}

// Parser ----------------------------------------------------------------------

Parser::Parser() :
	glibParser(this)
{
}

void Parser::parse(Parser * parser, const Glib::ustring & data) {
	Glib::Markup::ParseContext context(parser->glibParser);
	context.parse(data);
	context.end_parse();
}

void Parser::GlibParser::raiseError(const Glib::Exception & x) {
	throw Glib::MarkupError(Glib::MarkupError::INVALID_CONTENT, x.what());
}

void Parser::GlibParser::on_start_element(Glib::Markup::ParseContext & context, const Glib::ustring & name, const AttributeMap & properties) {
	try {
		parser->onStartElement(name, properties);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_end_element(Glib::Markup::ParseContext & context, const Glib::ustring & name) {
	try {
		parser->onEndElement(name);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_text(Glib::Markup::ParseContext & context, const Glib::ustring & text) {
	try {
		parser->onText(text);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

void Parser::GlibParser::on_passthrough(Glib::Markup::ParseContext & context, const Glib::ustring & text) {
	try {
		parser->onPassthrough(text);
	} catch(const Glib::Exception & x) {
		raiseError(x);
	}
}

// UiElementExtract ------------------------------------------------------------

UiElementExtract::emap UiElementExtract::extract(const Glib::ustring & str) {
	depth= 0;
	result.clear();
	try {
		Parser::parse(this, str);
	} catch(const Glib::Error & x) {
		FatalError("UI parsing failed: " + x.what());
	}
	return result;
}

void UiElementExtract::onStartElement(const Glib::ustring & element, const Glib::Markup::Parser::AttributeMap & attributes) {

	++depth;

	if(depth == 2) {

		Glib::Markup::Parser::AttributeMap::const_iterator iname= attributes.find("name");
		Glib::Markup::Parser::AttributeMap::const_iterator iaction= attributes.find("action");
		Glib::ustring name, action;
	
		if(iname!=attributes.end())
			name= iname->second;
		if(iaction!=attributes.end())
			action= iaction->second;

		Glib::ustring realname;
		
		if(!name.empty())
			realname= name;
		else if(!action.empty())
			realname= action;
		else
			realname= element;

		result.insert(std::make_pair(realname, element));
	
	}

}

}
