/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "views.h"

#include "palette.h"
#include "controller.h"
#include "widgets.h"

namespace Crow {

// Property --------------------------------------------------------------------

Property::Property() {
	role= prNone;
	type= InvalidTypeId;
	state= sfNone;
	editor= InvalidEditorId;
	elementEditor= InvalidEditorId;
}

Property::Property(
	const std::string & n,
	PropertyRole r,
	TypeId t
) {

	name= n;
	role= r;
	type= t;
	state= sfNone;

	switch(role) {
		
		case prVector:
			editor= Palette::get()->findEditor("VECTOR");
			elementEditor= Palette::get()->getEntry(type)->getPropertyEditor();
			inert= defaultValue= CAny::createVector(AnyVector());
			break;
	
		case prEntity:
			inert= defaultValue= CAny::createObject();
		
		default:
			editor= Palette::get()->getEntry(type)->getPropertyEditor();
			elementEditor= InvalidEditorId;

	}

}

Any Property::getValue() {
	Any value= valueGet(this);
	return value;
}

void Property::setValue(Any value) {
	Check(!hasFlag(sfReadOnly));
	Any oldv= getValue();
	if(hasFlag(sfOverwrite) || !Equal(value, oldv))
		valueSet(this, value);
}

void Property::configureView(PEntityView view) {
	if(!view->isConfigured()) {
		view->setConfigured();
		if(viewConfigure)
			viewConfigure(this, view);
	}
}

PNode Property::insertElement(int index) {
	if(elementInsert)
		return elementInsert(this, index);
	else
		return PNode();
}

std::string Property::getLabel(PNode node) {
	if(labelGet)
		return labelGet(this, node);
	else
		return *node->getName();
}

bool Property::canEdit() {
	return isEditingAllowed() && !isReadOnly();
}

Property & Property::setEditor(const std::string & ename) {
	editor= Palette::get()->findEditor(ename);
	return *this;
}

Property & Property::setElementEditor(const std::string & ename) {
	elementEditor= Palette::get()->findEditor(ename);
	return *this;
}

Any Property::getDefault() {
	return defaultValue;
}

Property & Property::setDefault(Any dvalue) {
	defaultValue= dvalue;
	return *this;
}

Any Property::getInert() {
	return inert;
}

Property & Property::setInert(Any value) {
	inert= value;
	return *this;
}

Property & Property::setInertDefault(Any value) {
	inert= value;
	defaultValue= value;
	return *this;
}

// EntityView ------------------------------------------------------------------

EntityView::EntityView() {
	controller= NULL;
	type= InvalidTypeId;
	configured= false;
	internal= false;
	onceWritten= false;
}

void EntityView::prepare(TypeId tid) {
	type= tid;
}

void EntityView::init(PNode node, Controller * ctrlr, Any val) {
	modelNode= node;
	controller= ctrlr;
	Check(!onceWritten);

	if(val) {
		internal= true;
		setValue(val);
	} else
		setValue(createInstance());

	initInstance();
	initDefaultValues();

	if(!internal && isEditing())
		initDesign();

}

void EntityView::done() {
	doneInstance();
	clearValue();
	modelNode.clear();
	controller= NULL;
}

bool EntityView::reconstruct() {
	Check(value && onceWritten);
	for(Properties::iterator it= properties.begin(); it!=properties.end(); ++it) {
		Property * property= &(*it);
		if(property->hasFlag(sfConstruct)) {
			PNode node= getModel()->find(getModelNode(), *property->getName());
			if(node && node->isModified())
				return true;
		}
	}
	return false;
}

GType EntityView::getGType() {
	GType gtype= Palette::get()->getEntry(getType())->getGType();
	return gtype;
}

PNode EntityView::getModelNode() {
	return modelNode;
}

Any EntityView::getValue() {
	return value;
}

void EntityView::setConfigured() {
	Check(!configured);
	configured= true;
}

void EntityView::update() {

	Check(!modelNode->isInactive());

	for(Properties::iterator it= properties.begin(); it!=properties.end(); ++it) {

		Property * property= &(*it);

		if((property->getRole()!=prEntity && property->isReadOnly()) || property->hasFlag(sfConstruct))
			continue;

		PNode node= getModel()->find(modelNode, *property->getName());

		if(!(node && (node->isModified() || !onceWritten)))
			continue;

		Any propertyValue;
		bool set= false;

		switch(property->getRole()) {

			case prScalar:
				set= controller->modelGetScalar(
					node,
					&propertyValue,
					modelNode,
					*property->getName(),
					property->getNodeState()
				);
				break;

			case prVector:
				set= controller->modelGetVector(
					node,
					&propertyValue,
					modelNode,
					*property->getName(),
					property->getType(),
					property->getNodeState()
				);
				break;

			case prEntity:
				set= controller->modelGetEntity(
					node,
					&propertyValue,
					modelNode,
					*property->getName(),
					property->getType(),
					property
				);
				break;

			default:
				Check(false);

		}

		if(set && !property->hasFlag(sfReadOnly))
			property->setValue(propertyValue);

	}

}

void EntityView::write() {

	Check(!modelNode->isInactive());

	for(Properties::iterator it= properties.begin(); it!=properties.end(); ++it) {

		Property * property= &(*it);

		switch(property->getRole()) {

			case prScalar:
				controller->modelSetScalar(
					modelNode,
					*property->getName(),
					property->getValue(),
					property
				);
				break;

			case prVector:
				controller->modelSetVector(
					modelNode,
					property->getType(),
					*property->getName(),
					property->getInert()
				);
				break;

			case prEntity:
				controller->modelSetEntity(
					modelNode,
					*property->getName(),
					property->getValue(),
					property
				);
				break;

			default:
				Check(false);

		}

	}

	onceWritten= true;

}

Property * EntityView::findProperty(const std::string & pname) {
	for(Properties::iterator it=properties.begin(); it!=properties.end(); ++it) {
		std::string nn= *it->getName();
		if(nn == pname)
			return &(*it);
	}
	return NULL;
}

Model * EntityView::getModel() {
	return controller->getModel();
}

bool EntityView::isEditing() {
	UpdateMode umode= controller->getModel()->getUpdateMode();
	return umode==umNormal;
}

bool EntityView::isSeeking() {
	UpdateMode umode= controller->getModel()->getUpdateMode();
	return umode==umSeek;
}

bool EntityView::isRoot() {
	return modelNode->getOwner() == controller->getModel()->getRoot();
}

void EntityView::setValue(Any val) {
	Check(!value && val);
	value= val;
}

void EntityView::clearValue() {
	value.clear();
}

Any EntityView::createInstance() {
	return Any();
}

void EntityView::initDefaultValues() {
	for(Properties::iterator it= properties.begin(); it!=properties.end(); ++it) {
		Property * property= &(*it);
		if(property->hasFlag(sfAutoDefault)) {
			Check(property->getRole()==prScalar);
			Any value= property->getValue();
			property->setDefault(value);
		}
	}
}

Property & EntityView::addProperty(
	const std::string & name,
	PropertyRole role,
	const std::string & type,
	Any dvalue
) {
	TypeId typeId= Palette::get()->findType(type);
	
	#ifdef CROW_SANDBOX
		Check(!findProperty(name));
	#endif
	
	properties.push_back(Property(name, role, typeId));
	if(dvalue)
		properties.back().setDefault(dvalue);
	return properties.back();
}

Property & EntityView::addInertProperty(
	const std::string & name,
	PropertyRole role,
	const std::string & type,
	Any dvalue
) {
	Property & prop= addProperty(name, role, type);
	prop.setValueGet(sigc::mem_fun(*this, &EntityView::getPropertyValueInert));
	prop.setValueSet(sigc::mem_fun(*this, &EntityView::setPropertyValueInert));
	if(dvalue)
		prop.setInertDefault(dvalue);
	return prop;
}

Any EntityView::getPropertyValueInert(Property * property) {
	return property->getInert();
}

void EntityView::setPropertyValueInert(Property * property, Any value) {
	property->setInert(value);
}

void EntityView::moveBefore(const std::string & pname, const std::string & before) {
	for(Properties::iterator it=properties.begin(); it!=properties.end(); ++it)
		if(*it->getName() == before) {
			Properties::iterator jt= it;
			for(++jt; jt!=properties.end(); ++jt)
				if(*jt->getName() == pname) {
					properties.splice(it, properties, jt);
					return;
				}
			Check(false);
		}
	Check(false);
}

void EntityView::touch(const std::string & str) {
	touch(modelNode, str);
}

void EntityView::touch(PNode node, const std::string & str) {
	PNode prop= controller->getModel()->find(node, str);
	if(prop)
		controller->touch(prop);
}

Property * EntityView::clearVectorProperty(const std::string & name) {
	Property * property= findProperty(name);
	Check(property->getRole()==prVector);
	property->setInert(CAny::createVector(AnyVector()));
	return property;
}

bool EntityView::propertyModelVectorEmpty(const std::string & name) {
	PNode node= controller->getModel()->find(modelNode, name);
	if(node) {
		Check(node->getRole()==nrVector);
		return node->getOut()->empty();
	}
	return true;
}

Any EntityView::getPropertyModel(const std::string & name) {
	PNode node= getModel()->find(getModelNode(), name);
	Check(!node || node->getRole()==nrScalar);
	return node ? node->getValue() : Any();
}

// GlibObjectView --------------------------------------------------------------

GlibObjectView::GlibObjectView() {
	reportDestruction= true;
}

Any GlibObjectView::getPropertyValue(Property * property) {
	TypeId type= property->getType();
	PaletteEntry * entry= Palette::get()->getEntry(type);
	Glib::ValueBase value;
	GType gt= entry->getGType();
	value.init(gt);
	getObject()->get_property_value(*property->getName(), value);
	return entry->createFromGValue(value);
}

void GlibObjectView::setPropertyValue(Property * property, Any value) {
	getObject()->set_property_value(*property->getName(), value->toGValue());
}

void GlibObjectView::setPropertyValueInertGet(Property * property, Any value) {
	property->setInert(value);
	setPropertyValue(property, value);
}

PNode GlibObjectView::insertObject(Property * property, int index) {
	PNode vector= getModel()->find(getModelNode(), *property->getName());
	PNode node= getModel()->vectorCreateLink(vector, property->getNodeState());
	getModel()->vectorMove(node, index - vector->getOut()->size() + 1);
	return node;
}

PNode GlibObjectView::insertEntity(Property * property, int index) {
	PNode vector= getModel()->find(getModelNode(), *property->getName());
	PNode node= getModel()->vectorCreateEntity(vector, property->getType());
	getModel()->vectorMove(node, index - vector->getOut()->size() + 1);
	return node;
}

PNode GlibObjectView::insertScalar(Property * property, int index, Any value) {
	PNode vector= getModel()->find(getModelNode(), *property->getName());
	PNode node= getModel()->vectorCreateScalar(vector, value, property->getNodeState());
	getModel()->vectorMove(node, index - vector->getOut()->size() + 1);
	return node;
}

std::string GlibObjectView::findSignal(const std::string & name) {
	return Palette::get()->findSignal(name, getGType());
}

PGlibObject GlibObjectView::getObject() {
	return getValue()->getObject();
}

void GlibObjectView::doneInstance() {
	PGlibObject obj= getObject();
	if(!isInternal() && reportDestruction)
		getController()->objectDestroyed(obj);
}

Property & GlibObjectView::addProperty(
	const std::string & name,
	PropertyRole role,
	const std::string & type,
	Any dvalue
) {
	return EntityView::addProperty(name, role, type, dvalue).
		setValueGet(sigc::mem_fun(*this, &GlibObjectView::getPropertyValue)).
		setValueSet(sigc::mem_fun(*this, &GlibObjectView::setPropertyValue));
}

Property & GlibObjectView::addInertGetProperty(
	const std::string & name,
	PropertyRole role,
	const std::string & type,
	Any dvalue
) {
	return EntityView::addInertProperty(name, role, type, dvalue).
		setValueSet(sigc::mem_fun(*this, &GlibObjectView::setPropertyValueInertGet));
}

PNode GlibObjectView::insertSignal(Property * property, int index) {
	SignalSelectDialog dialog;
	dialog.set_transient_for(*getController()->getContext()->getWindow());
	dialog.setChoices(Palette::get()->querySignals(getGType()));
	if(getController()->getContext()->runDialog(&dialog) == Gtk::RESPONSE_OK) {
		PNode vector= getModel()->find(getModelNode(), "signals");
		std::string signal= dialog.getSelection();
		Any emitter= CAny::createEmitter(Emitter(signal, "", true));
		PNode node= getModel()->vectorCreateScalar(vector, emitter, sfNone);
		getModel()->vectorMove(node, index - vector->getOut()->size() + 1);
		return node;
	} else
		return PNode();
}

std::string GlibObjectView::getSignalLabel(Property * property, PNode node) {
	const Emitter & emitter= *node->getValue()->get<Emitter>();
	return (emitter.after ? "" : "\342\200\242") + emitter.signame;
}

Property & GlibObjectView::addSignalsProperty() {
	Property * prop= findProperty("signals");
	if(prop)
		return *prop;
	else
		return addInertProperty("signals", prVector, "CrowEmitter").
			setElementInsert(sigc::mem_fun(*this, &GlibObjectView::insertSignal)).
			setLabelGet(sigc::mem_fun(*this, &GlibObjectView::getSignalLabel));
}

// GtkWidgetView ---------------------------------------------------------------

GtkWidgetView::GtkWidgetView() {
	interactive= false;
	addSignalsProperty();
	addProperty("size-request", prScalar, "CrowPoint", CAny::createPoint(Point(-1, -1))).
		setValueGet(sigc::mem_fun(*this, &GtkWidgetView::getSizeRequest)).
		setValueSet(sigc::mem_fun(*this, &GtkWidgetView::setSizeRequest));
	addProperty("sensitive", prScalar, "bool", CAny::createBool(true));
	addInertProperty("visible", prScalar, "bool", CAny::createBool(true));
	addInertProperty("design-size", prScalar, "CrowPoint", CAny::createPoint(Point(-1, -1))).
		setFlag(sfHidden);
}

PNode GtkWidgetView::createChild(const Place & place, PNode widget) {
	return PNode();
}

Rectangle GtkWidgetView::getChildArea() {
	return getWidgetRect(getObject());
}

PGtkWidget GtkWidgetView::getObject() {
	return PGtkWidget::cast_static(GlibObjectView::getObject());
}

void GtkWidgetView::initInstance() {
	GlibObjectView::initInstance();
	getObject()->show();
	getObject()->property_can_focus()= false;
}

Any GtkWidgetView::getSizeRequest(Property * property) {
	int width, height;
	getObject()->get_size_request(width, height);
	return CAny::createPoint(Point(width, height));
}

void GtkWidgetView::setSizeRequest(Property * property, Any value) {
	const Point & size= *value->getPoint();
	getObject()->set_size_request(size.x, size.y);
}

Glib::ustring GtkWidgetView::getDesignLabel() {
	return isRoot() ? *getModelNode()->getName() : "text";
}

void GtkWidgetView::addReceivesDefaultProperty(bool defval) {
	Property * property= findProperty("receives-default");
	if(!property)
		addProperty("receives-default", prScalar, "bool", CAny::createBool(defval));
	else
		property->setDefault(CAny::createBool(defval));
}

void GtkWidgetView::addCanFocusProperty(bool defval) {
	Property * property= findProperty("can-focus");
	if(!property)
		addInertProperty("can-focus", prScalar, "bool", CAny::createBool(defval));
	else
		property->setDefault(CAny::createBool(defval));
}

void GtkWidgetView::addCanDefaultProperty(bool defval) {
	Property * property= findProperty("can-default");
	if(!property)
		addProperty("can-default", prScalar, "bool", CAny::createBool(defval));
	else
		property->setDefault(CAny::createBool(defval));
}

void GtkWidgetView::addAppPaintableProperty() {
	Property * property= findProperty("app-paintable");
	if(!property)
		addInertProperty("app-paintable", prScalar, "bool", CAny::createBool(false));
}

void GtkWidgetView::addEventProperties() {
	Property * prop1= findProperty("events");
	if(!prop1)
		addInertProperty("events", prScalar, "GdkEventMask", CAny::createFlags("GdkEventMask", 0));
	Property * prop2= findProperty("extension-events");
	if(!prop2)
		addInertProperty("extension-events", prScalar, "GdkExtensionMode", CAny::createEnum("GdkExtensionMode", Gdk::EXTENSION_EVENTS_NONE));
}

Point GtkWidgetView::transform(const Point & p, PGtkWidget widget) {
	return Transform(p, GetPtr(getObject()), GetPtr(widget));
}

Rectangle GtkWidgetView::getWidgetRect(PGtkWidget widget) {
	int width= widget->get_allocation().get_width();
	int height= widget->get_allocation().get_height();
	Point p= transform(Point(0, 0), widget);
	return Rectangle(-p.x, -p.y, width, height);
}

// GtkContainerViewBase --------------------------------------------------------

GtkContainerViewBase::GtkContainerViewBase() {
	addProperty("border-width", prScalar, "unsigned", CAny::createUnsigned(0));
}

// GtkContainerView ------------------------------------------------------------

GtkContainerView::GtkContainerView() {
	childType= InvalidTypeId;
	asContainer= false;
	addProperty("as-container", prScalar, "bool", CAny::createBool(false)).
		setValueGet(sigc::mem_fun(*this, &GtkContainerView::getAsContainerProperty)).
		setValueSet(sigc::mem_fun(*this, &GtkContainerView::setAsContainerProperty));
	addInertProperty("children", prVector, "CrowChild").
		setFlag(sfNoEditing | sfReadOnly | sfHidden | sfOverwrite).
		setValueSet(sigc::mem_fun(*this, &GtkContainerView::setChildren));
	addInertProperty("focus-chain", prVector, "GtkWidget").
		setFlag(sfLinkOnly | sfWeakLink | sfReadOnly).
		setElementInsert(sigc::mem_fun(*static_cast<GlibObjectView *>(this), &GlibObjectView::insertObject));
	addInertProperty("focus-child", prEntity, "GtkWidget").
		setFlag(sfLinkOnly | sfWeakLink | sfReadOnly);
}

bool GtkContainerView::findFreePlace(const Point & p, FreePlace * fplace) {
	if(isAsContainer()) {
		Container * container= getContainer();
		Children children= findChildren(p, 0);
		PChild child= children.empty() ? PChild() : children.front();
		if(child && container->isPlaceholder(child->widget)) {
			fplace->place= child->place;
			fplace->area= getWidgetRect(child->widget);
			return true;
		}
	}
	return false;
}

FreePlaces GtkContainerView::findFreePlaces(const Point & p, const Nodes & nodes, const Point & p0) {
	FreePlaces fplaces;
	if(isAsContainer()) {
		Container * container= getContainer();
		Children children= findChildren(p, nodes.size()-1);
		if(!children.empty() && container->isPlaceholder(children.front()->widget)) {
			fplaces.resize(nodes.size());
			for(int i=0; i<(int)children.size(); ++i) {
				fplaces[i].place= children[i]->place;
				if(isTopChild(children[i]))
					fplaces[i].area= getWidgetRect(children[i]->widget);
			}
		}
	}
	return fplaces;
}

PNode GtkContainerView::createChild(const Place & place, PNode widget) {
	Check(isAsContainer());
	Model * model= getModel();
	PNode children= model->find(getModelNode(), "children");
	PNode child= model->vectorCreateEntity(children, childType);
	PNode link= model->createLink(child, "widget", sfNone);
	model->setChildNode(widget, child);
	return child;
}

PGtkContainer GtkContainerView::getObject() {
	return PGtkContainer::cast_static(GlibObjectView::getObject());
}

void GtkContainerView::initInstance() {
	GtkContainerViewBase::initInstance();
	if(!isRoot())
		findProperty("as-container")->setFlag(sfReadOnly);
}

void GtkContainerView::setChildType(const std::string & tname) {
	childType= FindType(tname);
}

bool GtkContainerView::placeholdersSet() {
	return isAsContainer();
}

void GtkContainerView::setAsContainerProperty(Property * property, Any value) {
	asContainer= value->getBool();
	setupAsContainer();
}

Any GtkContainerView::getAsContainerProperty(Property * property) {
	return CAny::createBool(asContainer);
}

void GtkContainerView::setupAsContainer(bool value) {
	setAsContainer(value);
	setupAsContainer();
}

void GtkContainerView::setupAsContainer() {
	Check(isRoot());
	setupContainer();
	if(isAsContainer()) {
		findProperty("children")->clearFlag(sfReadOnly);
		findProperty("focus-chain")->clearFlag(sfReadOnly);
		findProperty("focus-child")->clearFlag(sfReadOnly);
	} else {
		clearVectorProperty("children")->setFlag(sfReadOnly);
		clearVectorProperty("focus-chain")->setFlag(sfReadOnly);
		findProperty("focus-child")->
			setFlag(sfReadOnly).
			setInert(CAny::createObject());
	}
	touch("focus-chain");
	touch("focus-child");
	onAsContainerSetup();
}

PGtkWidget GtkContainerView::getWidget(PNode node) {
	return getController()->getView(node)->getValue()->getObject<Gtk::Widget>();
}

Children GtkContainerView::findChildren(const Point & p, int after) {

	Container * container= getContainer();
	Children children= container->get(true);
	Children result;
	result.reserve(1 + after);

	for(int i=0; i<(int)children.size(); ++i) {

		Rectangle rect= children[i]->widget->get_allocation();
		rect.x= rect.y= 0;

		if(isTopChild(children[i]) && transform(p, children[i]->widget).inside(rect)) {

			result.push_back(children[i]);
			++i;

			if(after > 0)
				for(int j=i; j<i+(int)children.size()-1; ++j) {
					PChild child= children[j%children.size()];
					if(container->isPlaceholder(child->widget)) {
						result.push_back(child);
						if((int)result.size() == 1 + after)
							break;
					}
				}

			if((int)result.size() == 1 + after)
				return result;
			else
				return Children();

		}

	}

	return Children();
}

Children GtkContainerView::makeChildren(Any value) {
	const AnyVector & vec= *value->getVector();
	Children children;
	children.reserve(vec.size());
	for(int i=0; i<(int)vec.size(); ++i) {
		PChild child= PChild::cast_dynamic(vec[i]->getObject());
		children.push_back(child);
	}
	return children;
}

Children GtkContainerView::getChildren() {
	Any value= findProperty("children")->getInert();
	return value ? makeChildren(value) : Children();
}

// ChildView -------------------------------------------------------------------

ChildView::ChildView() {
	addProperty("widget", prEntity, "GtkWidget").
		setValueGet(sigc::mem_fun(*this, &ChildView::getWidget)).
		setValueSet(sigc::mem_fun(*this, &ChildView::setWidget)).
		setFlag(sfLinkOnly | sfNoEditing | sfChildDef);
}

PChild ChildView::getObject() {
	return PChild::cast_static(GlibObjectView::getObject());
}

Any ChildView::getWidget(Property * property) {
	return CAny::createObject(getObject()->widget);
}

void ChildView::setWidget(Property * property, Any value) {
	getObject()->widget= PGtkWidget::cast_dynamic(value->getObject());
}

Property & ChildView::addIndexProperty() {
	return addProperty("index", prScalar, "int").
		setFlag(sfNoEditing | sfChildDef).
		setValueGet(sigc::mem_fun(*this, &ChildView::getIndex)).
		setValueSet(sigc::mem_fun(*this, &ChildView::setIndex));
}

Any ChildView::getIndex(Property * property) {
	return CAny::createInt(getObject()->place.a);
}

void ChildView::setIndex(Property * property, Any value) {
	getObject()->place.a= value->getInt();
}

}
