/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "application.h"

namespace CrowDesignerApp {

// -----------------------------------------------------------------------------

void CheckFailed(const char * expr, const char * file, unsigned line) {

	#ifdef CROW_SANDBOX
		assert(0);
	#endif

	std::ostringstream oss;
	oss << "Check failed: " << expr << ", file: " << file << ", line: " << line;
	FatalError(oss.str());
}

void FatalError(const std::string & message) {
	g_critical(message.c_str());
	Gtk::MessageDialog dialog(
		"Fatal Error [" G_LOG_DOMAIN "]", false, 
		Gtk::MESSAGE_ERROR, Gtk::BUTTONS_CLOSE, true
	);
	dialog.set_secondary_text(message);
	dialog.set_keep_above(true);
	dialog.run();
	abort();
}

std::vector<std::string> Arguments(int argc, char ** argv) {
	std::vector<std::string> arguments;
	arguments.reserve(argc);
	for(int i=0; i<argc; ++i)
		arguments.push_back(argv[i]);
	return arguments;
}

std::vector<Glib::ustring> Split(const Glib::ustring & str, gunichar ch) {
	std::vector<Glib::ustring> result;
	gchar ** vec= g_strsplit(str.c_str(), Glib::ustring(1, ch).c_str(), 0);
	int len= g_strv_length(vec);
	result.reserve(len);
	for(int i=0; i<len; ++i)
		result.push_back(vec[i]);
	g_strfreev(vec);
	return result;
}

GHashTable * SettingsToHashTable(const Settings & settings) {
	GHashTable * table= g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	for(Settings::const_iterator it= settings.begin(); it!=settings.end(); ++it) {
		gchar * key= g_strdup(it->first.c_str());
		gchar * value= g_strdup(it->second.c_str());
		g_hash_table_insert(table, key, value);
	}
	return table;
}

Settings HashTableToSettings(GHashTable * table) {
	Settings result;
	struct tmp { static void add(gpointer key, gpointer value, gpointer user_data) {
		(*static_cast<Settings *>(user_data))[static_cast<gchar *>(key)]= static_cast<gchar *>(value);
	} };
	g_hash_table_foreach(table, &tmp::add, &result);
	return result;
}

int GetKeyInt(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	Check(it!=settings.end());
	std::istringstream iss(it->second);
	int value;
	iss >> value;
	Check(!iss.fail());
	return value;
}

void SetKeyInt(Settings * settings, const std::string & key, int value) {
	std::ostringstream oss;
	oss << value;
	(*settings)[key]= oss.str();
}

Glib::ustring GetKeyString(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	Check(it!=settings.end());
	return it->second;
}

void SetKeyString(Settings * settings, const std::string & key, const char * const value) {
	(*settings)[key]= value;
}

void EnsureKeyString(Settings * settings, const std::string & key, const char * const value) {
	Settings::const_iterator it= settings->find(key);
	if(it==settings->end())
		(*settings)[key]= value;
}

bool GetKeyBool(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	Check(it!=settings.end());
	return it->second == "true" ? true : false;
}

void SetKeyBool(Settings * settings, const std::string & key, bool value) {
	(*settings)[key]= value ? "true" : "false";
}

Gdk::Color GetKeyColor(const Settings & settings, const std::string & key) {
	Settings::const_iterator it= settings.find(key);
	Check(it!=settings.end());
	return Gdk::Color(it->second);
}

void SetKeyColor(Settings * settings, const std::string & key, const Gdk::Color & value) {
	unsigned pixel=
		((value.get_red() >> 8) << 16) |
		((value.get_green() >> 8) << 8) |
		(value.get_blue() >> 8);
	std::ostringstream oss;
	oss << "#";
	oss << std::setw(6) << std::setfill('0');
	oss << std::hex << std::uppercase;
	oss << pixel;
	(*settings)[key]= oss.str();
}

// ErrorDialog -----------------------------------------------------------------

ErrorDialog::ErrorDialog() {
	set_title("GTK+ Messages");
	set_default_size(400, 200);
	set_type_hint(Gdk::WINDOW_TYPE_HINT_DIALOG);
	set_position(Gtk::WIN_POS_CENTER_ON_PARENT);

	scroll.set_policy(Gtk::POLICY_NEVER, Gtk::POLICY_ALWAYS);
	add(scroll);
	scroll.show();

	textView.set_editable(false);
	textView.set_cursor_visible(false);
	textView.set_wrap_mode(Gtk::WRAP_WORD);
	scroll.add(textView);
	textView.show();
}

void ErrorDialog::onErrorMessage(const Glib::ustring & message) {
	if(!is_visible())
		textView.get_buffer()->set_text("");
	textView.get_buffer()->insert(textView.get_buffer()->end(), message);
	textView.get_buffer()->place_cursor(textView.get_buffer()->end());
	Gtk::TextBuffer::iterator it= textView.get_buffer()->end();
	textView.scroll_to(it);
	show();
}

// SettingsDialog --------------------------------------------------------------

SettingsDialog::SettingsDialog() {

	SetKeyColor(&defaults, "$wce_vc", Gdk::Color("forest green"));
	SetKeyColor(&defaults, "$wce_wc", Gdk::Color("indian red"));
	SetKeyBool(&defaults, "$show_dw", false);
	SetKeyBool(&defaults, "$show_hp", false);
	SetKeyBool(&defaults, "$supp_msg", false);

	loader= Gui::Loader::create();

	GUIXML_DECL
		#include "settings.gui.h"
	GUIXML_LOAD_CPP(loader)
	
	loader->bind_object("window", &window);
	loader->bind_object("widget", &widget);
	loader->bind_object("validColor", &validColor);
	loader->bind_object("wrongColor", &wrongColor);
	loader->bind_object("showDebugWidgets", &showDebugWidgets);
	loader->bind_object("showHiddenProperties", &showHiddenProperties);
	loader->bind_object("suppressMessages", &suppressMessages);
	loader->bind_object("actions", &actions);
	loader->bind_object("folders", &folders);
	loader->bind_object("defaultsButton", &defaultsButton);

	loader->connect_signals("onAdd", sigc::mem_fun(*this, &SettingsDialog::onAdd));
	loader->connect_signals("onRemove", sigc::mem_fun(*this, &SettingsDialog::onRemove));
	loader->connect_signals("onMoveUp", sigc::mem_fun(*this, &SettingsDialog::onMoveUp));
	loader->connect_signals("onMoveDown", sigc::mem_fun(*this, &SettingsDialog::onMoveDown));
	loader->connect_signals("onColorChanged", sigc::mem_fun(*this, &SettingsDialog::onColorChanged), 2);
	loader->connect_signals("onCheckChanged", sigc::mem_fun(*this, &SettingsDialog::onCheckChanged), 3);
	loader->connect_signals("onSwitchPage", sigc::mem_fun(*this, &SettingsDialog::onSwitchPage));
	loader->connect_signals("onDefaults", sigc::mem_fun(*this, &SettingsDialog::onDefaults));

	Gtk::TreeModelColumn<Glib::ustring> column;
	Gtk::TreeModel::ColumnRecord columns;
	columns.add(column);
	model= Gtk::ListStore::create(columns);

	folders->set_model(model);
	folders->append_column("", column);

	folders->get_selection()->signal_changed().connect(sigc::mem_fun(*this, &SettingsDialog::onSelectionChanged));

}

void SettingsDialog::setSettings(const Settings & settings) {
	original= settings;
	original.insert(defaults.begin(), defaults.end());
	original["$icon_folders"];
	updateUi(original);
}

Settings SettingsDialog::getSettings() {
	Settings current= original;

	SetKeyColor(&current, "$wce_vc", validColor->get_color());
	SetKeyColor(&current, "$wce_wc", wrongColor->get_color());
	SetKeyBool(&current, "$show_dw", showDebugWidgets->get_active());
	SetKeyBool(&current, "$show_hp", showHiddenProperties->get_active());
	SetKeyBool(&current, "$supp_msg", suppressMessages->get_active());

	Glib::ustring vec;
	Gtk::TreeModel::Children rows= model->children();
	for(Gtk::TreeModel::Children::iterator it= rows.begin(); it!=rows.end(); ++it) {
		Glib::ustring str;
		it->get_value(0, str);
		vec += str + '*';
	}
	if(!vec.empty())
		vec.resize(vec.size()-1);
	SetKeyString(&current, "$icon_folders", vec.c_str());

	return current;
}

void SettingsDialog::onAdd() {
	Gtk::FileChooserDialog dialog(*window, "File Chooser", Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER);
	dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);
	dialog.set_current_folder(Glib::get_home_dir());
	if(dialog.run() == Gtk::RESPONSE_OK) {
		Glib::ustring folder= Glib::filename_to_utf8(dialog.get_filename());
		Gtk::TreeRow row= *model->insert_after(getPosition());
		row.set_value(0, folder);
		folders->get_selection()->select(row);
		actionCompleted();
	}
}

void SettingsDialog::onRemove() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter sel= it;
	++sel;
	model->erase(it);
	if(!sel && !model->children().empty())
		--sel;
	if(sel)
		folders->get_selection()->select(sel);
	actionCompleted();
}

void SettingsDialog::onMoveUp() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter dest= it;
	--dest;
	model->move(it, dest);
	actionCompleted();
	updateActions();
}

void SettingsDialog::onMoveDown() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter dest= it;
	++dest;
	++dest;
	model->move(it, dest);
	actionCompleted();
	updateActions();
}

void SettingsDialog::onColorChanged() {
	updateButtons();
}

void SettingsDialog::onCheckChanged() {
	updateButtons();
}

void SettingsDialog::onSelectionChanged() {
	updateActions();
}

void SettingsDialog::onSwitchPage(GtkNotebookPage *, guint p) {
	updateButtons();
}

void SettingsDialog::onDefaults() {
	Settings settings= getSettings();
	for(Settings::iterator it= defaults.begin(); it!=defaults.end(); ++it)
		settings[it->first]= it->second;
	updateUi(settings);
}

void SettingsDialog::updateUi(const Settings & settings) {
	validColor->set_color(GetKeyColor(settings, "$wce_vc"));
	wrongColor->set_color(GetKeyColor(settings, "$wce_wc"));
	showDebugWidgets->set_active(GetKeyBool(settings, "$show_dw"));
	showHiddenProperties->set_active(GetKeyBool(settings, "$show_hp"));
	suppressMessages->set_active(GetKeyBool(settings, "$supp_msg"));

	std::vector<Glib::ustring> vec= Split(GetKeyString(settings, "$icon_folders"), '*');
	model->clear();
	for(std::vector<Glib::ustring>::iterator it=vec.begin(); it!=vec.end(); ++it) {
		Gtk::TreeModel::iterator iter= model->append();
		iter->set_value(0, *it);
	}

	updateButtons();
	updateActions();
}

void SettingsDialog::updateButtons() {
	Settings current= getSettings();
	window->set_response_sensitive(Gtk::RESPONSE_OK, original != current);
	bool canDefault= false;
	if(widget->get_current_page()==0)
		for(Settings::iterator it= defaults.begin(); it!=defaults.end(); ++it)
			if(current[it->first]!=it->second) {
				canDefault= true;
				break;
			}
	defaultsButton->set_sensitive(canDefault);
}

void SettingsDialog::updateActions() {
	Gtk::TreeIter it= getPosition();
	Gtk::TreeIter next;
	if(it) {
		next= it;
		++next;
	}
	actions->get_action("Add")->set_sensitive(true);
	actions->get_action("Remove")->set_sensitive(it);
	actions->get_action("MoveUp")->set_sensitive(it && it!=model->children().begin());
	actions->get_action("MoveDown")->set_sensitive(it && next!=model->children().end());
}

Gtk::TreeIter SettingsDialog::getPosition() {
	if(folders->get_selection()->count_selected_rows() == 0)
		return model->children().end();
	else
		return folders->get_selection()->get_selected();
}

void SettingsDialog::actionCompleted() {
	if(folders->get_selection()->count_selected_rows() != 0) {
		Gtk::TreeIter it= folders->get_selection()->get_selected();
		folders->scroll_to_row(model->get_path(it));
	}
	updateButtons();
}

// AppWindow -------------------------------------------------------------------

AppWindow::State AppWindow::getState() {
	State state;
	get_size(state.width, state.height);
	state.maximized= getMaximized();
	return state;
}

void AppWindow::setState(const State & state) {
	deiconify();
	if(state.maximized==1) {
		unfullscreen();
		maximize();
	} else if(state.maximized==2) {
		unmaximize();
		fullscreen();
	} else {
		unfullscreen();
		unmaximize();
	}
	resize(state.width, state.height);
}

int AppWindow::getMaximized() {
	Gdk::WindowState state= getGdkWindow()->get_state();
	if(state & Gdk::WINDOW_STATE_MAXIMIZED)
		return 1;
	else if(state & Gdk::WINDOW_STATE_FULLSCREEN)
		return 2;
	else
		return 0;
}

Glib::RefPtr<Gdk::Window> AppWindow::getGdkWindow() {
	Glib::RefPtr<Gdk::Window> window= get_window();
	Check(window);
	return window;
}

// Application -----------------------------------------------------------------

Application::Application(const std::vector<std::string> & args1, const std::vector<std::string> & args2) {

	{
		#include "icons.h"
		Gtk::IconTheme::add_builtin_icon("crow-icon", 16, Gdk::Pixbuf::create_from_inline(sizeof(crow16), crow16));
		Gtk::IconTheme::add_builtin_icon("crow-icon", 32, Gdk::Pixbuf::create_from_inline(sizeof(crow32), crow32));
		Gtk::IconTheme::add_builtin_icon("crow-icon", 48, Gdk::Pixbuf::create_from_inline(sizeof(crow48), crow48));
		gtk_window_set_icon_name(window.gobj(), "crow-icon");
	}

	defaultIconPath= currentIconPath();

	errorDialog.set_transient_for(window);

	setupHandler("GLib");
	setupHandler("GLib-GObject");
	setupHandler("Gdk");
	setupHandler("GdkPixbuf");
	setupHandler("Pango");
	setupHandler("Gtk");
	setupHandler("GuiLoader");
	setupHandler("glibmm");
	setupHandler("gdkmm");
	setupHandler("pangomm");
	setupHandler("gtkmm");

	Glib::add_exception_handler(sigc::mem_fun(*this, &Application::onException));

	startupArgs1= args1;
	startupArgs2= args2;
	startupDirectory= Glib::get_current_dir();

	loadSettings();
	AppWindow::State state;
	state.width= GetKeyInt(settings, "window_width");
	state.height= GetKeyInt(settings, "window_height");
	state.maximized= GetKeyInt(settings, "window_maximized");
	window.setState(state);
	window.signal_delete_event().connect(sigc::mem_fun(*this, &Application::onDeleteEvent));

	createActionGroup();
	createUIManager();

	window.add(vBox);
	vBox.show();

	Gtk::MenuBar * menuBar= dynamic_cast<Gtk::MenuBar *>(uiManager->get_widget("/MainMenubar"));
	vBox.pack_start(*menuBar, Gtk::PACK_SHRINK);

	Gtk::Toolbar * toolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/MainToolbar"));
	toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	vBox.pack_start(*toolbar, Gtk::PACK_SHRINK);

	statusbar.push("");
	vBox.pack_end(statusbar, Gtk::PACK_SHRINK);
	statusbar.show();
	
	statusbar.pack_end(infoFrame, Gtk::PACK_SHRINK);

	infoLabel.set_text("");
	infoFrame.add(infoLabel);
	infoLabel.show();

	designer= crow_designer_new();
	g_signal_connect(designer, "actions-updated", G_CALLBACK(&Application::on_actions_updated), this);
	g_signal_connect(designer, "status-updated", G_CALLBACK(&Application::on_status_updated), this);
	g_signal_connect(designer, "info-updated", G_CALLBACK(&Application::on_info_updated), this);
	g_signal_connect(designer, "show-url", G_CALLBACK(&Application::on_show_url), this);

	Gtk::Widget * dwidget= designer_get_widget();
	vBox.pack_start(*dwidget, Gtk::PACK_EXPAND_WIDGET);

	Gtk::MenuItem * recent= dynamic_cast<Gtk::MenuItem *>(*actionGroup->get_action("Recent")->get_proxies().begin());
	Gtk::RecentFilter * filter= Gtk::manage(new Gtk::RecentFilter);
	filter->add_pattern("*.gui");
	filter->add_pattern("*.gui.h");
	Gtk::RecentChooserMenu * chooser= Gtk::manage(new Gtk::RecentChooserMenu);
	chooser->set_sort_type(Gtk::RECENT_SORT_MRU);
	chooser->set_filter(*filter);
	chooser->set_show_icons(false);
	chooser->set_show_numbers(true);
	chooser->signal_item_activated().connect(sigc::bind<Gtk::RecentChooser *>(
		sigc::mem_fun(*this, &Application::onOpenRecent), chooser
	));
	recent->set_submenu(*chooser);
	chooser->show();

	designer_apply_settings(settings);
	window.add_accel_group(uiManager->get_accel_group());

	updateTitle();

	if(args2.size()==2)
		openFile(File(args2[1]));

}

Application::~Application() {
	designer_store_settings(&settings);
	saveSettings(true);
	g_object_unref(designer);
	for(Hids::iterator it=hids.begin(); it!=hids.end(); ++it)
		g_log_remove_handler(it->first, it->second);
}

void Application::setupHandler(const char * domain) {
	int hid= g_log_set_handler(domain,
		GLogLevelFlags(
			G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING |
			G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION
		),
		&Application::logHandler, this
	);
	hids.push_back(std::make_pair(domain, hid));
}

void Application::logHandler(
	const gchar * logDomain,
	GLogLevelFlags logLevel,
	const gchar * message,
	gpointer userData
) {
	static_cast<Application *>(userData)->onErrorMessage(logDomain, logLevel, message);
}

void Application::onErrorMessage(
	const Glib::ustring & domain,
	GLogLevelFlags level,
	const Glib::ustring & message
) {
	Glib::ustring errstr= "[" + domain;
	if(level & G_LOG_LEVEL_CRITICAL)
		errstr += " CRITICAL";
	if(level & G_LOG_LEVEL_WARNING)
		errstr += " WARNING";
	errstr += "] " + message + "\n\n";
	std::cerr << errstr << std::flush;
	if(!GetKeyBool(settings, "$supp_msg"))
		errorDialog.onErrorMessage(errstr);
}

void Application::createActionGroup() {

	Glib::RefPtr<Gtk::ActionGroup> actions= Gtk::ActionGroup::create();

	actions->add(Gtk::Action::create("Save", Gtk::Stock::SAVE, "_Save", "Save the current file"), sigc::mem_fun(*this, &Application::onSave));

	actions->add(Gtk::Action::create("Undo", Gtk::Stock::UNDO, "_Undo", "Undo the last action"), Gtk::AccelKey("<control>Z"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_UNDO));
	actions->add(Gtk::Action::create("Redo", Gtk::Stock::REDO, "_Redo", "Redo the undone action"), Gtk::AccelKey("<shift><control>Z"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_REDO));
	actions->add(Gtk::Action::create("Cut", Gtk::Stock::CUT, "Cu_t", "Cut the selection"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_CUT));
	actions->add(Gtk::Action::create("Copy", Gtk::Stock::COPY, "_Copy", "Copy the selection"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_COPY));
	actions->add(Gtk::Action::create("Paste", Gtk::Stock::PASTE, "_Paste...", "Paste the clipboard"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PASTE));
	actions->add(Gtk::Action::create("PasteToRoot", "Paste to _Root"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PASTE_TO_ROOT));
	actions->add(Gtk::Action::create("Delete", Gtk::Stock::DELETE, "_Delete", "Delete the selection"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DELETE));
	
	actions->add(Gtk::Action::create("About", Gtk::Stock::INFO, "_About"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_ABOUT));
	
	actions->add(Gtk::Action::create("Preview", Gtk::Stock::EXECUTE, "_Preview...", "Show widgets in a preview window"), Gtk::AccelKey("<control>T"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_PREVIEW));
	actions->add(Gtk::Action::create("DefaultSize", Gtk::Stock::ZOOM_100, "D_efault Size", "Restore default widget size"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DEFAULT_SIZE));
	actions->add(Gtk::Action::create("ShiftLeft", Gtk::Stock::GO_BACK, "Shift _Left", "Shift the widget left"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_SHIFT_LEFT));
	actions->add(Gtk::Action::create("ShiftRight", Gtk::Stock::GO_FORWARD, "Shift _Right", "Shift the widget right"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_SHIFT_RIGHT));
	actions->add(Gtk::Action::create("Translatables", Gtk::Stock::EDIT, "_Strings...", "Manage translatable strings"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_TRANSLATABLES));
	
	actions->add(Gtk::Action::create("Default", Gtk::Stock::CLEAR, "_Default", "Set the property to its default value"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_DEFAULT));
	actions->add(Gtk::Action::create("Add", Gtk::Stock::ADD, "_Add", "Add a New Element"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_ADD));
	actions->add(Gtk::Action::create("Remove", Gtk::Stock::REMOVE, "Re_move", "Remove the element"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_REMOVE));
	actions->add(Gtk::Action::create("MoveUp", Gtk::Stock::GO_UP, "Move _Up", "Move the element up"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_MOVE_UP));
	actions->add(Gtk::Action::create("MoveDown", Gtk::Stock::GO_DOWN, "Move Do_wn", "Move the element down"), sigc::bind<CrowActions>(sigc::mem_fun(*this, &Application::designer_run_action), CROW_ACTION_MOVE_DOWN));

	{
		std::vector<Glib::RefPtr<Gtk::Action> > V= actions->get_actions();
		for(int i=0; i<(int)V.size(); ++i)
			V[i]->set_sensitive(false);
	}

	actions->add(Gtk::Action::create("File", "_File"));
	actions->add(Gtk::Action::create("Clear", Gtk::Stock::NEW, "_New", "Start a new file"), sigc::mem_fun(*this, &Application::onClear));
	actions->add(Gtk::Action::create("Clone", "New _Window"), sigc::mem_fun(*this, &Application::onClone));
	actions->add(Gtk::Action::create("Open", Gtk::Stock::OPEN, "_Open", "Open a file"), sigc::mem_fun(*this, &Application::onOpen));
	actions->add(Gtk::Action::create("Recent", "Open _Recent"));
	actions->add(Gtk::Action::create("SaveAs", "Save _As..."), sigc::mem_fun(*this, &Application::onSaveAs));
	actions->add(Gtk::Action::create("Quit", Gtk::Stock::QUIT, "_Quit"), sigc::mem_fun(*this, &Application::onQuit));

	actions->add(Gtk::Action::create("Edit", "_Edit"));
	actions->add(Gtk::Action::create("Settings", Gtk::Stock::PREFERENCES, "_Settings..."), sigc::mem_fun(*this, &Application::onSettings));
	actions->add(Gtk::Action::create("Help", "_Help"));
	actions->add(Gtk::Action::create("Home", Gtk::Stock::HOME, "_Home"), Gtk::AccelKey("<control>H"), sigc::mem_fun(*this, &Application::onHome));
	actions->add(Gtk::Action::create("API", Gtk::Stock::HELP, "API _Reference"), Gtk::AccelKey("F1"), sigc::mem_fun(*this, &Application::onApi));
	actions->add(Gtk::Action::create("Actions", "_Actions"));

	actionGroup= actions;
}

void Application::createUIManager() {

	Glib::RefPtr<Gtk::UIManager> manager= Gtk::UIManager::create();
	manager->insert_action_group(actionGroup);

	Glib::ustring uistr=
		"<ui>"
		"  <menubar name='MainMenubar'>"
		"    <menu action='File'>"
		"      <menuitem action='Clear'/>"
		"      <menuitem action='Clone'/>"
		"      <separator/>"
		"      <menuitem action='Open'/>"
		"      <menuitem action='Recent'/>"
		"      <separator/>"
		"      <menuitem action='Save'/>"
		"      <menuitem action='SaveAs'/>"
		"      <separator/>"
		"      <menuitem action='Quit'/>"
		"    </menu>"
		"    <menu action='Edit'>"
		"      <menuitem action='Undo'/>"
		"      <menuitem action='Redo'/>"
		"      <separator/>"
		"      <menuitem action='Cut'/>"
		"      <menuitem action='Copy'/>"
		"      <menuitem action='Paste'/>"
		"      <menuitem action='PasteToRoot'/>"
		"      <menuitem action='Delete'/>"
		"      <separator/>"
		"      <menuitem action='Settings'/>"
		"    </menu>"
		"    <menu action='Actions'>"
		"      <menuitem action='Preview'/>"
		"      <menuitem action='DefaultSize'/>"
		"      <menuitem action='ShiftLeft'/>"
		"      <menuitem action='ShiftRight'/>"
		"      <menuitem action='Translatables'/>"
		"      <separator/>"
		"      <menuitem action='Default'/>"
		"      <menuitem action='Add'/>"
		"      <menuitem action='Remove'/>"
		"      <menuitem action='MoveUp'/>"
		"      <menuitem action='MoveDown'/>"
		"    </menu>"
		"    <menu action='Help'>"
		"      <menuitem action='Home'/>"
		"      <menuitem action='API'/>"
		"      <separator/>"
		"      <menuitem action='About'/>"
		"    </menu>"
		"  </menubar>"
		"  <toolbar  name='MainToolbar'>"
		"    <toolitem action='Clear'/>"
		"    <toolitem action='Open'/>"
		"    <toolitem action='Save'/>"
		"    <separator/>"
		"    <toolitem action='Cut'/>"
		"    <toolitem action='Copy'/>"
		"    <toolitem action='Paste'/>"
		"    <toolitem action='Delete'/>"
		"    <separator/>"
		"    <toolitem action='Undo'/>"
		"    <toolitem action='Redo'/>"
		"    <separator/>"
		"    <toolitem action='Preview'/>"
		"    <toolitem action='DefaultSize'/>"
		"    <toolitem action='ShiftLeft'/>"
		"    <toolitem action='ShiftRight'/>"
		"    <toolitem action='Translatables'/>"
		"  </toolbar>"
		"  <toolbar name='ExplorerToolbar'>"
		"    <toolitem action='Default'/>"
		"    <separator/>"
		"    <toolitem action='Add'/>"
		"    <toolitem action='Remove'/>"
		"    <toolitem action='MoveUp'/>"
		"    <toolitem action='MoveDown'/>"
		"  </toolbar>"
		"  <popup name='MainPopup'>"
		"    <menuitem action='Cut'/>"
		"    <menuitem action='Copy'/>"
		"    <menuitem action='Delete'/>"
		"    <separator/>"
		"    <menuitem action='DefaultSize'/>"
		"    <menuitem action='ShiftLeft'/>"
		"    <menuitem action='ShiftRight'/>"
		"  </popup>"
		"</ui>"
	;

	try {
		manager->add_ui_from_string(uistr);
	} catch(const Glib::Error & x) {
		FatalError("Creating UIManager failed: " + x.what());
	}

	uiManager= manager;
}

void Application::setupGtkAction(CrowActions actions, CrowActions changed, CrowActions ca, const char * str) {
	if(changed & ca)
		actionGroup->get_action(str)->set_sensitive(actions & ca);
} 

void Application::on_actions_updated(CrowDesigner * obj, CrowActions changed, Application * app) {
	app->onActionsUpdated(changed);
}

void Application::on_status_updated(CrowDesigner * obj, const gchar * str, Application * app) {
	app->onStatusUpdated(str);
}

void Application::on_info_updated(CrowDesigner * obj, const gchar * str, Application * app) {
	app->onInfoUpdated(str);
}

void Application::on_show_url(CrowDesigner * obj, const gchar * str, Application * app) {
	app->onShowUrl(str);
}

void Application::onActionsUpdated(CrowActions changed) {
	CrowActions actions= designer_get_enabled_actions();
	if(changed & CROW_ACTIONS_SAVE)
		setupGtkAction(actions, changed, CROW_ACTION_SAVE, "Save");
	if(changed & CROW_ACTIONS_BASIC) {
		setupGtkAction(actions, changed, CROW_ACTION_UNDO, "Undo");
		setupGtkAction(actions, changed, CROW_ACTION_REDO, "Redo");
		setupGtkAction(actions, changed, CROW_ACTION_CUT, "Cut");
		setupGtkAction(actions, changed, CROW_ACTION_COPY, "Copy");
		setupGtkAction(actions, changed, CROW_ACTION_PASTE, "Paste");
		setupGtkAction(actions, changed, CROW_ACTION_PASTE_TO_ROOT, "PasteToRoot");
		setupGtkAction(actions, changed, CROW_ACTION_DELETE, "Delete");
		setupGtkAction(actions, changed, CROW_ACTION_ABOUT, "About");
		setupGtkAction(actions, changed, CROW_ACTION_PREVIEW, "Preview");
		setupGtkAction(actions, changed, CROW_ACTION_DEFAULT_SIZE, "DefaultSize");
		setupGtkAction(actions, changed, CROW_ACTION_SHIFT_LEFT, "ShiftLeft");
		setupGtkAction(actions, changed, CROW_ACTION_SHIFT_RIGHT, "ShiftRight");
		setupGtkAction(actions, changed, CROW_ACTION_TRANSLATABLES, "Translatables");
	}
	if(changed & CROW_ACTIONS_EXPLORER) {
		setupGtkAction(actions, changed, CROW_ACTION_DEFAULT, "Default");
		setupGtkAction(actions, changed, CROW_ACTION_ADD, "Add");
		setupGtkAction(actions, changed, CROW_ACTION_REMOVE, "Remove");
		setupGtkAction(actions, changed, CROW_ACTION_MOVE_UP, "MoveUp");
		setupGtkAction(actions, changed, CROW_ACTION_MOVE_DOWN, "MoveDown");
	}
}

void Application::onStatusUpdated(const Glib::ustring & str) {
	statusbar.pop();
	statusbar.push(str);
}

void Application::onInfoUpdated(const Glib::ustring & str) {
	infoFrame.property_visible()= !str.empty();
	infoLabel.set_text(str);
}

void Application::onShowUrl(const Glib::ustring & str) {
	showUrl(str);
}

void Application::onClear() {
	if(askProceed()) {
		file.clear();
		designer_clear();
		updateTitle();
	}
}

void Application::onClone() {

	std::vector<std::string> args= startupArgs1;
	for(int i=1; i<(int)startupArgs2.size(); ++i) {
		const std::string & arg= startupArgs2[i];
		std::vector<std::string>::iterator it= std::find(args.begin(), args.end(), arg);
		if(it!=args.end())
			args.erase(it);
	}

	try {
		Glib::spawn_async(startupDirectory, args, Glib::SPAWN_SEARCH_PATH);
	} catch(const Glib::Error & x) {
		std::string str= "spawn_async(" + startupDirectory + ",";
		for(int i=0; i<(int)args.size(); ++i)
			str += " " + args[i];
		showMessage("'New Window' action failed: " + x.what() + "\n" + str + ")");
	}
}

void Application::onOpen() {
	if(askProceed())
		openFile(askFilename(Gtk::FILE_CHOOSER_ACTION_OPEN));
}

void Application::onOpenRecent(Gtk::RecentChooser * chooser) {
	if(askProceed()) {
		std::string uri= chooser->get_current_uri();
		if(uri.compare(0, 7, "file://") == 0)
			openFile(uri.substr(7));
		else
			showMessage("Invalid URI: " + uri);
	}
}

void Application::openFile(const File & filename) {
	File name= filename;
	if(!name.empty()) {
		Glib::ustring data;
		bool ok= loadFile(name, &data);
		if(ok) {
			try {
				designer_load(data);
				name.header= data[0]=='"';
				file= name;
				addRecentFile();
			} catch(const Glib::Exception & x) {
				file.clear();
				showMessage("Loading failed: " + x.what());
			}
			updateTitle();
		}
	}
}

void Application::onSave() {
	File name= file;
	if(name.empty())
		name= askFilename(Gtk::FILE_CHOOSER_ACTION_SAVE);
	if(!name.empty()) {
		Glib::ustring data= name.header ? designer_save_header() : designer_save();
		bool ok= saveFile(name, data);
		if(ok) {
			file= name;
			addRecentFile();
			designer_saved();
			updateTitle();
		}
	}
}

void Application::onSaveAs() {
	File name= askFilename(Gtk::FILE_CHOOSER_ACTION_SAVE);
	if(!name.empty()) {
		Glib::ustring data= name.header ? designer_save_header() : designer_save();
		bool ok= saveFile(name, data);
		if(ok) {
			file= name;
			addRecentFile();
			designer_saved();
			updateTitle();
		}
	}
}

void Application::onQuit() {
	if(askProceed())
		window.hide();
}

void Application::onSettings() {
	Glib::ustring path= settings["$icon_folders"];
	designer_store_settings(&settings);
	SettingsDialog dialog;
	dialog.getWindow()->set_transient_for(window);
	dialog.setSettings(settings);
	if(dialog.getWindow()->run() == Gtk::RESPONSE_OK) {
		settings= dialog.getSettings();
		if(settings["$icon_folders"]!=path)
			updateIconPath();
		if(GetKeyBool(settings, "$supp_msg"))
			errorDialog.hide();
		designer_apply_settings(settings);
		designer_store_settings(&settings);
		saveSettings();
	}
}

void Application::onHome() {
	showUrl("http://crow-designer.sourceforge.net");
}

void Application::onApi() {
	showUrl("http://developer.gnome.org/doc/API/2.0/gtk/index.html");
}

bool Application::onDeleteEvent(GdkEventAny *) {
	onQuit();
	return true;
}

void Application::onException() {
	Glib::ustring msg;
	try {
		throw;
	} catch(const Glib::Exception & x) {
		msg= x.what();
	} catch(...) {
		msg= "no message";
	}
	FatalError(msg);
}

bool Application::loadFile(const File & name, Glib::ustring * data) {
	gchar * contents;
	if(g_file_get_contents(name.name.c_str(), &contents, NULL, NULL)) {
		*data= contents;
		g_free(contents);
		return true;
	} else {
		showMessage("Loading failed");
		*data= "";
		return false;
	}
}

bool Application::saveFile(const File & name, const Glib::ustring & data) {
	if(g_file_set_contents(name.name.c_str(), data.c_str(), data.bytes(), NULL)) {
		return true;
	} else {
		showMessage("Saving failed");
		return false;
	}
}

Application::File Application::askFilename(Gtk::FileChooserAction action) {
	Gtk::FileChooserDialog dialog(window, "File Chooser", action);
	Gtk::CheckButton checkHeader("Save as a C header");

	dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	if(action==Gtk::FILE_CHOOSER_ACTION_SAVE) {
		checkHeader.set_active(file.header);
		dialog.set_extra_widget(checkHeader);
		dialog.add_button(Gtk::Stock::SAVE, Gtk::RESPONSE_OK);
		dialog.set_do_overwrite_confirmation(true);
		checkHeader.signal_toggled().connect(
			sigc::bind<Gtk::FileChooser *, Gtk::CheckButton *>(
				sigc::mem_fun(*this, &Application::onCheckHeader), &dialog, &checkHeader
			)
		);
	} else
		dialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	Gtk::FileFilter filter_gui;
	filter_gui.set_name("GuiXml files");
	filter_gui.add_pattern("*.gui");
	filter_gui.add_pattern("*.gui.h");
	dialog.add_filter(filter_gui);

	Gtk::FileFilter filter_any;
	filter_any.set_name("Any files");
	filter_any.add_pattern("*");
	dialog.add_filter(filter_any);

	if(!file.empty())
		dialog.set_filename(file.name);
	else
		dialog.set_current_folder(Glib::get_home_dir());

	if(dialog.run() == Gtk::RESPONSE_OK) {
		File F(dialog.get_filename(), checkHeader.get_active());
		if(action==Gtk::FILE_CHOOSER_ACTION_SAVE)
			if(Glib::path_get_basename(F.name).find('.') == std::string::npos)
				F.name += F.header ? ".gui.h" : ".gui";
		return F;
	} else
		return File();
}

void Application::onCheckHeader(Gtk::FileChooser * chooser, Gtk::CheckButton * button) {
	bool header= button->get_active();
	std::string name= Glib::path_get_basename(chooser->get_filename());
	if(header && g_str_has_suffix(name.c_str(), ".gui")) {
		name += ".h";
		chooser->set_current_name(name);
	} else if(!header && g_str_has_suffix(name.c_str(), ".gui.h")) {
		name.resize(name.size() - 2);
		chooser->set_current_name(name);
	}
}

bool Application::askProceed() {
	if(designer_get_enabled_actions() & CROW_ACTION_SHOULD_SAVE) {

		Gtk::Dialog dialog("Document is not saved", window);
		dialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_NONE);
		dialog.add_button(Gtk::Stock::YES, Gtk::RESPONSE_OK);

		Glib::ustring text= file.empty() ? "Abandon new document?" : "Abandon changes?";

		Gtk::Label * label= Gtk::manage(new Gtk::Label("\n" + text + "\n"));
		label->show();
		dialog.get_vbox()->add(*label);

		return dialog.run() == Gtk::RESPONSE_OK;

	} else
		return true;
}

void Application::updateTitle() {
	Glib::ustring name= file.empty() ? "" : " - " + Glib::filename_to_utf8(file.name);
	window.set_title("Crow" + name);
}

std::vector<Glib::ustring> Application::currentIconPath() {
	return Gtk::IconTheme::get_default()->get_search_path();
}

void Application::updateIconPath() {
	std::vector<Glib::ustring> custom= Split(settings["$icon_folders"], '*');
	std::vector<Glib::ustring> path;
	path.reserve(defaultIconPath.size() + custom.size());
	path.insert(path.end(), defaultIconPath.begin(), defaultIconPath.end());
	path.insert(path.end(), custom.begin(), custom.end());
	if(path != currentIconPath())
		Gtk::IconTheme::get_default()->set_search_path(path);
}

void Application::loadKeys(Settings * skeys) {

	#ifdef CROW_WIN32

		HKEY hKey;
		LONG result= RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\crow\\settings", 0, KEY_READ, &hKey);
		if(result==ERROR_SUCCESS) {
			for(DWORD index= 0; ; ++index) {
				CHAR name[100]; DWORD nameSize= sizeof(name);
				CHAR value[1000]; DWORD valueSize= sizeof(value);
				LONG result= RegEnumValue(hKey, index, name, &nameSize, NULL, NULL, reinterpret_cast<BYTE *>(value), &valueSize);
				if(result==ERROR_NO_MORE_ITEMS)
					break;
				if(result==ERROR_SUCCESS)
					(*skeys)[name]= value;
			}
			RegCloseKey(hKey);
		}

	#else

		Glib::ustring path= Glib::ustring(g_get_home_dir()) + "/.crow";
		GKeyFile * file= g_key_file_new();
		if(g_key_file_load_from_file(file, path.c_str(), G_KEY_FILE_NONE, NULL)) {
			gsize length;
			gchar ** keys= g_key_file_get_keys(file, "settings", &length, NULL);
			if(keys) {
				for(int i=0; i<(int)length; ++i) {
					char * str= g_key_file_get_string(file, "settings", keys[i], NULL);
					if(str) {
						(*skeys)[keys[i]]= str;
						g_free(str);
					} 
				}
				g_strfreev(keys);
			}
		}
		g_key_file_free(file);

	#endif

} 

void Application::loadSettings() {

	loadKeys(&settings);

	EnsureKeyString(&settings, "window_width", "800");
	EnsureKeyString(&settings, "window_height", "450");
	EnsureKeyString(&settings, "window_maximized", "0");
	EnsureKeyString(&settings, "hpaned", "160");
	EnsureKeyString(&settings, "hpaned_ce", "400");
	EnsureKeyString(&settings, "vpaned_hp", "200");
	EnsureKeyString(&settings, "hier_split", "90");
	EnsureKeyString(&settings, "expl_split", "90");

	EnsureKeyString(&settings, "$supp_msg", "false");

	updateIconPath();

}

void Application::saveSettings(bool quit) {

	AppWindow::State state= window.getState();
	SetKeyInt(&settings, "window_width", state.width);
	SetKeyInt(&settings, "window_height", state.height);
	SetKeyInt(&settings, "window_maximized", state.maximized);

	Settings skeys;
	loadKeys(&skeys);

	if(quit)
		for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it) {
			Check(!it->first.empty());
			if(it->first[0] == '$') {
				Settings::iterator jt= skeys.find(it->first);
				if(jt!=skeys.end())
					it->second= jt->second;
			}
		}

	if(skeys == settings)
		return;

	#ifdef CROW_WIN32

		HKEY hKey;
		LONG result= RegCreateKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\crow\\settings", 0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hKey, NULL);
		if(result==ERROR_SUCCESS) {
			for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it) {
				const std::string & key= it->first;
				const Glib::ustring & str= it->second;
				RegSetValueEx(hKey, key.c_str(), 0, REG_SZ, reinterpret_cast<const BYTE *>(str.c_str()), str.size()+1);
			}      
			RegCloseKey(hKey);
		} else
			showMessage("Saving settings failed");

	#else

		Glib::ustring path= Glib::ustring(g_get_home_dir()) + "/.crow";
		GKeyFile * file= g_key_file_new();
		for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it)
			g_key_file_set_string(file, "settings", it->first.c_str(), it->second.c_str());
		gsize length;
		gchar * data= g_key_file_to_data(file, &length, NULL);
		if(!g_file_set_contents(path.c_str(), data, length, NULL))
			showMessage("Saving settings failed");
		g_free(data);
		g_key_file_free(file);

	#endif

}

void Application::addRecentFile() {
	Gtk::RecentManager::get_default()->add_item("file://" + file.name);
}

void Application::showMessage(const Glib::ustring & message) {
	Gtk::MessageDialog dialog(window, message);
	dialog.run();
}

void Application::showUrl(const Glib::ustring & url) {
	#ifdef CROW_WIN32
		HINSTANCE result= ShellExecute(NULL, NULL, url.c_str(), NULL, NULL, SW_SHOWNORMAL);
		if(reinterpret_cast<UINT_PTR>(result) <= 32)
			showMessage("Error opening url:\n" + url);
	#else
		int result= -1;
		try {
			Glib::spawn_command_line_sync("xdg-open '" + url + "'", NULL, NULL, &result);
		} catch(const Glib::Error & x) {
			result= -1;
		}
		if(result!=0)	
			showMessage("Error opening url:\n" + url);
	#endif
}

void Application::designer_apply_settings(const Settings & settings) {	
	GHashTable * table= SettingsToHashTable(settings);
	crow_designer_apply_settings(designer, table);
	g_hash_table_destroy(table);
}

void Application::designer_store_settings(Settings * settings) {
	GHashTable * table= SettingsToHashTable(*settings);
	crow_designer_store_settings(designer, table);
	*settings= HashTableToSettings(table);
	g_hash_table_destroy(table);
}

void Application::designer_clear() {
	crow_designer_clear(designer);
}

void Application::designer_load(const Glib::ustring & data) {
	GError * error= NULL;
	gboolean result= crow_designer_load(designer, data.c_str(), &error);
	if(!result) {
		Glib::ustring msg= error->message;
		g_error_free(error);
		throw Glib::Error(CROW_ERROR, CROW_ERROR_RUNTIME, msg);
	}
}

Glib::ustring Application::designer_save() {
	Glib::ustring result;
	gchar * data= crow_designer_save(designer);
	result= data;
	g_free(data);
	return result;
}

Glib::ustring Application::designer_save_header() {
	Glib::ustring result;
	gchar * data= crow_designer_save_header(designer);
	result= data;
	g_free(data);
	return result;
}

void Application::designer_saved() {
	crow_designer_saved(designer);
}

Gtk::Widget * Application::designer_get_widget() {
	return Glib::wrap(crow_designer_get_widget(designer));
}

CrowActions Application::designer_get_enabled_actions() {
	return crow_designer_get_enabled_actions(designer);
}

void Application::designer_run_action(CrowActions action) {
	crow_designer_run_action(designer, action);
}

}
