/*
 *  Crow, a GUI designer for GTK+
 *  Copyright (c) 2007 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <crow.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include "container.h"

#include "glue.h"
#include "marshal.h"

void CheckFailed(const gchar * expr, const gchar * file, unsigned line) {
	gchar msg[256];
	#ifdef CROW_SANDBOX
		assert(0);
	#endif
	g_snprintf(msg, sizeof(msg), "Check failed: %s, file: %s, line: %d", expr, file, line);
	FatalError(msg);
}

void FatalError(const gchar * message) {
	GtkWidget * dialog;
	g_critical(message);

	dialog= gtk_message_dialog_new(
		NULL,
		GTK_DIALOG_MODAL,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_CLOSE,
		"Fatal Error [" G_LOG_DOMAIN "]"
	);

	gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog), message);
	gtk_window_set_keep_above(GTK_WINDOW(dialog), TRUE);
	gtk_dialog_run(GTK_DIALOG(dialog));
	gtk_widget_destroy(dialog);

	abort();
}

GType crow_container_get_type(void) {
	static GType object_type= 0;
	if(object_type==0) {
		static const GTypeInfo object_info= {
			sizeof(CrowContainerClass), 
			NULL, 
			NULL,	
			(GClassInitFunc)crow_container_class_init,
			NULL, 
			NULL, 
			sizeof(CrowContainer), 
			0,
			(GInstanceInitFunc)crow_container_init,
		};
	  object_type= g_type_register_static(G_TYPE_OBJECT, "CrowContainer", &object_info, 0);
	}
	return object_type;
}

void crow_container_class_init(CrowContainerClass * meta) { 
	GObjectClass * object_class;
	meta->parent_class= g_type_class_peek_parent(meta);
	object_class= G_OBJECT_CLASS(meta);
	object_class->finalize= crow_container_finalize;

	g_signal_new(
		"actions_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_INT64,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_INT64
	);
	g_signal_new(
		"status_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"info_updated",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"show_url",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		crow_marshal_VOID__INT_STRING,
		G_TYPE_NONE, 2,
		G_TYPE_INT, G_TYPE_STRING
	);
	g_signal_new(
		"modal_enter",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1,
		G_TYPE_INT
	);
	g_signal_new(
		"modal_leave",
		G_OBJECT_CLASS_TYPE(object_class),
		G_SIGNAL_RUN_LAST, 0,
		NULL, NULL,
		g_cclosure_marshal_VOID__INT,
		G_TYPE_NONE, 1,
		G_TYPE_INT
	);

}

void crow_container_init(CrowContainer * obj) { 
	ContainerInit(obj);
}

void crow_container_finalize(GObject * object) {
  CrowContainer * container= CROW_CONTAINER(object);
  CrowContainerClass * meta= CROW_CONTAINER_GET_CLASS(container);
	ContainerDone(container);
	(*meta->parent_class->finalize)(object);
}

CrowContainer ** ContainerPtr(void) {
	static CrowContainer * ptr= NULL;
	return &ptr;
}

CrowContainer * ContainerGet(void) {
	return *ContainerPtr(); 
}

void ContainerEnter(void) {
	dbus_g_object_type_install_info(CROW_TYPE_CONTAINER, &dbus_glib_crow_container_object_info);
  dbus_g_error_domain_register(CROW_ERROR, NULL, CROW_TYPE_ERROR);
	*ContainerPtr()= g_object_new(CROW_TYPE_CONTAINER, NULL); 
}

void ContainerLeave(void) {
	g_object_unref(ContainerGet());
	*ContainerPtr()= NULL;
}

void ContainerInit(CrowContainer * obj) {
	obj->counter= 1;
	obj->designers= g_hash_table_new_full(NULL, NULL, NULL, RemoveDesigner);
	gtk_icon_theme_get_search_path(gtk_icon_theme_get_default(), &obj->defIconPath, NULL);
}

void ContainerDone(CrowContainer * obj) {
	g_strfreev(obj->defIconPath);
	g_hash_table_remove_all(obj->designers);
	g_hash_table_destroy(obj->designers);
	#ifdef CROW_SANDBOX
		g_print("ContainerDone()\n");
	#endif
}

CrowDesigner * CreateDesigner(GdkNativeWindow handle) {
	CrowDesigner * designer= crow_designer_new();
	GtkWidget * widget= crow_designer_get_widget(designer);
	GtkWidget * plug= gtk_plug_new(handle);
	gtk_container_add(GTK_CONTAINER(plug), widget);
	return designer;
}

void RemoveDesigner(gpointer value) {
	CrowDesigner * designer= CROW_DESIGNER(value);
	g_object_unref(designer);
}

CrowDesigner * FindDesigner(CrowContainer * obj, gint id) {
	CrowDesigner * designer;
	Check(obj && id>0);
	designer= g_hash_table_lookup(obj->designers, GINT_TO_POINTER(id));
	Check(CROW_IS_DESIGNER(designer));
	return designer;
}

GHashTable * CloneSettings(GHashTable * settings) {
	GHashTable * result= g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
	g_hash_table_foreach(settings, CloneSetting, result);
	return result;
}

void CloneSetting(gpointer key, gpointer value, gpointer result) {
	g_hash_table_insert(result, g_strdup(key), g_strdup(value));
}

gchar ** VStrMerge(gchar ** vstr1, gchar ** vstr2) {
	int len1= g_strv_length(vstr1);
	int len2= g_strv_length(vstr2);
	gchar ** result= g_new(gchar *, len1+len2+1);
	int i;
	for(i=0; i<len1; ++i)
		result[i]= g_strdup(vstr1[i]);
	for(i=0; i<len2; ++i)
		result[i+len1]= g_strdup(vstr2[i]);
	result[len1+len2]= NULL;
	return result;
}

gboolean VStrEqual(gchar ** vstr1, gchar ** vstr2) {
	int len1= g_strv_length(vstr1);
	int len2= g_strv_length(vstr2);
	int i;
	if(len1!=len2)
		return FALSE;
	for(i=0; i<len1; ++i)
		if(strcmp(vstr1[i], vstr2[i])!=0)
			return FALSE;
	return TRUE;
}

void OnActionsUpdated(CrowDesigner * obj, CrowActions changed, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "actions-updated", id, changed);
}

void OnStatusUpdated(CrowDesigner * obj, const gchar * status, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "status-updated", id, status);
}

void OnInfoUpdated(CrowDesigner * obj, const gchar * info, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "info-updated", id, info);
}

void OnShowUrl(CrowDesigner * obj, const gchar * url, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "show-url", id, url);
}

void OnModalEnter(CrowDesigner * obj, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "modal-enter", id);
}

void OnModalLeave(CrowDesigner * obj, CrowContainer * container) {
	gint id= GPOINTER_TO_INT(g_object_get_data(G_OBJECT(obj), "CrowContainerId"));
	g_signal_emit_by_name(container, "modal-leave", id);
}

gboolean crow_container_get_version(CrowContainer * obj, gint * version, GError ** error) {
	*version= CROW_VERSION;
	return TRUE;
}

gboolean crow_container_shutdown(CrowContainer * obj, GError ** error) {
	gtk_main_quit();
	return TRUE;
}

gboolean crow_container_create(CrowContainer * obj, gint64 handle, gint * id, GError ** error) {
	CrowDesigner * designer= CreateDesigner((GdkNativeWindow)handle);
	gpointer key= GINT_TO_POINTER(obj->counter);
	g_hash_table_insert(obj->designers, key, designer);
	*id= obj->counter;
	++obj->counter;
	g_signal_connect(designer, "actions-updated", G_CALLBACK(OnActionsUpdated), obj);
	g_signal_connect(designer, "status-updated", G_CALLBACK(OnStatusUpdated), obj);
	g_signal_connect(designer, "info-updated", G_CALLBACK(OnInfoUpdated), obj);
	g_signal_connect(designer, "show-url", G_CALLBACK(OnShowUrl), obj);
	g_signal_connect(designer, "modal-enter", G_CALLBACK(OnModalEnter), obj);
	g_signal_connect(designer, "modal-leave", G_CALLBACK(OnModalLeave), obj);
	g_object_set_data(G_OBJECT(designer), "CrowContainerId", key);
	return TRUE;
}

gboolean crow_container_remove(CrowContainer * obj, gint id, GError ** error) {
	gboolean removed= g_hash_table_remove(obj->designers, GINT_TO_POINTER(id));
	Check(removed);
	return TRUE;
}

gboolean crow_container_apply_settings(CrowContainer * obj, gint id, GHashTable * settings, GError ** error) {
	CrowDesigner * designer= FindDesigner(obj, id);
	GtkWidget * plug= gtk_widget_get_parent(crow_designer_get_widget(designer));
	gboolean visible= GTK_WIDGET_VISIBLE(plug);
	
	gchar * folders= g_hash_table_lookup(settings, "$icon_folders");
	gchar ** custom= g_strsplit(folders ? folders : "", "*", 0);
	gchar ** path= VStrMerge(obj->defIconPath, custom);
	gchar ** current= NULL;
	gtk_icon_theme_get_search_path(gtk_icon_theme_get_default(), &current, NULL);
	if(!VStrEqual(path, current))
		gtk_icon_theme_set_search_path(gtk_icon_theme_get_default(), (gpointer)path, g_strv_length(path));
	g_strfreev(custom);
	g_strfreev(path);
	g_strfreev(current);
	
	if(!visible) {
		gchar * winw= g_hash_table_lookup(settings, "window_width");
		gchar * winh= g_hash_table_lookup(settings, "window_height");
		if(winw && winh)
			gtk_window_set_default_size(GTK_WINDOW(plug), atoi(winw), atoi(winh));
	}

	crow_designer_apply_settings(designer, settings);

	if(!visible)
		gtk_widget_show(plug);

	return TRUE;
}

gboolean crow_container_store_settings(CrowContainer * obj, gint id, GHashTable * original, GHashTable ** settings, GError ** error) {
	*settings= CloneSettings(original);
	crow_designer_store_settings(FindDesigner(obj, id), *settings);
	return TRUE;
}

gboolean crow_container_clear(CrowContainer * obj, gint id, GError ** error) {
	crow_designer_clear(FindDesigner(obj, id));
	return TRUE;
}

gboolean crow_container_load(CrowContainer * obj, gint id, const gchar * data, GError ** error) {
	return crow_designer_load(FindDesigner(obj, id), data, error);
}

gboolean crow_container_save(CrowContainer * obj, gint id, gchar ** data, GError ** error) {
	*data= crow_designer_save(FindDesigner(obj, id));
	return TRUE;
}

gboolean crow_container_save_header(CrowContainer * obj, gint id, gchar ** data, GError ** error) {
	*data= crow_designer_save_header(FindDesigner(obj, id));
	return TRUE;
}

gboolean crow_container_saved(CrowContainer * obj, gint id, GError ** error) {
	crow_designer_saved(FindDesigner(obj, id));
	return TRUE;
}

gboolean crow_container_get_enabled_actions(CrowContainer * obj, gint id, gint64 * actions, GError ** error) {
	*actions= crow_designer_get_enabled_actions(FindDesigner(obj, id));
	return TRUE;
}

gboolean crow_container_run_action(CrowContainer * obj, gint id, gint64 action, GError ** error) {
	crow_designer_run_action(FindDesigner(obj, id), action);
	return TRUE;
}
